// Global Header file of the CForth  project **********************************
//
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "main.h", "cforth.c",
// "cforth.h", "dmx.c" and "usb-fs-431.c"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#ifndef GLOBALDEFINES
#define GLOBALDEFINES

#define REVNUM  63			//describes firmware revision

//#define MINI  //compact hardware with DMX but no RS-232
//#ifndef MINI: standard PCB layout with option RS_232, DMX or not

#ifdef MINI
//ifndef MICRO: more expanded layout on Veroboard

//#define MICRO				//small PCB assembly --not published anymore
							//PA0,1,2,3,9,10 not accessible as UserI/O !
//							//Hex code and DIY instruction not published anymore
							//PB5 and PA15 exchanged with respect to older versions
							//major difference else is MOT1,2 pin assignment

#define HARDWARE_VERSION 3	
#define KERNELSIZE		159
#define DMXFIRST		139

#define DMX		//don't comment out!! MINI always uses DMX !!

//#define CPU_CLOCK144
//ifndef: CPU Clock is 48 MHz (default)

#else  //not MINI
//EXACTLY one of these options must be UNCOMMENTED EXCLUSIVELY !
//#define STANDARD		//Veroboard without DMX
#define STANDARD_DMX	//Veroboard w.DMX additional feature
//#define LQFP48		//expanded IO access (e.g.STM32G441CBT6),
						//else same code as STANDARD
						//uses USART3 instead of STANDARD USART1 (PB10,11)
//#define LQFP48_DMX	//expanded IO access (e.g.STM32G441CBT6),
						//else same code as STANDARD_DMX
						//uses USART3 instead of STANDARD USART1 (PB10,11)
#endif

#ifdef STANDARD
#define HARDWARE_VERSION 7	//describes project specific hardware version
#define CPU_CLOCK144
//ifndef: CPU Clock is 48 MHz (default)
#define KERNELSIZE		150
#endif

#ifdef STANDARD_DMX
#define HARDWARE_VERSION 7	//describes project specific hardware version
#define DMX			//else DMX is not defined and not included in firmware
//#define CPU_CLOCK144
//ifndef: CPU Clock is 48 MHz (default)
#define KERNELSIZE		169
#define DMXFIRST		149
#endif

#ifdef LQFP48			//GreyPill=BlackPill compatible LED is PB15
#define CASE48

//#define CPU_CLOCK144
//ifndef: CPU Clock is 48 MHz (default)

//BlackPill compatible PCB layout
//special hardware features:
#define EXTERN_TXINVERT	//w.external 74HC14 inverter,STM32F411 compatible
#define HARDWARE_VERSION 3	//describes project specific hardware version
//no Vusb Test
//USART1 ,PA9/10TX inverted

#define KERNELSIZE		150
#endif

#ifdef LQFP48_DMX	//GreyPill=BlackPill compatible LED is PB15
#define CASE48
#define DMX			//else DMX is not defined and not included in firmware
//#define CPU_CLOCK144
//ifndef: CPU Clock is 48 MHz (default)

//BlackPill compatible PCB layout
//special hardware features:
#define EXTERN_TXINVERT	//w.external 74HC14 inverter,STM32F411 compatible
#define HARDWARE_VERSION 3	//describes project specific hardware version
//no Vusb Test
//USART1 ,PA9/10TX inverted
//DMX: USART2 PA2 (BlackPill compatible), DMX-BREAK PB0
#define KERNELSIZE		169
#define DMXFIRST		149	//line break in OPS printout
#endif

//the device will appear as "STMicroelectronics Virtual Com Port"
#define DEFAULT_VIDH	0x04
#define DEFAULT_VIDL	0x83
#define DEFAULT_VID		((DEFAULT_VIDH << 8) | DEFAULT_VIDL)
#define DEFAULT_PIDH	0x57
#define DEFAULT_PIDL	0x40
#define MIDI_PLUS	3	//added to actual VID/PID in USB is MIDI interface
#define DEFAULT_PID		((DEFAULT_PIDH << 8) | DEFAULT_PIDL)

#define DEFAULT_NUMBASE	10 //Decimal, enter 16 for HEX
#define DEFAULT_BAUD	11	//expanded to 115200, 
							//enter first 2 places, e.g. 96 for 9600

//attributes of KernelOperators
//attribute entries are merged with some other functional flags !!
#define A_IME		0x10000 //ATTRIBUTE_IMMEDIATE, cannot be compiled
#define A_CPO		0x20000 //ATTRIBUTE_COMPILEONLY,refuse work interpreting
#define A_BAK		0x40000 //may be parsed in background at executing state
#define A_CI		(A_CPO | A_IME) //COMPILEONLY IMMEDIATE
							//performs complex operations during compilation
#define A_IB		(A_IME | A_BAK)
#define A_RUN		0x80000 //ATTRIBUTE_RUNTIME,
				//compiled by other operators for special runtime behaviour
				//no direct user access, essential for viewing SEE and STEP
//only attribute of UserToken
#define A_VARCON	0xC0000000

#define ERR_STRINGTOKEN	1
#define ERR_SYNTAX		2
#define ERR_STACK		4
#define ERR_COMPONLY	8
#define ERR_SYSTEM		0x10
#define ERR_USERCODE	0x20
#define ERR_COMMACC		0x40
#define ERR_CONTEXT 	0x80
#define ERR_USERCONTEXT	0xA0	//combination of both, save code memory
#define ERR_USERSTRING	0x21	//combination of both, save code memory
#define ERR_CANCEL		0x100
#define ERR_RUNTOKEN	0x200
#define ERR_NUMVAL		0x400
#define ERR_DIVZERO		0x800

#define MAXUSER			168//*6 words p.struct=1008 words=4032 bytes=ca.2pages
#define MAXSTACK		64
#define MAXCODE			0xC00	// =3072 words =12288 bytes
								//takes 6 pages each 2048 bytes
#define CODETOP			MAXCODE - 1
#define USERTOP			(uint32_t)&gfUserHeader[MAXUSER]

#define CDC_CLASS		0		//initalizes device as virtual COM port
#define MIDI_CLASS		1		//initalizes device as native MIDI interface

#ifdef DMX
#define RXBUFLEN 512 //768
#define MIDI_BUFLEN 128
#else
#define RXBUFLEN 1024
#define MIDI_BUFLEN 256
#endif
#define DMX_BUFLEN 512 // index 0 is DMX addr. 1, TX from index 0 to 512
						//formally necessary for compiler even w.o. DMX

#define CTRL_B  2
#define CTRL_C  3
#define BKSP  8
#define TAB   9
#define LFF  10		//for Macintosh OS: simply exchange values of CRR and LFF
#define CRR  13		//then New_Line will be interpreted correctly
#define CTRL_R  18
#define ESC  27
#define SPACE  32
#define BACKSLASH 92

//not relevant for this project, but necessary for compilation:
#define EOX		0xF7
#define RUNNINGSTATE  //uncomment if not wanted
//works only on MIDI msgs which are transmitted from MIDI OUT
//Running State is always supported with MIDI msgs received at MIDI IN

//bitmask shortcuts: (used in usb-fs-431.c)
#define m0      1
#define m1      (1<<1)
#define m2      (1<<2)
#define m3      (1<<3)
#define m4      (1<<4)
#define m5      (1<<5)
#define m6      (1<<6)
#define m7      (1<<7)
#define m8      (1<<8)
#define m9      (1<<9)
#define m10     (1<<10)
#define m11     (1<<11)
#define m12     (1<<12)
#define m13     (1<<13)
#define m14     (1<<14)
#define m15     (1<<15)
#define m16     (1<<16)
#define m17     (1<<17)
#define m18     (1<<18)
#define m19     (1<<19)
#define m20     (1<<20)
#define m21     (1<<21)
#define m22     (1<<22)
#define m23     (1<<23)
#define m24     (1<<24)
#define m25     (1<<25)
#define m26     (1<<26)
#define m27     (1<<27)
#define m28     (1<<28)
#define m29     (1<<29)
#define m30     (1<<30)
#define m31     (1<<31)

#endif  //GLOBALDEFINES

// ##########################################################################

// Version changes:
//REVNUM 1: 03April2020 copied from old PC-based C-project.
//	STM32G431KB MCU specific parts added
//	16May20: HARDWARE_VERSION 2: USART1 is PA10&PB6, VUSB detection PA9
//	15June20: HARDWARE_VERSION3: alternative Veroboard compatible design
//	VUSB detection:PA8, LED:PA9
//REVNUM43: 20June2020 first published release. Intermediate work not documented
//REVNUM44: 22June20: 2 CONSTants,2VARiables precompiled as User Token 
//	25June20: bugfix "W":parameters exchanged
//	VAL? renamed to VAL and may be compiled, too
//REVNUM45: 02JULY2020: some debug messages removed
//	bugfix and improvements in save and load project in Flash
//	default number base changed to Decimal
//	minor improvements en detail
//REVNUM46: 03July2020:  RANDOM number generator new
//	05July20:TFREQ,PWM,TIME,PULS,RANDOM, I2C, StackFlush new	
//	08July20: doCONST renamed doVALUE, doVAR renamed doADDR
//	11July20: P" and TX" bytes stored more compact
//  PWM1(PA6) and PWM2(PA7) independent setup but same TFREQ
//	15July2020: DMX parts added from STM32-USB-DMX module w.STM32F042
//	and modified for data input in Forth postfix notation.
//	HARDWARE_VERSION 4, HW version3 enhanced with DMX output
//	16July20: indices of kernel operators restructured
//	17July20: circuit design and firmware for "MINI" version
//	18July20: all versions (+MINI, +DMX) merged together
//released 20July2020
//REVNUM47: 01August2020
//	01Aug20	USB and USART transmit buffers overflow at 144MHz clock fixed
//	03Aug20	DMX op "XCYC" new (number of transmitted "slots" per DMX cycle)
//	05Aug20	reorg. of op "Q" (message of system parameters)
//REVNUM48: 09August2020
//			SPI init parameters added to output of "Q"
//			send0X32() new, code more compact
//	13Aug20	bugfix in interpretUserToken(): 
//			TIB pointer was not restored after background process
//			improvement of FREQ initialisation, printout with op "Q"
//			problem gfSpiClock initialisation fixed, SpiClock limit=0xFFFF
//			SPI and PWM storage in Flash reorganized
//			for MINI PA13 as I/O allowed
//	16Aug20	terminal text interpreter converts TAB into SPACE
//release 16Aug20
//REVNUM49: 31 October2020
//	07Nov20	return value of "interpretUserToken()" inverted
//			return of "interpretKernelToken()" changed to "void"
//			return of "executeTCode()" changed to "void"
//			opNODBG introduced,background process not executed,higher speed
//			if USART RX (PA10) shorted during startup, AUTOEXE->"none"
//			(MINI: pull PA3 low during startup, AUTOEXE->"none")
//	08Nov20	subQuery() rejects (TIB bytes == 0), this simplifies scanNext()
//			return of scanNext() changed to "void", improvement of op_AUTO
//	18Nov20 Option MINI in many cases replaced by RS_232.Option VUSB_ONLY new.
//			Code restructured correspondingly
// 	23Nov20	Kernel Ops DEBUG, CASE, <CASE new. ENDIF replaced by THEN
// 	24Nov20	MIDI features improved
//release 03Dec20
//REVNUM50:
//	07Dec20	executeTcode etc. optimized for runtime improvement
//			Z-BREAK renamed ZBREAK and doZ-BRK renamed doZBRK
//			 IF()...LOOP() and doIF()...doLOOP() improved
//			for better runtime behaviour and better readability of MEM()
//			;(SEMIS) optimized
//	10Dec20	pullup/down switched off when I/O is output (save current)
//	12Dec20	user terminal communication during NO-DBG improved (CTRL_D)
//	15Dec20	parameter format of sendX8() changed to uint32_t
//	16Dec20	time operators reorg. (MS, DS are blocking, BLOCK removed)		
//          improvement of CTRL_B (unblock RX, MS and DS) behaviour
//	23Dec20	USART Tx inverted (no external inverter) by default
//			(MINI hardware not concerned)
//			HARDWARE_VERSION changed to 5
//			//BAUD param.changed:add 100(dec) or 0x100(hex) for ext.Inverter
//	27Dec20	MiniDmx startup and ACK reorganized, only USB Com supported now
//release 28Dec20
//REVNUM51:
// 01Jan21	Bugfix Baud Rate storage in Flash
//release 01Jan21
//REVNUM52:
// 20Jan21:	bugfix "RANA": assignment of ADC channels and number 
//release 22Jan21
//REVNUM53:
//	MAY2021
//	development system changed to "Segger Embedded Studio for ARM 5.44"
//	no changes of source code and project setup necessary
//	option VUSB_ONLY removed
//	new option LQFP48 for STM32G431,STM32G441 and STM32G474 w.48pin LQFP case
//	 (different pins for USART and VUSB det,more I/O supported.Else same code)
//	new kernel ops ROLL, INJECT, U>.  ops 2SWAP and 2DUP removed
//	op_REV behaviour modified: can be compiled now
//	start behaviour of op AUTO improved
//	19MAY21	Kernel ops ROLL,INJECT,U> added, 2SWAP,2DUP removed
//	22MAY21	%X and %D "Immediate,Background" now
//			op VAL reduced to Background Process value message
//	25May21	op_QUEST and AUTOEXE check moved to "popFlashProject" 
//			for better start of AUTOEXE when another project is loaded
//	29May21	HARDWARE_VERSION changed to 6: PA8 and PA exchanged
//			for better fitness with BOOT0 and easier wiring
//			automatic ADC oversampling introduced
//	30May21	PICK, ROLL, INJECT modified to Forth79 style
//			FS ("flush data stack") modified to "NO attributes"
//released 31MAY2021
//REVNUM 54:
//21Jun21	option XRGB added, option LQFP48_DMX added
//			number of DMX scenes in Flash increased to 96
//			for this reason only projects 0,1,2 available for DMX versions.
//13ul21	bugfix LQFP48 portA:op_PINOH,op_PINOL,op_PINIPU,op_PINIPD,op_PINIZ
//			LQFP48:+VREF must be connected with VDDA,else op_DOA4,5 won't work
//			clearance of LED pin connection, especially STM32G431 "Mini"
//12Aug21	bugfix DAC initialisation
//21Sept21	LED of MINI versions changed to PA8 (HARDWARE_VERSION -->3)
//			new MINI Veroboard compatible hardware, - same firmware
//			option XNRGB added
//23Oct21	bugfix: change USB_CLASS following op_SPLIT()
//release 24Oct21
//REVNUM 55: 26 Oct 2021
//11Nov21	USB stability improved (SET_LINE_CODING)
//			new:Stepper motor drive as background task
//12Nov21   bugfix load empty project with modified globals
//13Nov21   MOT1,MOT4,N1,MOT2,N2 new (background motor control)
//			op_VAL improved(not restricted to STEP mode anymore,but IMMEDIATE)
//			PWM3,4 (PA15,PA1) new
//23Nov21	STANDARD new HARDWARE_VERSION 7:
//			CN4,CN5,PB0,PA0,PA3,DMX wiring changed
//			Bugfix op_XCYC (DMX cycle length)
//			STANDARD_DMX: DMX_SPACE control modified: now PA0. Option OLDIE:PA3
//04Dec21	bugfix and improvement subPrintString()
//			improvement op_SPLIT(), emptyForth();
//release 06 Dec 2021
//REVNUM56 13Dec2021
//			op_MOT1(), subMOT1()simplified
//14Dec21	AUTOEXE bahaviour changed (-1 removed, index not increased)
//			gfUserToken[0] not deletable
//15Dec21	reorganisation of MOT1,MOT4,MOT2 code
//			Bugfix OH,OL,IPU,IPD,IZ,RANA of STANDARD_DMX(PA0,PA2 vs.PA2,PA3)
//16Dec21	bugfix "popFlashProject", gfPWM1,2
//19Dec21	code of PWM1,2,3,4 and code of MOT1,4,2 straightened
//			checkBack() new, control of background processes centralized
//release 21Dec21
//REVNUM57 04Jan2022
//28Apr22	except MINI version, RS-232 always included
//			(option '#ifdef RS_232' removed, replaced by '#ifndef MINI')
//PWM Freq bei CPU_CLOCK144 testen = ok, aber warum ??
//06May22   %x renamed $, %d renamed %
//08May22   preloaded V2 removed, now P1,P2,P3 preloaded
//16May22	op_PRINTSTACK modified (CRR removed)
//27May22	string output of op_USER modified
//28May22	gfTICKCOUNT -> gfMSCOUNT and gfTIXCOUNT (decouple TIX from  MS,DS)
//07June22	VAR and CONST replaced by VARCON,
//			op_REV compiles op_doREV instead of op_doADDR, saves runtime
//			op_VAL removed, op_TRACE new
//			P1,P2,P3 renamed to K1,K2,K3,K4
//			due to these changes, some reorganisation of code
//09June22	struct "gfUserToken" renamed to "gfUserHeader"
//10June22	option "USART_TX_INTERRUPT" removed, has never been used
//			MY1..MY4 finally included
//			kernel op TERM? renamed to KEY?
//11June22	bugfix interpretNumber(), parsing big negative numbers
//			CTRL_B/CTRL_D handler in USB from pullUsbRxBuf() to servEP2_Out()
//			kernel op ARRAY added
//13June22	improvement of compileUserToken() and restore()
//							(gfTCode offset calculation made more transparent)
//15June22	background process behaviour of KEY? and KEY improved
//release 17June22
//REVNUM58:	18June2022
//22March23	kernel op "NEG" renamend to "+/-"
//			kernel op "=" renamend to "=="
//30Sept23	kernel op "%" renamend to "DZ"
//			kernel op "$" renamend to "HX"
//02Oct2	kernel ops UNTIL and correseponding doUNTIL added
//			not MIDI version: PICK replaced by ROLL
//			option "Cold Reset" (CTRL_R) removed
//			independent Watchdog removed
//07Oct23	"RANA"(raed analog in) improved, new: disableADC(),saves current
//			P renamed . , PH renamed .H , P" renamed .", .D  implemented new
//			kernel Op 'Q' renamed to '?'
//			>CASE and do>CASE implemented new
//07Oct23	numberBase saved in general preset, always reactived at startup
//15Oct23	pin arrangement for PWM and MOT improved
//			kernel op '&' renamed to 'AND', kernel op '|' renamed to 'OR
//16Oct23	kernel ps NOP. TIME, PULS removed
//			MOT code rearranged.Now MOT1,2,3 each 2 wire.Option 4wire removed
//17Oct23	USER output reformatted
//18Oct23	REPLACE and RESTORE improved
//20Oct23	MOT1,2,3 code improved
//21Oct23	REGW, REGBSET, REGBCLR,REGBTST stack parameters exchanged
//25OCT23	RANA renamed RADC, RDIG renamed RDIN, WAIT renamed TIME
//26Oct23	kernel ops "<<" and ">>" replaced by single "SHIFT"
//			"VID" , "PID" combined to "VIDPID"
//28Oct23	CONTI, BREAK, ZBREAK optimized, doCONTI,doBREAK,doZBREAK removed
//			K1,K2,K3,K4 renamed V1,V2,V3,V4
//30Oct23	servSystem() transferred from checkTerminal() to showPreStep()
//09Nov23	order of Kernel Ops rearranged
//12Nov23	RP-1 and RP-4 only "COMPILING" flag now.To be used ahead of RETURN
//release 12Nov23
//REVNUM 59:13Nov2023
//19Dec23	MEM attributes changed to A_BAK
//20Dec23	bugfix compileTCode: protection against CODETOP overflow
//21Dec23	digital port I/O improved
//released	25Dec23
//REVNUM->60
//31Dec23	modification compilation user header with kernel name (op_colon)
//			new gfBackProcess=2: return to higher thread level with CTRL_R
//01Jan24	changed: internally gfAuto = target UserFunction + 1
//			op_MEM now sends block of 128 bytes(32 words)
//			improvement countdown MSTP1,2,3 if stepper is not rotating
//04Jan24	PWM3 output now PB3, PWM4 output.now PA5
//			LQFP48: pins for LED and Vusb detect changed HARDWARE_VERSION-->2
//			op_MOTx pins changed/unified for all HW versions
//11Jan24	send0X32() modified: always with leading 0x even at small numbers
//			query() made void
//12Jan24	bugfix:popFlashProject()/AUTOEXE was called before USB initialized
//13Jan24	USB setup with user modified VID/PID improved
//			MIDI_PLUS for MIDI interface PID increment introduced
//released 14Jan24
//REVNUM-->61
//18Jan24	bugfix op_SEE()
//19Jan24	decimal numbers can be entered with leading '%' at any muber base
//			see interpretNumber()
//21Jan24	servSystem() updates DMX fader with 100Hz speed
//08Feb24	kernelOps ITX1 and IRXN implemented. IRXAN renamed to IADRXN
//24Feb24	MiniDMX option removed
//26Feb24	program flow IADRXN improved
//27Feb24	first completely working version of GREYPILL
//07March24	kernel op RADC: only 1 single sample used for readout
//09March24 flash storage of gdFadeTime and gfBootScene simplified & improved
//			reorg kernel op SAVE and LOAD more Forth typically
//released 14March24
//REVNUM-->62
//31March24	bugfix cforth.h: DZ and HX have IMMEDIATE flag now
//02Apr24	doVALUE renamed doVAL
//released 09April202424
//REVNUM-->63
//04Apr24	op_MOT modified: NSTP=-2-->permanent rotation
//20Apr24	doLOOP() end index behaviour changed
//06Sept24	GreyPill LED changed to PB15 (PA8 reserved for MCO output)
//22Dec24	old "BluePill" compatible version (unpublished) not supported any more
//			RS-232 now USART1 in any case (GreyPill PA9/PA10, Mini PA2,PA3, else PB6/PA10)
//			DMX output PA2 now (F411 compatible), except MiniDMX: here PB6
//			DMX Break:PA0(Std),PB0(case48), PA15(Mini)
//27Dec24	FlashPage23: order gfNumbase, gBaud exchanged
//			LED not accessible as user I/O anymore
//			DMX option OLDIE cancelled, option MICRO "unpublished" now
//29Dec24	checkSpecialIO() renamed checkPWM()
//			subPWMx() reorganized
//04Jan25	op_LED()
//			servSystem(): LED blink behaviour changed for better user access
//			op_SPI() simplified
//16Jan25	{"REV"} renamed {"W"}. Previous kernel operators "W" and "R" removed
//			op_DOREV() renamed op_WVA()
//19Jan25	VARCONs directly stored in user code. op_VALUE unnecessary/removed
//20Jan25	executeTCode() speed optimization
//25Jan25	bugfix Mini/PA10 for  gfQAuto in popFlashProject()
//			bugfix op_READDIG: wrong pinNo filter mask
//26Jan25	code modifications in executeTCode() and op_ABORT
//01Feb25	DEBUG renamed BACKOP, NO_DBG renamed NOBACK
//			CTRL_B->CTRL_C, CTRL_D->CTRL_B
//07Feb25	XBUS, gfDmxBusy introduced: speed up kernel when DMX is not needed

// *****************************

