// CForth  project ************************************************************
//
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.h","cforth.c","cforth.h",
// "usb-fs-431.c", "dmx.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"
#include "main.h"

int main(void)    // ********************************************************
// **************************************************************************
{
uint32_t k;

	RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;	//necessary for what?? 
											//PWREN of the other busses?
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOAEN;	// Enable Port A clock
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOBEN;	// Enable Port B clock
#ifdef CASE48
	RCC->AHB2ENR |= RCC_AHB2ENR_GPIOCEN;	//Enable Port C clock
	GPIOC->MODER = 0x00000000;		//primarily all input	
	GPIOC->PUPDR = 0xA8000000;		//primarily all inputs pulldown
#endif
	GPIOA->MODER = 0x00000000;		//primarily all input 
	GPIOB->MODER = 0x00000000;		//primarily all input
#ifdef MINI
  #ifdef MICRO
	GPIOA->PUPDR = 0xA802AA6A;		//primarily all inputs pulldown,
									//except PA9,10,11,12. PA3 pull-Up.
  #else
	GPIOA->PUPDR = 0xA81AAAAA;		//primarily all inputs pulldown,
									//except PA10,11,12
					//bugfix 25Jan25: PA10 pull-up for gfAuto check
  #endif
#else
	GPIOA->PUPDR = 0xA802AAAA;		//primarily all inputs pulldown,
									//except PA9,10,11,12.
#endif
	GPIOB->PUPDR = 0xAAAAAAAA;		//primarily all inputs pulldown

	SetSysClock();

//pin config repeated specifically:
#ifdef CASE48
//BlackPill compatible special LQFP48 version
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPD15);	//pull-down off, wasts current
	GPIOB->MODER &= ~(GPIO_MODER_MODE15);
	GPIOB->MODER |= GPIO_MODER_MODE15_0 ;  	//Set PortB15 (LED) as output
	GPIOB->BSRR =  GPIO_BSRR_BS_15;        	//PB15 (LED) ON
//no USB voltage detect
#else
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPD8);	//pull-down off, wasts current
	GPIOA->MODER &= ~(GPIO_MODER_MODE8);
	GPIOA->MODER |= GPIO_MODER_MODE8_0 ;  	//Set PortA8 (LED) as output
	GPIOA->BSRR =  GPIO_BSRR_BS_8;        	//PA8 (LED) ON
  #ifndef MINI
	GPIOA->PUPDR &= ~(GPIO_PUPDR_PUPD9);	//PA9=USB detect pull-down off
  #endif
#endif

	gSYSCOUNT = 20;
	gLEDCOUNT = 0;
	gDELAYCOUNT = 0;
	gTIMEOUT = 0;
	gVid = DEFAULT_VID;
	gPid = DEFAULT_PID;
	guDeviceClass = CDC_CLASS;
	gUsbChecked = 0;
	gProject = 0;
	gfRandom = 314159265;		//seed for xorshift32 algorithm
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
	gLedBusy = 2;
#ifndef MINI
	gSplit = 0;					//USB and USART in parallel as terminal
	gBaud = DEFAULT_BAUD;
	gTempBaud = DEFAULT_BAUD;
	gBaudSet = 0;
	gChannel = 0;
	gMIDI_HEAD = 0;
    gMIDI_TAIL = 0;
#endif

#ifdef DMX
	gdDmxMax = 512;
#endif

		
	SysTick_Config(SystemCoreClock/1000);

	msDelay(50);

//start Independent Watchdog  **************
	IWDG->KR = 0x00005555;
	IWDG->RLR = 0x0000FFFF;
	msDelay(2);
	IWDG->PR = 0x00000002;		//resulting timeout ca. 1-2 seconds
	msDelay(2);
	IWDG->KR = 0x0000CCCC;		//start Independent Watchdog

//init I/O state  *********************
	popFlashPage23();			//load only project independent parameters
	msDelay(2);
//init communication  *********************
#ifdef MINI  //standard version
	initMINI();
#else
	initUsart1();
	msDelay(5);
//separate init routines, else code would look confusing
  #ifdef STANDARD
	initSTANDARD();
  #endif
  #ifdef STANDARD_DMX
	initSTANDARD_DMX();
  #endif
  #ifdef LQFP48
	initLQFP48();
  #endif
  #ifdef LQFP48_DMX
	initLQFP48_DMX();
  #endif
#endif
//re-organized 12Jan24:
	popFlashProject();				//load complete project now
#ifdef DMX
	if (gdBootScene)  {
		op_FLUSH();
		gfSP++;
		*gfSP = gdBootScene;
		op_DMXLOAD();
	}
#endif

	while(1) {
		servSystem();
		quit();							//checks for errors
		query();						//collects new terminal command line
										//else handles background processes
		if (gfTibMax == 0) continue;	//formally, in practice impossible
										//if terminal input collected:
		eval();							//performs command line
	}

//emgergeny system reset (should be impossible)
	while(1);       //force cold reset by watchdog
}	

//##########################################################################

void sendModel(void)  //  ****************************************************
//  **************************************************************************
{
#ifdef STANDARD
const uint8_t typeStr [] = "CForth for STM32G4xx LQFP32 v";
#endif
#ifdef STANDARD_DMX
const uint8_t typeStr [] = "CForth-DMX512 for STM32G4xx LQFP32 v";
#endif
#ifdef MINI
const uint8_t typeStr [] = "CForth-DMX512 Mini for STM32G4xx LQFP32 v";
#endif
#ifdef LQFP48
const uint8_t typeStr [] = "CForth for STM32G4xx LQFP48 v";
#endif
#ifdef LQFP48_DMX
const uint8_t typeStr [] = "CForth-DMX512 for STM32G4xx LQFP48 v";
#endif
const uint8_t copyStr [] = "  (c)2020-25 W.Schemmert";
#ifndef MINI
	sendTB(CRR);
#endif
	sendString(typeStr);
	sendTB(HARDWARE_VERSION + 0x30);
	sendTB('.');
	sendNS32(REVNUM);
	sendString(copyStr);
	sendTB(CRR);
}

// Misc Low Level Routines
//************************

#ifdef STANDARD
void initSTANDARD(void)  // **************************************************
//  **************************************************************************
{
	sendTB(CRR);
	sendModel();
	guConfigured = 0;
	if ((GPIOA->IDR&GPIO_IDR_ID9) && (gVid != 0) && (gPid != 0)) startUSB();
	else  {
		sendString(gNoUsbStr);
		sendTB(CRR);
	}
}
#endif

#ifdef STANDARD_DMX
void initSTANDARD_DMX(void)  // **********************************************
//  **************************************************************************
{
// init DMX transmitter
	startDmxTx();
	DmxHeader();	
	msDelay(2);

	sendTB(CRR);
	sendModel();	

	guConfigured = 0;
	if ((GPIOA->IDR&GPIO_IDR_ID9)&&(gVid != 0)&&(gPid != 0)) startUSB();
	else {
		sendString(gNoUsbStr);
		sendTB(CRR);
	}
}
#endif

#ifdef MINI
void initMINI(void)  // ******************************************************
//  **************************************************************************
{
// init DMX transmitter
	startDmxTx();
	DmxHeader();
	msDelay(2);

	guConfigured = 0;
	startUSB();
}
#endif

#ifdef LQFP48
void initLQFP48(void)  // ***************************************************
//  *************************************************************************
{
	sendTB(CRR);
	sendModel();

	guConfigured = 0;
//no USB voltage check
	if ((gVid != 0)&&(gPid != 0)) startUSB();
	else  {	
		sendString(gNoUsbStr);
		sendTB(CRR);
	}
}
#endif

#ifdef LQFP48_DMX
void initLQFP48_DMX(void)  // ************************************************
//does not support miniDMX due to lack of flash.
//  **************************************************************************
{
// init DMX transmitter
	startDmxTx();
	DmxHeader();	
	msDelay(2);

	sendTB(CRR);
	sendModel();	

	guConfigured = 0;
//no USB voltage check
	if ((gVid != 0)&&(gPid != 0)) startUSB();
	else {
		sendString(gNoUsbStr);
		sendTB(CRR);
	}
}
#endif

void SetSysClock(void)  // ***************************************************
//assumes "Vcore range 1", HSE clock 8.000 MHz, SystemClock=AHB=48MHz
//APB1+2 = 48MHz
//adopted from "system_stm32f0xx.c" and modified according to STM32G431 FRefMan
//Reset SW[1:0], HPRE[3:0], PPRE1[2:0], PPRE2[2:0], MCOSEL[2:0] and MCOPRE[2:0]
// **************************************************************************
{
uint32_t StartUpCounter = 0, OscStatus = 0;

#define STARTUP_TIMEOUT   0x50000 //Time out for HSE start up

    RCC->APB1ENR1 |= RCC_APB1ENR1_PWREN;  //Power Interface clock enable
//Vcore Range1 is default after Reset

// SYSCLK, HCLK, PCLK configuration ----------------------------------------
	RCC->CR &= ~((uint32_t)RCC_CR_HSION);			//precaution:switch OFF
// Enable HSE
    RCC->CR |= RCC_CR_HSEON;

// Wait till HSE is ready or Timeout is reached
    do  {
      OscStatus = RCC->CR & RCC_CR_HSERDY;
      StartUpCounter++;
	} while((OscStatus == 0) && (StartUpCounter  < STARTUP_TIMEOUT));

#ifdef CPU_CLOCK144
     if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
	FLASH->ACR =
     FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_7WS;

//System Clock shall be 144MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //8 or 6 MHz PLL input for project, no pre-division
#define PLL_N      36       //8*36=288MHz for project

//#define PLL_N      48       //6*48=288MHz for 6 MHz crystal, special option

// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      2        //divides PLL/6 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 144MHz SYSCLK

 // HCLK = SYSCLK/
    RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 144MHz
 // PCLK2 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE2_DIV2;   //set for project to 72MHz
 //PCLK1 = HCLK / 2
    RCC->CFGR |= RCC_CFGR_PPRE1_DIV2;   //set for project to 72MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
     RCC->PLLCFGR = PLL_N << 8;         //PLL_N = 36=0x12 = 001 0010 (bits 8-14)
     RCC->PLLCFGR |= PLL_Q << 21;        //PLL_Q = 2 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;        //PLL_R = 00  (bits 25,26)

    RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
    RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1;            //select PLL_Q as USB clock
    RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN | RCC_PLLCFGR_PLLREN);

#else		//Clock 48MHz

    if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
	FLASH->ACR =
        FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |FLASH_ACR_ICEN |FLASH_ACR_LATENCY_2WS;

//System Clock shall be 48MHz
//  PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N
#define PLL_M      0        //8 or 6 MHz PLL input for project, no pre-division
#define PLL_N      12       //8*12=96MHz for project

//#define PLL_N      16       //6*16=96MHz for 6 MHz crystal, special option

// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_Q      0        //divides PLL/2 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_R      0        //divides PLL/2 for 48MHz SYSCLK

 // HCLK = SYSCLK/
	RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 48MHz
 // PCLK2 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;   //set for project to 48MHz
 //PCLK1 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;   //set for project to 48MHz

//Configure the main PLL
//PLL input is 8MHz, so PLL_M is set correctly with next command:
//implies correct setting of PLL_R and PLL_Q, too
	RCC->PLLCFGR = PLL_N << 8;	//PLL_N = 12=0x0C = 000 1100 (bits 8-14)
//    RCC->PLLCFGR |= PLL_Q << 21;	//PLL_Q = 00 (bits 21,22)
//    RCC->PLLCFGR |= PLL_R << 25;	//PLL_R = 00  (bits 25,26)

	RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
	RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1; //select PLL_Q as USB clock
	RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN |  RCC_PLLCFGR_PLLREN);
	
#endif	//CPU_CLOCK144 or 48

//Select HSE as PLL source
	RCC->PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSE; //set bits 0,1

// Enable the main PLL
	RCC->CR |= RCC_CR_PLLON;
// Wait till the main PLL is ready
	while((RCC->CR & RCC_CR_PLLRDY) == 0) ;

// Select PLL as system clock source
     RCC->CFGR |= RCC_CFGR_SW_PLL;

// Wait till PLL is used as system clock source
	while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL) ;

	SystemCoreClockUpdate(); //update in correspondence with parameters set above

	RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN; //necessary for Flash read/write action
    }
	
#ifndef  CPU_CLOCK144
    else  {
//init HSI, synchronize with USB, if possible
		RCC->CR &= ~RCC_CR_HSEON;
		RCC->CR |= RCC_CR_HSION;

// Wait till HSI is ready or Timeout is reached
		do  {
		OscStatus = RCC->CR & RCC_CR_HSIRDY;
		StartUpCounter++;
		} while((OscStatus == 0) && (StartUpCounter  < STARTUP_TIMEOUT));

		if (OscStatus) {
// Enable Prefetch Buffer and set Flash Latency
			FLASH->ACR = FLASH_ACR_PRFTEN| FLASH_ACR_DCEN |
										FLASH_ACR_ICEN |FLASH_ACR_LATENCY_2WS;

//System Clock shall be 96MHz
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_MI      1        //8 MHz PLL input for project
#define PLL_NI      12       //8*12=96MHz for project
// USB OTG FS, SDIO and RNG Clock =  PLL_VCO / PLLQ
#define PLL_QI      0        //divides PLL/2 for 48MHz USB clock
// SYSCLK = PLL_VCO / PLL_P
#define PLL_RI      0        //divides PLL/2 for 48MHz SYSCLK

 // HCLK = SYSCLK/
	RCC->CFGR |= RCC_CFGR_HPRE_DIV1;    //set for project to 48MHz
 // PCLK2 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE2_DIV1;   //set for project to 48MHz
 //PCLK1 = HCLK / 2
	RCC->CFGR |= RCC_CFGR_PPRE1_DIV1;   //set for project to 48MHz

//Configure the main PLL
			RCC->PLLCFGR = PLL_MI << 4;
			RCC->PLLCFGR = PLL_NI << 8;		//PLL_N = 36=0x12 (bits 8-14)
			RCC->PLLCFGR |= PLL_QI << 21;	//PLL_Q = 2 (bits 21,22)
			RCC->PLLCFGR |= PLL_RI << 25;	//PLL_R = 00  (bits 25,26)

			RCC->CCIPR &= ~RCC_CCIPR_CLK48SEL;
			RCC->CCIPR |= RCC_CCIPR_CLK48SEL_1; //select PLL_Q as USB clock
			RCC->PLLCFGR |= (RCC_PLLCFGR_PLLQEN |  RCC_PLLCFGR_PLLREN);

//Select HSI as PLL source
			RCC->PLLCFGR |= RCC_PLLCFGR_PLLSRC_HSI; //set bit 1 only
// Enable the main PLL
			RCC->CR |= RCC_CR_PLLON;
// Wait till the main PLL is ready
			while((RCC->CR & RCC_CR_PLLRDY) == 0) ;
// Select PLL as system clock source
			RCC->CFGR |= RCC_CFGR_SW_PLL;

// Wait till PLL is used as system clock source
			while ((RCC->CFGR & RCC_CFGR_SWS) != RCC_CFGR_SWS_PLL) ;

			SystemCoreClockUpdate();
							//update in correspondence with parameters set above

			RCC->AHB1ENR |= RCC_AHB1ENR_FLASHEN;
							//necessary for Flash read/write action

		}
    }
#endif
}


void LED_Off(void)  // ******************************************************
// built as subroutine to supply service to "usb-fs-431.c" hardware independent
// **************************************************************************
{
	if (gLedBusy > 1)  {
#ifdef CASE48
		GPIOB->BSRR = GPIO_BSRR_BR_15;	// clear bit PB15,LED OFF
#else
		GPIOA->BSRR = GPIO_BSRR_BR_8;	// clear bit PA8,LED OFF
#endif
		gLEDCOUNT = 4;  //= ca 300 milliseconds
	}
}


void msDelay(uint32_t count)   // *******************************************
//blocking delay by 'count' milliseconds
// **************************************************************************
{
	gDELAYCOUNT = count;
	while(gDELAYCOUNT)  servSystem();
}


void wait5us(void)   // *****************************************************
// blocking delay ca 5us
// *************************************************************************
{
uint32_t countdown;

#ifdef CPU_CLOCK144
	countdown = 87;
#else
	countdown = 51;
#endif

    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}


void servSystem(void)   // ************************************************
//system relevant, but not extremely time critical services
//are handled quasi in the background
//especially called in pullUsbRxBuf
//but watchdog resets when it hangs somewhere else
//additionally slowed down by SysTick to 100, 20 and 1 times/second
// ****************************************************************************
{
static uint32_t secCount, checkCount;
const uint8_t plugStr [] = "USB plugged ... " ;
const uint8_t unplugStr [] = "USB unplugged" ;

#ifdef DMX
	if (gfDmxBusy)  {
//max. high speed part for DMX:
  #ifdef MINI
		if (USART1->ISR & USART_ISR_TXE) {
			USART1->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
			gdDmxTxIndex++;
			if (gdDmxTxIndex >= gdDmxMax)  DmxHeader() ;
		}
  #else
//  #ifdef DMX
		if (USART2->ISR & USART_ISR_TXE) {
			USART2->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
			gdDmxTxIndex++;
			if (gdDmxTxIndex >= gdDmxMax)  DmxHeader() ;
		}
//  #endif
  #endif
	}

//lower speed services:  **********************************

	if (gSYSCOUNT)  return;		//countdown in SysTick_Handler()

//else if (gSYSCOUNT==0) restart 100Hz service interval
	gSYSCOUNT = 10;

	IWDG->KR = 0x0000AAAA;  //fed watchdog - timeout ca. 1-2 seconds
	
//#ifdef DMX
	if (gfDmxBusy)  {
//handle DMX faders
		if (gdFaderBusy)  fadeProcess();
	}
#endif

	checkCount++;
	if (checkCount == 5)  {					//from here 20Hz interval
		checkCount = 0;

//LED control
		if (gLEDCOUNT){//04Jan25 LED blink behaviour changed,better usr access
			gLEDCOUNT--;

			if (gLEDCOUNT==0)  {
#ifdef CASE48
				if (gLedBusy >= 2)  GPIOB->BSRR = GPIO_BSRR_BS_15;	
													//(LED ON again)
#else
				if (gLedBusy <= 2)  GPIOA->BSRR = GPIO_BSRR_BS_8;	
													//(LED ON again)
#endif
			}
		}
// complex structure of different versions --------------------
#ifdef CASE48
//USB mode control
		secCount++;
		if (secCount == 20)  {          //restart 1 second service interval
			secCount = 0;
			if (guConfigured == 0) {
				if (gUsbChecked == 0) {
					sendString(plugStr);
					sendTB(CRR);
					startUSB();         //USB plugged, enforce reset by Watchdog
				}
			}
			else  {
				if ((gSplit == 4) && (guDeviceClass == CDC_CLASS))  {
					stopUSB();
					guDeviceClass = MIDI_CLASS;
					msDelay(500);
					startUSB();
				}
				if ((gSplit < 4) && (guDeviceClass == MIDI_CLASS))  {
					stopUSB();
					guDeviceClass = CDC_CLASS;	
					msDelay(500);
					startUSB();
				}
			}
		}
#else
  #ifndef MINI
//USB mode control
		secCount++;
		if (secCount == 20)  {          //restart 1 second service interval
			secCount = 0;
			if (guConfigured == 0) {
				if ((GPIOA->IDR & GPIO_IDR_ID9) && (gUsbChecked == 0)) {
					sendString(plugStr);
					sendTB(CRR);
					startUSB();         //USB plugged, enforce reset by Watchdog
			}
		}
			else  {
				if ((GPIOA->IDR & GPIO_IDR_ID9) == 0)  {
					sendString(unplugStr);
					sendTB(CRR);
					stopUSB();          //USB un-plugged, deactivate peripheral
					gUsbChecked = 0;
				}
				else   {                //USB plugged
					if ((gSplit == 4) && (guDeviceClass == CDC_CLASS))  {
						stopUSB();
						guDeviceClass = MIDI_CLASS;	//added 23Oct21
						msDelay(500);
						startUSB();
					}
					if ((gSplit < 4) && (guDeviceClass == MIDI_CLASS))  {
						stopUSB();
						guDeviceClass = CDC_CLASS;	//added 23Oct21
						msDelay(500);
						startUSB();
					}
				}
			}
		}
  #endif		//not MINI
#endif			//CASE48 or not
	}
}


void startUSB(void)  // ******************************************************
// **************************************************************************
{
uint32_t oldTimeout;

#ifdef MINI
	guDeviceClass = CDC_CLASS;
	initUSB();
	gTIMEOUT = 5000;
	while (gTIMEOUT)  {
		servSystem();   //serve DMX process, LED, watchdog
		if (guConfigured)  break;
	}
#else
const uint8_t cdcClassStr [] = "USB as virtual COM=" ;
const uint8_t midiClassStr [] = "USB as MIDI interface=" ;

	gUsbChecked = 1;
	if (gSplit == 4) guDeviceClass = MIDI_CLASS; //double safe init
	else guDeviceClass = CDC_CLASS;
	initUSB();
	gTIMEOUT = 3000;
	oldTimeout = 3000;
	while (gTIMEOUT)  {
		IWDG->KR = 0x0000AAAA;		//feed watchdog, timeout ca.0.5sec
		if (gTIMEOUT != oldTimeout)  {	//check once per ms
			oldTimeout = gTIMEOUT;
			if ((gTIMEOUT & 0x00000003FF) == 0x3FF)  sendTB('.');
		}
		if (guConfigured)  break;
	}

	if (gSplit == 4)  sendString(midiClassStr);
	else  sendString(cdcClassStr);
	if (guConfigured)  {
		sendTB('o');
		sendTB('k');
	}
	else {
		sendTB('E');
		sendTB('r');
		sendTB('r');
		sendTB('o');
		sendTB('r');
	}
	sendTB(CRR);
#endif  //MINI or not
    msDelay(200);
}


void popFlashPage23(void)   // **********************************************
//read project independent system parameters
// **************************************************************************
{
uint32_t * flashPtr;
uint32_t test, vid, pid;
uint32_t *sramPtr;

	emptyForth();		//init default system parameters (=Project #0)

	flashPtr = (uint32_t *)PAGE23;

	test = ~(*flashPtr);//virgin test cannot be==0 if project was saved in Flash
	flashPtr++;
	vid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;
	pid =  ~(*flashPtr);    //cannot be==0 if any project was saved in Flash
	flashPtr++;

	if ((test | vid | pid)==0) {//Flash is virgin, load default empty project
#ifndef MINI
		gBaud = DEFAULT_BAUD;	//as precaution, should be already set
		gTempBaud = DEFAULT_BAUD;
#endif
		return;					//defaults already init
	}
	gVid = vid;
	gPid = pid;

#ifdef MINI
	gfNumBase = ~(*flashPtr);
	flashPtr++;
	if ((gfNumBase != 10) && (gfNumBase != 16))  gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = gfNumBase;
	flashPtr++;					//dummy for double word adjustment
#else
	test = ~(*flashPtr);
	flashPtr++;							//order gfNumbase,gBaud exchanged 27Dec24
	gfNumBase = test & 0x0000001F;
	if ((gfNumBase != 10) && (gfNumBase != 16))  gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = gfNumBase;
	gBaud = (test >> 16) & 0x0000007F;	//baud rate is project independent
	if (gBaud == 0) gBaud = DEFAULT_BAUD;
	gTempBaud = gBaud;
	gChannel = ~(*flashPtr);	//MIDI channel is project independent
	flashPtr++;
#endif

	gProject =  ~(*flashPtr);	//gProject=0 loads Empty Project
	if (gProject == 0)  emptyForth();
	flashPtr++;
#ifdef DMX
	gdFadeTime = ~*flashPtr;
	flashPtr++;
	gdBootScene = ~*flashPtr;
	flashPtr++;
#endif
}


uint32_t pushFlashPage23(void)   //******************************************
//save project independent system parameters permanently
// **************************************************************************
{
uint32_t *flashPtr, *sramPtr;
uint32_t k, word;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
    else return 0;
 
//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca.0.5 seconds

//erase Page23
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = ((23<<3) | FLASH_CR_EOPIE | FLASH_CR_ERRIE | FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write mirror into Page23:
//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word at readout
//i.e. first word in = first word read
//only write USB Vid/Pid, gBaud and gProject into page23 (+ DMX start param)
    flashPtr = (uint32_t *)PAGE23;

	FLASH->CR |= FLASH_CR_PG;

	*flashPtr = 0;                 //for flash virgin test
	flashPtr++;
	*flashPtr = ~(gVid & 0x0000FFFF);
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;

	*flashPtr = ~(gPid & 0x0000FFFF);
	flashPtr++;

#ifdef MINI
	*flashPtr = ~(gfNumBase);
#else
//order gfNumbase,gBaud exchanged 27Dec24
	*flashPtr = ~((gBaud << 16) | gfNumBase);
#endif
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
#ifdef MINI
	*flashPtr = 0xFFFFFFFF;		//empty for double word burn adjustment
#else
	*flashPtr = ~gChannel;
#endif
	flashPtr++;
	*flashPtr = ~gProject;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
#ifdef DMX
	*flashPtr = ~gdFadeTime;
	flashPtr++;
	*flashPtr = ~gdBootScene;
	flashPtr++;
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
#endif

//terminate
    FLASH->CR &= ~FLASH_CR_PG;
//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(20);
    return  1;
}


void popFlashProject(void)   // *********************************************
//first read some project specific globals
//next read 2 pages to gfUserHeader[], then read 6 pages to gfTCode
// **************************************************************************
{
uint32_t k, m, n;
uint32_t word, test;
uint32_t * flashPtr;

//emptyForth() already updated in popFlashPage23() (default system parameters)

	if (gProject)   {
		flashPtr = (uint32_t *)(PAGE24 + 0x4000 * (gProject - 1));
//first 32 bytes = 8 words with system & peripheral config parameters
		test =  ~(*flashPtr);	//cannot be==0 if any project was saved in Flash
		flashPtr++;
		if (test == 0) {		//project Flash is virgin
			emptyForth();		//implies gProject=0 -> empty project
		}
		else  {
		#ifndef MINI
			gMIDI_HEAD = 0;		//flush MIDI buffer
			gMIDI_TAIL = 0;
		#endif
			gfNumBase = ~(*flashPtr);
			flashPtr++;
			if ((gfNumBase != 10)&&(gfNumBase != 16)) gfNumBase = DEFAULT_NUMBASE;
			gfTempNumBase = gfNumBase;

			gfAuto = ~(*flashPtr);
			flashPtr ++;
			
			k = ~(*flashPtr);
			flashPtr ++;
			gfSpiMode = (k >> 24) & 0x00000003;
			gfSpiBits = (k >> 16) & 0x00000003F;
			gfSpiClock = k & 0x00000FFFF;
			
			gfFreq12 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM1 = (int32_t)(k >> 16);
			if ((gfPWM1 == 0) || (gfPWM1 == 0x0000FFFF)) gfPWM1 = -1;
			gfPWM2 = (int32_t)(k & 0x0000FFFF); //Bugfix 16Dec21,before gfPWM1
			if ((gfPWM2 == 0) || (gfPWM2 == 0x0000FFFF)) gfPWM2 = -1;
			flashPtr ++;
			if (gfFreq12)  {
				subFREQ12();
				subPWM1();
				subPWM2();
			}
			gfFreq34 = ~(*flashPtr);
			flashPtr ++;
			k = ~(*flashPtr);
			gfPWM3 = (int32_t)(k >> 16);
			if ((gfPWM3 == 0) || (gfPWM3 == 0x0000FFFF)) gfPWM3 = -1;
			gfPWM4 = (int32_t)(k & 0x0000FFFF);
			if ((gfPWM4 == 0) || (gfPWM4 == 0x0000FFFF)) gfPWM4 = -1;
			flashPtr ++;
			if (gfFreq34)  {
				subFREQ34();
				subPWM3();
				subPWM4();
			}
			
			for (k = 0;k < MAXUSER/2;k++)  {  
										//reads 84*24 bytes=2016bytes +32=2048
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
			flashPtr = (uint32_t *)(PAGE25 + 0x4000 * (gProject - 1));
			for (k = (MAXUSER/2);k < MAXUSER;k++) {  //reads 2016 bytes!
				n = 0;
				for (m = 0; m < 4; m++) {
					word = ~(*flashPtr);
					flashPtr++;
					gfUserHeader[k].name[n]= word & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 8) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 16) & 0x000000FF;
					n++;
					gfUserHeader[k].name[n]= (word >> 24) & 0x000000FF;
					n++;
				}
				gfUserHeader[k].toCompile= ~(*flashPtr);
				flashPtr ++;
				gfUserHeader[k].attributes = ~(*flashPtr);
				flashPtr ++;
			}
#ifdef DMX
			gdFadeTime = ~(*flashPtr);
			flashPtr++;
			gdBootScene = ~(*flashPtr);
			flashPtr++;
#endif

		//loading code is possible over page limits.
			flashPtr = (uint32_t *)(PAGE26 + 0x4000 * (gProject - 1));
			for (k = 0;k < MAXCODE;k++)  {	 //reads 6 pages each 2048 bytes!
				gfTCode[k] = ~(*flashPtr);
				flashPtr ++;
			}
		}
	}

//from here in any case if(gProject) or virgin or not:
	wait5us();
	op_ABORT();

//emergency brake to avoid AUTO deadlock
#ifdef MICRO
		if ((GPIOA->IDR & 8) == 0) gfAuto = 0;	//if PA3(special feature) is LOW
		GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD3;
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD3_1;		//pull-down = default state
#else
		if ((GPIOA->IDR & 0x400) == 0) gfAuto = 0;	//if PA10(RX input) is low
  #ifdef MINI
		GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD10;			//bugfix 25Jan25
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD10_1;		//pull-down = default state
  #endif
#endif

#ifndef MINI						//has USB terminal only
	op_QUEST();
#endif

	if (gfAuto) {
        gfParsed = gfAuto - 1;	//-1 to distinguish from empty AUTOEXE == 0
		interpretUserToken();
	}
}


uint32_t pushFlashProject(void)   //*****************************************
// **************************************************************************
{
uint32_t k, m, n, offset;
uint32_t * flashPtr;
uint32_t word[12];

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    if (gProject == 0) return 1;   //=default empty project, don't save

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 0.5 seconds

    offset = (gProject - 1) * 8;     //project no.0 is empty, not treated here
                                    //project no.1 starts with page#24
//erase Page24 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((24 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//this microcontroller allows exclusively doubleword-wide write in flash
//flash least word is first word in readout
//i.e. first word in = first word read
//first write virgin test, relocalisation, gfNumBase and gfAuto into Page32

    flashPtr = (uint32_t *)(PAGE24 + 0x800 * offset); //0x4000*(project-1));
    FLASH->CR |= FLASH_CR_PG;
//first 32 bytes with system & peripheral config parameters
    *flashPtr = 0x00000000; //virgin flash test
    flashPtr++;
    *flashPtr = ~gfNumBase;
    flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;

    *flashPtr = ~gfAuto;
    flashPtr++;
    *flashPtr = (~((gfSpiMode << 24) | (gfSpiBits << 16 ) |gfSpiClock));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq12;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM1 << 16) & 0xFFFF0000) | 
											((uint32_t)gfPWM2 & 0x0000FFFF));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    *flashPtr = ~gfFreq34;
	flashPtr++;
    *flashPtr = ~((((uint32_t)gfPWM3 << 16) & 0xFFFF0000) | 
											((uint32_t)gfPWM4 & 0x0000FFFF));
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

//write first half of gfUserHeader[] into Page24 + offset
    for (k = 0;k < MAXUSER/2; k+=2) {	//MAXUSER/2 = 84 ->2016bytes+32=2048
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;

        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;

        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
//burn
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca.0.5 seconds
    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page25 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((25 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second half of gfUserHeader[] into Page25 + offset
    flashPtr = (uint32_t *)(PAGE25 + 0x800 * offset); //0x4000 * (project - 1));

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = (MAXUSER/2);k < MAXUSER; k+=2) {     //MAXUSER/2 = 84 ->2016bytes
        n = 0;
        for (m = 0; m < 4; m++) {
            word[m] = gfUserHeader[k].name[n];
            n++;
            word[m] |= gfUserHeader[k].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k].name[n] << 24;
            n++;
        }
        word[4] = gfUserHeader[k].toCompile;
        word[5] = gfUserHeader[k].attributes;
        n = 0;
        for (m = 6; m < 10; m++) {
            word[m] = gfUserHeader[k+1].name[n];
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 8;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 16;
            n++;
            word[m] |= gfUserHeader[k+1].name[n] << 24;
            n++;
        }
        word[10] = gfUserHeader[k+1].toCompile;
        word[11] = gfUserHeader[k+1].attributes;
        for (m = 0; m < 12; m+=2)  {
// this microcontroller allows exclusively doubleword-wide write in flash
//first word
            *flashPtr = ~word[m]; //~for better compatibility with empty SRAM
            flashPtr++;
//second word
            *flashPtr = ~word[m+1]; //~for better compatibility with empty SRAM
            flashPtr++;
//burn
            while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
            FLASH->SR |= FLASH_SR_EOP;
            if (FLASH->SR & FLASH_SR_OPERR)  return 0;
        }
    }
#ifdef DMX
    *flashPtr = ~gdFadeTime;
    flashPtr++;
    *flashPtr = ~gdBootScene;
	flashPtr++;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
#endif

//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca.0.5 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }

//erase Page26 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((26 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write first part of gTCode into Page26 + offset
    flashPtr = (uint32_t *)(PAGE26 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0; k < 0x200; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 0.5 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page27 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((27 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write second part of gTCode into Page27 + offset

    flashPtr = (uint32_t *)(PAGE27 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x200; k < 0x400; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k]; //~for better compatibility with empty SRAM
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];//~for better compatibility with empty SRAM
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 0.5 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page28 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((28 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;
	
    FLASH->CR &= ~FLASH_CR_PER;

//write third part of gTCode into Page28
    flashPtr = (uint32_t *)(PAGE28 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x400; k < 0x600; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca.0.5 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page29 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((29 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write fourth part of gTCode into Page29 + offset
    flashPtr = (uint32_t *)(PAGE29 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x600; k < 0x800; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr= ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR) return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca.0.5 seconds

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page30 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((30 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
     FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR) return 0;
    FLASH->CR &= ~FLASH_CR_PER;

//write fifth part of gTCode into Page30 + offset
    flashPtr = (uint32_t *)(PAGE30 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0x800; k < 0xA00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//feed watchdog
    IWDG->KR = 0x0000AAAA;  //timeout ca. 0.5 seconds

    sendTB('.');

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
	FLASH->KEYR = KEY1;
	FLASH->KEYR = KEY2;
    }
//erase Page31 + offset
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((31 + offset)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;

//write sixth part of gTCode into Page31 + offset
    flashPtr = (uint32_t *)(PAGE31 + 0x800 * offset);

    FLASH->CR |= FLASH_CR_PG;
// this microcontroller allows exclusively doubleword-wide write in flash
    for (k = 0xA00; k < 0xC00; k += 2)  {
//first word
	*flashPtr = ~gfTCode[k];
	flashPtr++;
//second word
	*flashPtr = ~gfTCode[k+1];
	flashPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(20);
    return 1;
}

#ifdef DMX
uint32_t pushDmxFlash(uint32_t page)   //************************************
// **************************************************************************
{
uint32_t k;
uint32_t * flashPtr;
uint32_t * dmxPtr;

//unlock flash
#define KEY1    0x45670123      //mandatory unlock key defined by STM
#define KEY2    0xCDEF89AB

    while ((FLASH->SR &FLASH_SR_BSY) !=0); //wait until no operation is going on
    if ((FLASH->CR & FLASH_CR_LOCK) != 0) {
		FLASH->KEYR = KEY1;
		FLASH->KEYR = KEY2;
    }

//feed watchdog
    IWDG->KR = 0x0000AAAA;          //timeout ca. 0.5 seconds

    sendTB('.');

//erase Page40 + page
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;  //wait until flash is not busy
    FLASH->CR = (((40 + page)<<3)|FLASH_CR_EOPIE|FLASH_CR_ERRIE|FLASH_CR_PER);
                                //EOPIE is set formally to get SR_EOP set
    FLASH->CR |= FLASH_CR_STRT;
    while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
    FLASH->SR |= FLASH_SR_EOP;
    if (FLASH->SR & FLASH_SR_OPERR)  return 0;

    FLASH->CR &= ~FLASH_CR_PER;
//this microcontroller allows exclusively doubleword-wide write in flash
	FLASH->CR |= FLASH_CR_PG;
	flashPtr = (uint32_t *)(PAGE40 + 0x800 * page);
	dmxPtr = (uint32_t *)gfTempBuf;
    for (k = 0; k < 512; k += 2)  {
	*flashPtr = ~(*dmxPtr);
	flashPtr++;
	dmxPtr++;	
	*flashPtr = ~(*dmxPtr);
	flashPtr++;
	dmxPtr++;
//burn
	while ((FLASH->SR & FLASH_SR_BSY) != 0) ;
	FLASH->SR |= FLASH_SR_EOP;
	if (FLASH->SR & FLASH_SR_OPERR)  return 0;
    }
//terminate
    FLASH->CR &= ~FLASH_CR_PG;

    sendTB('.');


//lock flash
    FLASH->CR |= FLASH_CR_LOCK;
    msDelay(10);
    return 1;	
}	
#endif //DMX

//###########################################################################

//USART Support
//*************

#ifdef MINI
// RS_232 not defined  *********************
void sendTB(uint8_t data) // ************************************************
//send one byte via serial terminal port (USB and/or USART)
//0=USB and USART both terminal, 1+3= USB is terminal, 2+4=USART is terminal
// **************************************************************************
{
	sendUsbByte(data);
}


void sendNibble(uint8_t data) // ********************************************
//0=USB and USART both terminal, 1= USART is terminal, 2=USB is terminal
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
	sendUsbByte(asciiNibble);
}

#else  //no MINI but RS-232

  #ifdef CASE48  //GREYPILL needs external TX inverter
 
void initUsart1(void) // ****************************************************
// USART1-TxD is PortA,9
// USART1-RxD is PortA,10
// **************************************************************************
{
uint16_t baudset;

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ;   // USART1 clock enable
//set PA9/PA10 in alternate function mode
    GPIOA->MODER &= ~(GPIO_MODER_MODE9_0 | GPIO_MODER_MODE10_0);
    GPIOA->MODER |= (GPIO_MODER_MODE9_1 | GPIO_MODER_MODE10_1);

//USART1 is AF7 on PortA
    GPIOA->AFR[1] &= 0xFFFFF00F;        //PA9 is TX, PA10 is RX
    GPIOA->AFR[1] |= 0x00000770;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

      #ifdef CPU_CLOCK144
    switch(gTempBaud) {
    case 96:
        baudset = 0x1D4C;       //9600Baud: 468.8 at 72 MHz APB2 Clock
        break;
    case 19:
        baudset = 0x0EA6;       //19200Baud: 234.4 at 72 MHz APB2 Clock
        break;
    case 31:
        baudset = 0x0900;       //MIDI=31250Baud: 144.0 at 72 MHz APB2 Clock
        break;
    case 38:
        baudset = 0x0753;       //38400Baud: 117.2 at 72 MHz APB2 Clock
        break;
    case 57:
        baudset = 0x04E2;       //57600Baud: 78.1 at 72 MHz APB2 Clock
        break;
    case 11:
    default:
        baudset = 0x0271;       //default (virgin flash)
        break;                  //115200Baud: 39.1 at 36 MHz APB2 Clock
    }
    USART1->BRR = baudset; //default 115200Baud: 19.3 at 36 MHz APB2 Clock

      #else
    switch(gTempBaud) {
    case 96:
        baudset = 0x1380;       //9600Baud: 312.0 at 48 MHz APB2 Clock
        break;
    case 19:
        baudset = 0x09C0;       //19200Baud: 156.0 at 48 MHz APB2 Clock
        break;
    case 31:
        baudset = 0x0600 ;      //MIDI=31250Baud: 96.0 at 48 MHz APB2 Clock
        break;
    case 38:
        baudset = 0x04E0 ;      //38400Baud: 78.0 at 48 MHz APB2 Clock
        break;
    case 57:
        baudset = 0x0340;       //57600Baud: 52.0 at 48 MHz APB2 Clock
        break;
    case 11:
    default:
        baudset = 0x01A0;       //default (virgin flash)
                  				//115200Baud: 26.0 at 48 MHz APB2 Clock
    }
// baudset = 0x01A0;       //default (virgin flash)
    USART1->BRR = baudset; //default 115200Baud: 26.0 at 48 MHz APB2 Clock
      #endif
#ifdef EXTERN_TXINVERT			//hardare with external inverter
	USART1->CR2 = 0;				//default:1 Stopbit
#else
	USART1->CR2 = USART_CR2_TXINV; 
								//1 Stopbit, TX level inverted
#endif
    USART1->CR3 = 0;			//default:no Handshake
    USART1->CR1 = USART_CR1_UE|USART_CR1_RXNEIE|USART_CR1_TE|USART_CR1_RE;
								//8Bit, no Parity, RX Int
	gBaudSet = 1;
    NVIC_EnableIRQ(USART1_IRQn);
}

  #else		//not CASE48

void initUsart1(void) // ****************************************************
// USART1-TxD is PortB,6
// USART1-RxD is PortA,10 (hardware version 2)
// **************************************************************************
{
uint16_t baudset;

    RCC->APB2ENR |=  RCC_APB2ENR_USART1EN ;   // USART1 clock enable
//set PA10 in alternate function mode
    GPIOA->MODER &= ~GPIO_MODER_MODE10_0 ;
    GPIOA->MODER |= GPIO_MODER_MODE10_1;
//set PB6 in alternate function mode
    GPIOB->MODER &= ~GPIO_MODER_MODE6_0 ;
    GPIOB->MODER |= GPIO_MODER_MODE6_1;
    //GPIOx->OTYPER is push-pull by default
    //GPIOx->OSPEEDR is 2 MHz by default
//USART1 is AF7 on PortA
    GPIOA->AFR[1] &= 0xFFFFF0FF;        //PA10 is RX
    GPIOA->AFR[1] |= 0x00000700;
//and AF7 on PortB
    GPIOB->AFR[0] &= 0xF0FFFFFF;        //PB6 is TX
    GPIOB->AFR[0] |= 0x07000000;

    gUSART_RXHEAD = gUSART_RXBUF;
    gUSART_RXTAIL = gUSART_RXBUF;

    #ifdef CPU_CLOCK144
    switch(gTempBaud) {
    case 96:
        baudset = 0x1D4C;       //9600Baud: 468.8 at 72 MHz APB2 Clock
        break;
    case 19:
        baudset = 0x0EA6;       //19200Baud: 234.4 at 72 MHz APB2 Clock
        break;
    case 31:
        baudset = 0x0900;       //MIDI=31250Baud: 144.0 at 72 MHz APB2 Clock
        break;
    case 38:
        baudset = 0x0753;       //38400Baud: 117.2 at 72 MHz APB2 Clock
        break;
    case 57:
        baudset = 0x04E2;       //57600Baud: 78.1 at 72 MHz APB2 Clock
        break;
    case 11:
    default:
        baudset = 0x0271;       //default (virgin flash)
        break;                  //115200Baud: 39.1 at 36 MHz APB2 Clock
    }
    USART1->BRR = baudset; //default 115200Baud: 19.3 at 36 MHz APB2 Clock

    #else
    switch(gTempBaud) {
    case 96:
        baudset = 0x1380;       //9600Baud: 312.0 at 48 MHz APB2 Clock
        break;
    case 19:
        baudset = 0x09C0;       //19200Baud: 156.0 at 48 MHz APB2 Clock
        break;
    case 31:
        baudset = 0x0600 ;      //MIDI=31250Baud: 96.0 at 48 MHz APB2 Clock
        break;
    case 38:
        baudset = 0x04E0 ;      //38400Baud: 78.0 at 48 MHz APB2 Clock
        break;
    case 57:
        baudset = 0x0340;       //57600Baud: 52.0 at 48 MHz APB2 Clock
        break;
    case 11:
    default:
        baudset = 0x01A0;       //default (virgin flash)
                  				//115200Baud: 26.0 at 48 MHz APB2 Clock
    }
// baudset = 0x01A0;			//default (virgin flash)
    USART1->BRR = baudset; //default 115200Baud: 26.0 at 48 MHz APB2 Clock
    #endif
#ifdef EXTERN_TXINVERT			//hardare with external inverter
	USART1->CR2 = 0;			//default:1 Stopbit
#else
	USART1->CR2 = USART_CR2_TXINV; 
								//1 Stopbit, TX level inverted by CPU
#endif
    USART1->CR3 = 0;			//default:no Handshake
    USART1->CR1 = USART_CR1_UE|USART_CR1_RXNEIE|USART_CR1_TE|USART_CR1_RE;
								//8Bit, no Parity, RX Int
	gBaudSet = 1;
    NVIC_EnableIRQ(USART1_IRQn);
}

  #endif	//CASE48 or not

void sendB(uint8_t data) // *************************************************
//send 1 raw byte
// **************************************************************************
{
uint32_t fillstate;

    while(!(USART1->ISR&USART_ISR_TXE));
    USART1->TDR = data;
	if (gLedBusy == 3) LED_Off();		//new 04Jan24
}


void sendTB(uint8_t data) // ************************************************
//send one byte via serial terminal port (USB and/or USART)
//0=USB and USART both terminal, 1+3= USB is terminal, 2+4=USART is terminal
// **************************************************************************
{
    if (gSplit == 0)  {
        sendB(data);
        if (guConfigured)  sendUsbByte(data);
    }
	else  if ((gSplit == 2) || (gSplit == 4))sendB(data);
							//USART is terminal, USB is serial or MIDI I/O
    else if (guConfigured)  sendUsbByte(data);
                            //USB is terminal(virtual COM),USART is serial I/O

}


void sendIOB(uint8_t data) // ************************************************
//send one byte via serial I/O
// **************************************************************************
{
    if ((gSplit == 1) || (gSplit == 3)) { 		//USB is terminal(virtual COM),
        sendB(data); 							//USART is serial I/O
    }
    else if ((gSplit == 2) && (guConfigured)) { //USART=terminal, USB=serial I/O
        sendUsbByte(data);
    }
    else if ((gSplit == 4) && (guConfigured))  { //USART=terminal, USB=MIDI I/O
        enCodeMidi(data);
    }
	else  gfErrors |= ERR_CONTEXT;  //gSplit=0
}


void sendNibble(uint8_t data) // ********************************************
//0=USB and USART both terminal, 1= USART is terminal, 2=USB is terminal
// **************************************************************************
{
uint8_t asciiNibble;

    asciiNibble = data+0x30;
    if (asciiNibble>0x39) asciiNibble+= 7;
    if (gSplit == 0)  {
        sendB(asciiNibble);
        if (guConfigured)  sendUsbByte(asciiNibble);
    }
    else if (((gSplit == 1) || (gSplit == 3)) && (guConfigured))  {
							//USB is terminal(virtual COM),USART is serial I/O
        sendUsbByte(asciiNibble);
    }
    else sendB(asciiNibble);
}
#endif  //MINI or RS_232


void sendString(const uint8_t * str) // *************************************
// **************************************************************************
{
 uint8_t byte;

    while  ((byte = *str++) != 0)  sendTB(byte);
}


void sendX8(uint32_t data) // *********************************************
// send bits 0...7 in hex format (without leading 0x)
// **************************************************************************
{
    uint8_t temp;

    temp= (data&0xF0) >> 4;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    temp= data&0x0F;
    temp = temp+0x30;
    if (temp>0x39) temp+= 7;
    sendTB(temp);
    sendTB(0x20);
}


void sendX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
//but leading SPACE
// **************************************************************************
{
uint8_t nibble;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    sendNibble(nibble);
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
    sendTB(0x20);

}

void sendNX32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
uint8_t nibble, noLead;
    noLead = 0;
    nibble = (uint8_t) ((data>>28) & 0x0000000F);
    if (nibble)  {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>24) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>20) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>16) & 0x0000000F);
     if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>12) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>8) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) ((data>>4) & 0x0000000F);
    if (nibble || noLead) {
        sendNibble(nibble);
        noLead = 1;
    }
    nibble = (uint8_t) (data & 0x0000000F);
    sendNibble(nibble);
}


void send0X32(uint32_t data) // **********************************************
// send 1 word (32bit) in hex format (without leading 0x)
// **************************************************************************
{
//	if (data >= 10)  {			//modified 11Jan24
		sendTB('0');
		sendTB('x');
//	}
	sendNX32(data);
}


void sendS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendNS32(data);
    sendTB(0x20);
}


 void sendLS32(uint32_t data) // ********************************************
//send signed decimal number string. Leading zeroes are suppressed
//leading SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
    sendTB(0x20);
    sendNS32(data);
}


void sendNS32(uint32_t data) // *******************************************
//send signed decimal number string. Leading zeroes are suppressed
//no leading and trailing SPACE
//up to  +/- 2.147.483.647 Billions
// **************************************************************************
{
uint32_t quotient;

    if (data >= 0x80000000) {
        sendTB('-');
        data = ~data + 1 ;
    }
    if (data < 10) goto ONES;
    if (data < 100) goto TENS;
    if (data < 1000) goto HUNDREDS;
    if (data < 10000) goto THOUSANDS;
    if (data < 100000) goto TH10;
    if (data < 1000000) goto TH100;
    if (data < 10000000) goto MILLIONS;
    if (data < 100000000) goto MIL10;
    if (data < 1000000000) goto MIL100;
    quotient = data/1000000000;
    data = data - quotient*1000000000;
    sendTB((uint8_t)quotient+ 0x30);
MIL100:
    quotient = data/100000000;
    data = data - quotient*100000000;
    sendTB((uint8_t)quotient + 0x30);
MIL10:
    quotient = data/10000000;
    data = data - quotient*10000000;
    sendTB((uint8_t)quotient + 0x30);
MILLIONS:
    quotient = data/1000000;
    data = data - quotient*1000000;
    sendTB((uint8_t)quotient + 0x30);
TH100:
    quotient = data/100000;
    data = data - quotient*100000;
    sendTB((uint8_t)quotient+ 0x30);
TH10:
    quotient = data/10000;
    data = data - quotient*10000;
    sendTB((uint8_t)quotient + 0x30);
THOUSANDS:
    quotient = data/1000;
    data = data - quotient*1000;
    sendTB((uint8_t)quotient + 0x30);
HUNDREDS:
    quotient = data/100;
    data = data - quotient*100;
    sendTB((uint8_t)quotient+ 0x30);
TENS:
    quotient = data/10;
    data = data - quotient*10;
    sendTB((uint8_t)quotient + 0x30);
ONES:
    sendTB((uint8_t)data + 0x30) ;
}


#ifdef MINI
uint32_t pullTermRxBuf(void)  // ********************************************
// receive byte from serial terminal port with priority order USB > RS232
// **************************************************************************
{
	return pullUsbRxBuf();
}

#else	//not MINI

uint32_t pullUsartRxBuf(void) // ********************************************
// pull one byte asynchronously from USART reveive buffer
// check first if something received. If yes then set bit 15. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.

	if (gUSART_RXTAIL!=gUSART_RXHEAD)  {
		recv= (uint32_t)((*gUSART_RXTAIL++ & 0x000000FF)|0x00008000);
		if (gUSART_RXTAIL==gUSART_RXBUFTOP)  gUSART_RXTAIL=gUSART_RXBUF;
		LED_Off();
		return recv;
	}
	else return 0;
}


uint32_t pullTermRxBuf(void)  // ********************************************
// receive byte from serial terminal port with priority order USB > RS232
// **************************************************************************
{
uint32_t recv;

    if (gSplit == 0)  {
        if (guConfigured)  {            //ASART and USB are parallel terminal
            recv = pullUsbRxBuf();
            if (recv) return recv;      //USB has priority over USART
        }
        recv = pullUsartRxBuf();
        return recv;
    }
    else if ((gSplit == 2)||(gSplit == 4)) { //gSplit = 2 or 4:USART is terminal
            recv = pullUsartRxBuf();
            return recv;
    }
    else if (guConfigured) {            //gSplit = 1 or 3://USB is terminal
        recv = pullUsbRxBuf();  //if (guConfigured == 0): no terminal available
        return recv;
    }
}


void  MidiScanner(uint32_t byte)  // ****************************************
// accepts only MIDI channel as stored in "gChannel", 
// exception: gChannel = 0xFF, then all MIDI channels are accepted
// **************************************************************************
{
static uint32_t status, data1, msgtype;
static uint32_t progress;

    if (byte >= 0x80) {					//new status byte
			if ((byte == 0xF0) || (byte == EOX) || (byte == 0xF4) ||
						(byte == 0xF5) || (byte == 0xF9) ||(byte == 0xFD) ) {
//								kill SysEx(F0),EOX is end of SysEx message,
//								0xF4, 0xF5, 0xF9 and 0xFD are not defined
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
			if ((byte == 0xF6) || (byte >= 0xF8))  { //1 byte msg
				gMIDIBUF[gMIDI_HEAD] = byte << 16;		//data bytes = 0
				gMIDI_HEAD++;
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}
// "normal" MIDI Channel Msg:
			if (((byte & 0x0000000F) != (gChannel - 1)) && (gChannel != 0)) {
				progress = 0;			//reset parser
				status = 0;				//drop input until next status byte
				msgtype = 0;
				return;
			}	//else accepted status byte:
			status = byte;
			msgtype = byte & 0xF0;
			progress = 1;
	}
	else  {								//new data byte
		if (progress == 0)  {	//SysEx or byte out of context, reset parser
			status = 0;
			msgtype = 0;
			data1 = 0;
			return;
		}
		if (progress == 1)  {
			data1 = byte;
			if ((msgtype == 0xC0)||(msgtype == 0xD0) || (status == 0xF1)||
									(status == 0xF3)) { //2 byte msgs
				gMIDIBUF[gMIDI_HEAD] = (status << 16) | data1;
				gMIDI_HEAD++;					//bits 0..7 for easier parsing
				if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
				return;		//keep progress==1 and gStatus for Running State
			}
			progress = 2;
			return;
		}
		if (progress == 2)  {					//any 3 byte messages
			gMIDIBUF[gMIDI_HEAD] = (status << 16) | (data1 << 8) | byte;
			gMIDI_HEAD++;					//bits 0..7 for easier parsing
			if (gMIDI_HEAD == MIDI_BUFLEN) gMIDI_HEAD = 0;
			progress = 1;   //reset progress and keep status for Running State
		}
	}
}


uint32_t pullMidiBuf(void) // ********************************************
// pull one word asynchronously from MIDIBUF
// check first if something received. If yes then set bit 31. If not return 0
// less overhead and more safe than use of a separate "buffer filled" flag
// **************************************************************************
{
uint32_t recv;

	servSystem();   //serve watchdog etc.

	if (gMIDI_TAIL != gMIDI_HEAD)  {
		recv= gMIDIBUF[gMIDI_TAIL] | 0x80000000;
		gMIDI_TAIL++;
		if (gMIDI_TAIL== MIDI_BUFLEN)  gMIDI_TAIL= 0;
		LED_Off();
		return recv;
	}
	else return 0;
}

#endif  //MINI or not

// ##########################################################################
// Interrupt Handlers

void SysTick_Handler(void)  // ***********************************************
//1kHz repetition rate
// ***************************************************************************
{
	if (gSYSCOUNT)  gSYSCOUNT--;		//controls servSystem()
	if (gDELAYCOUNT) gDELAYCOUNT--;
	if (gTIMEOUT) gTIMEOUT--;			//to be used in any subroutine
	if (gfMSCOUNT) gfMSCOUNT --;		//used by op_MS,op_DS
	if (gfTIXCOUNT) gfTIXCOUNT --;		//used by op_TIX,op_WAIT
	if (gfSTEPCOUNT1) gfSTEPCOUNT1--;	//stepper timing
	if (gfSTEPCOUNT2) gfSTEPCOUNT2--;
	if (gfSTEPCOUNT3) gfSTEPCOUNT3--;
	gfRandom ^= gfRandom << 13;				//"RANDOM" xorshift32 algorithm
	gfRandom ^= gfRandom >> 17;
	gfRandom ^= gfRandom << 5;	
}

#ifndef MINI
void USART1_IRQHandler(void)    // *****************************************
// *************************************************************************
{
uint32_t recv;

	if (USART1->ISR & USART_ISR_RXNE)  {
		recv = USART1->RDR & 0x00FF;
		if ((gSplit != 1) && (gSplit != 3)) {	//	USART is terminal
			if (recv == CTRL_C)  {
				gfTermCtrlBC = CTRL_C;
				return;
			}
			if (recv != CTRL_B)  {
				*gUSART_RXHEAD = (uint8_t)recv;
				gUSART_RXHEAD++;
				if (gUSART_RXHEAD==gUSART_RXBUFTOP)gUSART_RXHEAD=gUSART_RXBUF;
			}
			else  op_BAKOP();
		}
		else {									//no special char handling
			*gUSART_RXHEAD = (uint8_t)recv;
			gUSART_RXHEAD++;
			if (gUSART_RXHEAD==gUSART_RXBUFTOP)  gUSART_RXHEAD=gUSART_RXBUF;
		}
	}
}
#endif	//no MINI but RS-232

// END #######################################################################