// Forth specific routines of of the CForth  project **************************
//
// Author and (c)2020-25 Wolfgang Schemmert. The copyright claim is restricted
// to this file and the depending files "main.c", "main.h", "cforth.h",
// "usb-fs-431.c", "dmx.c" and "globaldefines.h"
// Other depending material is published, copyrighted and licensed
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// Contact: <www.midi-and-more.de>
// A detailled hardware construction and operation manual
// is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32g431xx.h"
#include "globaldefines.h"
#include "cforth.h"


void quit(void)   // **************************************************
// ********************************************************************
{
uint32_t k;
const uint8_t ErrorStr [] = "Error:";
const uint8_t ColStr [] = " (col ";
const uint8_t NotFoundStr [] = "_not found";
const uint8_t SyntaxStr [] = "_Syntax";
const uint8_t StackStr [] = "_Stack";
const uint8_t CompoStr [] = "_Compile only";
const uint8_t CommStr [] = "_No I/OComm Access";
const uint8_t CodeAccessStr [] = "_Invalid Code access";
const uint8_t InvalContextStr [] = "_out of Context";
const uint8_t RunOnlyStr [] = "_Runtime only token";
const uint8_t CancelStr [] = "_Thread cancelled";
const uint8_t RangeStr [] = "_Number out of Range";
const uint8_t Div0Str [] = "_Division by Zero";
const uint8_t CompilingStr [] = "  compiling      ";

    if (gfSP >= gfDataStack + MAXSTACK - 4) gfErrors |= ERR_STACK;
    if (gfSP < gfDataStack + 3) gfErrors |= ERR_STACK;

	if (gfErrors != 0)  {
		if (gfCompiState) {
           for (k=0; k<16; k++) {
               gfUserHeader[gfNextNameCompi].name[k] = 0;
           }
               gfUserHeader[gfNextNameCompi].toCompile = 0;
               gfUserHeader[gfNextNameCompi].attributes = 0;
		}

		gfCompiState = 0;		//terminate compilation in any case

		if (gfTokenStart)  {
            for (k = 0;k < gfTokenStart;k++) sendTB(SPACE);
            sendTB('^');
            sendTB(CRR);
        }
        sendString(ErrorStr);
		if ((gfErrors & ERR_STRINGTOKEN) != 0)  sendString(NotFoundStr);
		if ((gfErrors & ERR_SYNTAX) != 0)  sendString(SyntaxStr);
		if ((gfErrors & ERR_STACK) != 0)  sendString(StackStr);
		if ((gfErrors & ERR_COMPONLY) != 0)  sendString(CompoStr);
		if ((gfErrors & ERR_COMMACC) != 0)  sendString(CommStr);
		if ((gfErrors & ERR_USERCODE) != 0)  sendString(CodeAccessStr);
		if ((gfErrors & ERR_CONTEXT) != 0)  sendString(InvalContextStr);
        if ((gfErrors & ERR_CANCEL) != 0)  sendString(CancelStr);
        if ((gfErrors & ERR_RUNTOKEN) != 0)  sendString(RunOnlyStr);
		if ((gfErrors & ERR_NUMVAL) != 0)  sendString(RangeStr);
		if ((gfErrors & ERR_DIVZERO) != 0)  sendString(Div0Str);
        sendString(ColStr);
		sendLS32(gfTokenStart);
        sendTB(0x29);  //')'
        op_ABORT();					//resets gfErrors, too
	}
	else  {
        if (gfCompiState) {			//compiling
            sendString(CompilingStr);
		}
		else if (gfComment != 1) {	//interpreting and no full line comment
            sendTB(SPACE);
            sendTB('o');
            sendTB('k');
			if (gfTempNumBase == 16)  sendTB('H');
		}
	}
	sendTB(CRR);
	gfBackNumValid = 0;
}


void query(void)   // ******************************************************
// *************************************************************************
{
uint32_t tibstate;
uint32_t recvword;

	gfTibPos = 0;
	gfTibMax = 0;
	gfComment = 0;          //reset and start every line not in comment state

	do  {
		serveBack();
		tibstate = subQuery();
	} while(tibstate == 0);
}


uint32_t subQuery(void)   // ************************************************
//return 0 is OK return. Other return values have special meaning
// **************************************************************************
{
uint32_t recvword;
uint8_t recvbyte;

	recvword = pullTermRxBuf();                 		//USB has priority
	if (recvword == 0)  return 0;
	recvbyte = (uint8_t)(recvword & 0x000000FF);
	if (gfBackProcess && gfExecUser)  {
		gfTermChar = recvbyte;
		gfTermCheck = 1;
	}
	if (recvbyte == LFF) return 0;	//ignore LFF
	if (recvbyte < CTRL_B)  return 0; //wrong TIB format,ignore and go on

//if not executing User Token, convert TAB -> SPACE
	if (gfExecUser == 0)  {
		if (recvbyte == TAB)  recvbyte = SPACE;
		if ((gfTibPos >= 64) && (recvbyte == SPACE))  recvbyte = CRR;
								//force line break if line is too long
	}

//block loading comments:
	if ((recvbyte == 0x28) && (gfComment == 0)) {       //0x28 = '('
		sendTB(0x28);
		while ((recvword = pullTermRxBuf()) == 0);
		recvbyte = (uint8_t)(recvword & 0x000000FF);
		 if (recvbyte == SPACE) {                       // C ..style comment
			gfComment = gfTibPos + 1;
			sendTB(BKSP);
			sendTB(SPACE);
			sendTB(BKSP);
		}
		else  {
			gfTib[gfTibPos] = 0x28;
			gfTibPos++;
			gfTib[gfTibPos] = recvbyte;
			gfTibPos++;
			sendTB(recvbyte);
		}
		return 0;
	}

	if ((recvbyte == 0x2F) && (gfComment == 0)) {	//0x2F = '/'
		sendTB(0x2F);
		while ((recvword = pullTermRxBuf()) == 0);
		recvbyte = (uint8_t)(recvword & 0x000000FF);
		if (recvbyte == 0x2F) {				// style comment
			gfComment = gfTibPos + 1;
			sendTB(BKSP);
			sendTB(SPACE);
			sendTB(BKSP);
		}
		else  {
			gfTib[gfTibPos] = 0x2F;
			gfTibPos++;
			gfTib[gfTibPos] = recvbyte;
			gfTibPos++;
			sendTB(recvbyte);
		}
		return 0;
	}

	if (recvbyte == ESC)  {
		if (gfExecUser)  {					//-->termination of User Thread
			sendTB(CRR);
			gfTib[gfTibPos] = SPACE;
			gfTib[gfTibPos + 1] = 0;
			gfTibMax = gfTibPos;
			return 2;
		}
		else return 0;
	}

	if (recvbyte == TAB)   return 3;
	if  (recvbyte == CTRL_R)  return 4;

	if ((recvbyte == BKSP) && (gfComment == 0)) {
		if (gfTibPos > 0)  {
			gfTibPos--;
			gfTib[gfTibPos]=SPACE;
			sendTB(BKSP);
			sendTB(SPACE);
			sendTB(BKSP);
		}
		return 0;
	}
//avoid line feed after empty comment lines:
	if (recvbyte == CRR)  {
		if (gfTibPos && (gfComment != 1)) sendTB(CRR);
		gfTib[gfTibPos] = SPACE;        //final token terminator for "eval"
		gfTib[gfTibPos + 1] = 0;
		gfTibMax = gfTibPos;
		return 1;
	}
//finally
	if (gfComment == 0)  {
		gfTib[gfTibPos] = (uint8_t)recvbyte;
		sendTB((uint8_t)recvbyte);
		gfTibPos++;
	}
    return 0;
}


void eval(void)   // ******************************************************
// ************************************************************************
{
	gfTokenStart = 0;
	gfTibPos = 0;
	while (gfTibPos < gfTibMax)  {
		scanNext();
		if (gfTokenstr[0] == 0) continue;	//ignore CTRL_B,TAB,LFF,ESC 

//return of interpretNumber():1:number, 0:look for token
		if (interpretNumber()) {		//valid number parsed
			if (gfCompiState == 0)  {	// interpreting
				gfSP ++;
				*gfSP = gfTokenNum;
			}
			else  {
				compileTCode(IDX_DOLIT);
				compileTCode(gfTokenNum);
			}
			gfTokenStart = gfTibPos;
			continue;
		}
//interpretNumber() == 0: could not parse as number, try to interpret as token
		else if (findUserToken()) {	//User Token found, index in gfParsed
			if(interpretUserToken())  return;	//returns 0 if ok !!
			gfTokenStart = gfTibPos;
			continue;
		}
		else  {
			if (findKernelToken() == 0) {	// Kernel Tokenstring not found
				gfErrors |= ERR_STRINGTOKEN;
				gfCodeIndex = gfCodeStartIndex;
				gfTokenStart = gfTibPos;
				return;
			}
			//Kernel Tokenstring found:
			interpretKernelToken(); //Kernel token found, index in gfParsed
			if (gfErrors) return;
			gfTokenStart = gfTibPos;
			continue;
		}
	}
}


void backProcEval(void)   // **********************************************
// ************************************************************************
{
uint32_t k;
const uint8_t invalidStr [] = " <-not a Kernel OP nor a number";
const uint8_t noBackStr [] = " <-invalid for Back-Ctrl";
const uint8_t runErrStr [] = " <-Runtime Error";

    for (k=82;k<gfTibMax;k++) sendTB(gfTib[k]); //for info, send parsed string

	gfTibPos = 82;
	gfTokenStart = gfTibPos;

	while (gfTibPos < gfTibMax)  {
		scanNext();
		if (gfTokenstr[0] == 0) continue;	//ignore CTRL_B,TAB,LFF,ESC 

		if (interpretNumber()) {
						//interpretNumber(): = 1: number, =0:look for token
						//valid number parsed, here only interpreting
            gfBackProcData = gfTokenNum;
            if (gfExecUser) gfBackNumValid = 1;
            gfTokenStart = gfTibPos;
			continue;
		}
		else  {
//interpretNumber() = 0: could not parse as number, try to interpret as token
            if (findKernelToken() == 0) { // Kernel Tokenstring not found
                sendString(invalidStr);
                goto ERROR_BREAK;
            }
            //else Kernel Tokenstring found;
            if (gfKernelAttributes[gfParsed] & A_BAK)  { //only BAK attribute !
                interpretKernelToken();	//Kernel token found, index in gfParsed
                if (gfErrors)  {
					gfErrors = 0;		//don't cancel background process
                    sendString(runErrStr);
					goto ERROR_BREAK;
                }
            }
            else  {
                sendString(noBackStr);
                goto ERROR_BREAK;
            }
            gfTokenStart = gfTibPos;
            continue;
		}
    }
ERROR_BREAK:
    sendTB(CRR);
    gfTibMax = 82;
    gfTibPos = 82;
    gfComment = 0;
}


void scanNext(void)   // **************************************************
// ************************************************************************
{
uint32_t k, tib;

	while (gfTib[gfTokenStart] == SPACE)  gfTokenStart++; 
										//eliminate leading SPACEs

	if (gfTib[gfTokenStart] == 0)  {	//is the case if CTRL_B,TAB,LFF,ESC
		gfTokenstr[0] = 0;
		gfTibPos = gfTokenStart;
		return;
	}

	k=0;
	while ((tib = gfTib[gfTokenStart+k]) != SPACE)  {
								//search trailing SPACE(s) after scanned token
		gfTokenstr[k] = tib;
		k++;
		if (k > 15) break;		//limit length of string
	}
	gfTokenstr[k] = 0;
	gfTibPos = gfTokenStart + k;
}


uint32_t interpretNumber(void)   // *****************************************
// hex numbers can be entered without leading 0X, gfTempNumBase is criterium
// optionally, when entered with leading 0X, Hex base is assumed temporarily
// leading '-'(after 0x if present) is accepted(good for small negative numbers)
// to simplify parameter transfer, succesfully parsed number is in "gfTokenNum"
//return:
// 1 = valid number
// 0 = no valid number, try to interpret as command token
// *************************************************************************
{
uint32_t numLen;

    numLen = 0;
    while (gfTokenstr[numLen] != 0) numLen++;

	if ((gfTokenstr[0] == '-') && (numLen == 1))  return 0;  //"-" operator	

	if ((gfTokenstr[0] == '0') &&
                    ((gfTokenstr[1] == 'X') || (gfTokenstr[1] == 'x'))) {
        if (gfTokenstr[2] == '-')  {
			//bugfix 11June22:wrong handling length of big negative numbers
			if (numLen > 11)  return 0;
		}
        else if (numLen > 10)  return 0;	//try to parse as command token
		if (string2Int(&gfTokenstr[2], 16))  return 1;
                    // success: interpreted as number, result in "gfTokenNum"
	}
	else if (gfTokenstr[0] == '%') {		//new feature 19Jan24
        if (gfTokenstr[1] == '-')  {
			if (numLen > 12)  return 0;
		}
        else if (numLen > 11)  return 0;	//try to parse as command token
		if (string2Int(&gfTokenstr[1], 10))  return 1;
                    // success: interpreted as number, result in "gfTokenNum"
	}
	else {
        if (gfTempNumBase == 16)  {
			if (gfTokenstr[0] == '-')  {
				if (numLen > 9)  return 0;
			}
            else if (numLen > 8)  return 0;	//try to parse as command token
        }
        else  {
			if (gfTokenstr[0] == '-')  {
				if (numLen > 11)  return 0;
			}
			else if (numLen > 10)  return 0;    //name is too long
		}
        if (string2Int(&gfTokenstr[0], gfTempNumBase))  return 1;
                    // success: interpreted as number, result in "gfTokenNum"
    }
	return 0; //string doesnt represent a number:try to parse as command token
}



uint32_t string2Int(uint8_t* numStr, uint32_t numBase)   // ******************
//return: 1= valid number, 0=no valid number, try to interpret as command token
// ***************************************************************************
{
uint32_t n, length;
uint8_t numString[16];
uint32_t digit;

    for (length = 0; length < 16;length++) {
        numString[length] = numStr[length];
    //transform into capital letters:
		if ((numString[length] >= 'a') && (numString[length] <= 'z'))
                                                    numString[length] -= 0x20;
        if (numString[length] == 0)  break;
    }
    numString[15] = 0;              //in any case, doesn't harm if length<15

    if ((numString[0] == '-') && (length == 1))  return 0;  //"-" operator

    gfTokenNum = 0;
    n = 0;
    if (numString[0] == '-') n = 1;
    while (n < length)  {
        digit = checkDigit(numString[n], numBase);
        if (digit == 0xFFFFFFFF) {
            return 0; //string doesn't represent a number
        }
        gfTokenNum = (int32_t)((gfTokenNum * numBase) + digit);
        n++;
    }
    if  (numString[0] == '-') {
        gfTokenNum = -gfTokenNum;
    }
    return 1;                       //string is parsed as valid number
}


uint32_t checkDigit(uint8_t digit, uint32_t numBase)   // *******************
// *************************************************************************
{
    if (digit < 0x30) return 0xFFFFFFFF;
    digit -= 0x30;
    if (digit < 10) return (int32_t)digit;
    if (numBase == 10) return 0xFFFFFFFF;
    else {
        if (digit < 0x11) return 0xFFFFFFFF; //error if digit < 0x41 (='A')
        if (digit > 0x16) return 0xFFFFFFFF; //error if digit > 0x46 (='F')
        digit -= 7;                         //range 0....0x0F
    }
    return digit;
}


uint32_t findKernelToken(void)   // ****************************************
// *************************************************************************
{
uint32_t k, n;
uint8_t kernelChar, refChar;

    k = 0;					//index 0 is op_ERR
	while (k < KERNELSIZE)  {
        for (n=0; n < 16;n++) {
            kernelChar = gfKernelNames[k][n];
            refChar = gfTokenstr[n];
//transform to capital letters:
            if ((kernelChar >= 'a')&&(kernelChar <= 'z')) kernelChar -= 0x20;
			if ((refChar >= 'a')&&(refChar <= 'z')) refChar -= 0x20;
            if (kernelChar == SPACE) kernelChar = 0;
            if (kernelChar != refChar) break;
            if (refChar != 0) continue;
            gfParsed = k;	//="gfKernelNames[]"index of found token
                    //=same index in "gfKernelOps[]" and "gfKernelAttibutes[]"
            return 1;
        }
		k++;
	}
	return 0;
}


uint32_t findUserToken(void)   // ******************************************
// *************************************************************************
{
int32_t k;
uint32_t n;
uint8_t refChar, userChar;

    k = 0;

	while (gfUserHeader[k].name[0] != 0)	k++;
	k--;                                //k is index of last compiled USER Token

	while (k != -1)  {                  //search DOWN from the LAST COMPILED
		for (n=0; n < 16;n++) {
            userChar = gfUserHeader[k].name[n];
            refChar = gfTokenstr[n];
//transform into capital letters:
            if ((refChar >= 'a')&&(refChar <= 'z')) refChar -= 0x20;
            if (userChar != refChar)  break;
            if (userChar == 0) {
                gfParsed =  k;          //="gfUserHeader[]" index of found token
                return 1;
            }
		}
		k--;
	}
	return 0;
}


void interpretKernelToken(void)   // ***************************************
// *************************************************************************
{
	if (gfKernelAttributes[gfParsed] & A_RUN) {
		gfErrors |= ERR_RUNTOKEN;
		return;
	}

	if(gfCompiState == 0) {		//interpreting, executing threaded code
		if (gfKernelAttributes[gfParsed] & A_CPO) {
			gfErrors |= ERR_COMPONLY;
			return;
		}
		gfKernelOps[gfParsed]();			//execute normally
	}
	else {				// compiling is active
		if (gfKernelAttributes[gfParsed] & A_IME) {
			gfKernelOps[gfParsed]();        //execute immediately,not compile
		}
		else  compileTCode(gfParsed);		//compile this token
	}
}


uint32_t interpretUserToken(void)   // *************************************
// unusual return value: 0= ok, else=Error
// *************************************************************************
{
uint32_t oldTibMax, oldTibPos;

	if(gfCompiState == 0) {	//interpreting, COMPONLY Option not accepted
		if ((gfUserHeader[gfParsed].attributes & A_VARCON) == 0) {
			gfIP = (uint32_t*)(gfUserHeader[gfParsed].toCompile & 0xFFF7FFFF);
												//delete RESTORE bit19 18Oct23
			gfExecUser = 1;			//marks "excuting UserToken"
									//modifies query() behaviour
			oldTibMax = gfTibMax;
			oldTibPos = gfTibPos;
			gfTibMax = 82;			//intermediate TIB for background process
			gfTibPos = 82;
			gfComment = 0;

			executeTCode();
			gfTibMax = oldTibMax;	
			gfTibPos = oldTibPos;
			return gfErrors;		//=0 if everything OK
		}
		else  {						//VARCON
			gfSP++;
			*gfSP = gfUserHeader[gfParsed].toCompile;			//special
			return 0;				//everything OK
		}
	}
	else {						// compiling
		if ((gfUserHeader[gfParsed].attributes & A_VARCON) == 0) {
            compileTCode(gfUserHeader[gfParsed].toCompile);
			return gfErrors;	//=0 if everything OK
		}
		else  {					//VARCON -- modified 19Jan25
			compileTCode((uint32_t)(&gfUserHeader[gfParsed].toCompile));
			return gfErrors;	//=0 if everything OK
		}
	}
    return ERR_SYSTEM;       //this exit should be impossible;
}


uint32_t compileUserSymbol(uint32_t attribute)   // *************************
// **************************************************************************
{
uint32_t n;
uint8_t userString[16];
const uint8_t warnStr [] = " Warning: will override Kernel OP w.same name! ";

    gfTokenStart = gfTibPos;
	scanNext();
    gfTokenStart = gfTibPos;

    if (interpretNumber() != 0)  {
        gfErrors |= ERR_CONTEXT;
        return 0;
    }

    if (findKernelToken())  {
		sendString(warnStr);
        if (conFirm() == 0)  return 2;    //not compiled, no error flag set
    }

    for (n = 0; n < 16;n++) {
		userString[n] = gfTokenstr[n];
//transform to capital letters:
		if ((userString[n] >='a')&&(userString[n] <='z')) userString[n]-=0x20;
		if (userString[n] == 0) break;
    }
    userString[15] = 0;              //in any case, doesn't harm if length<15
    n = 0;
    while ((userString[n] != 0) && (n < 15))  {
      gfUserHeader[gfNextNameCompi].name[n] = userString[n];
      n++;
    }
    gfUserHeader[gfNextNameCompi].name[n+1] = 0;
	gfUserHeader[gfNextNameCompi].toCompile =(gfCodeIndex *4)+(uint32_t)gfTCode;
	gfUserHeader[gfNextNameCompi].attributes = gfCodeIndex | attribute;
	return 1;
}


void compileTCode(uint32_t token)   // *************************************
// *************************************************************************
{
	gfTCode[gfCodeIndex] = token;
	gfCodeIndex ++;
	if (gfCodeIndex >= CODETOP)  gfErrors |= ERR_USERCODE;
}


void executeTCode(void)   // ***********************************************
// *************************************************************************
{
uint32_t tibstate;
const uint8_t finiStr [] = " debug finished ";
uint32_t *test;

    while (gfIP < gfTCode + USERTOP)  {
		if (gfBackProcess == 0)  {
//			if (gfTermCtrlBC == CTRL_B)  op_BAKOP();		
								//inserted 11June22 set in USB servEP2_Out()}
			if (*gfIP < KERNELSIZE)  gfKernelOps[*gfIP]();	//is Kernel Token
			else  {									//then it is a User Token
				if (*gfIP >= (uint32_t)&gfUserHeader[0]) {	//new 19Jan25
									//is address of userHeader[x].toCompile
					test = (uint32_t*)(*gfIP);
					if (*(test+1) & A_VARCON)  {  //->userHeader[x].attributes
						gfSP++;
						*gfSP = (uint32_t)(*(uint32_t*)*gfIP);
					}
				}
				else  intoUserCode();
			}
		}
		else  {
			if (checkTerminal())  {
				tibstate = subQuery();
				if (tibstate == 1)  {					//Terminal CRR input
					backProcEval();
				}
				if (tibstate == 2)  {					//Terminal ESC input
					gfErrors |= ERR_CANCEL;
					return;
				}
//tibstate =3(step) ignored in this case
			}
			if (*gfIP < KERNELSIZE)  {						//is Kernel Token
				if (gfBackProcess == 3)  execNext();
				else  gfKernelOps[*gfIP]();
			}
			else  {									//then it is a User Token
				if (checkVarcon() == 0) intoUserCode();	//new 19Jan25
			}
		}

//serve HW Background process
#ifdef DMX
	if (gfDmxBusy)  {
  #ifdef MINI
		if (USART1->ISR & USART_ISR_TXE) {
			USART1->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
			gdDmxTxIndex++;
			if (gdDmxTxIndex >= gdDmxMax)  DmxHeader() ;
		}
  #else
		if (USART2->ISR & USART_ISR_TXE) {
			USART2->TDR = gdDMXBUF[gdDmxTxIndex];
// the TXE bit is cleared automatically when TDR is written
			gdDmxTxIndex++;
			if (gdDmxTxIndex >= gdDmxMax)  DmxHeader() ;
		}
  #endif
		if (gdFaderBusy)  {
			if (gSYSCOUNT == 0)  {		//countdown in SysTick_Handler()
				gSYSCOUNT = 10;
				fadeProcess();			//handle DMX faders
			}
		}
	}
#endif

#ifdef LQFP48_DMX
		serveBack();
#else
	if (gfTermCtrlBC == CTRL_B)  op_BAKOP();	//set by USB lowlevel RX
  #ifndef MINI
		if (gSplit >= 3) Rx2Midi();
  #endif
		if  (gfStepSpeed1)  {
			if ((gfUserHeader[4].toCompile == 0xFFFFFFFE) || 
										(gfUserHeader[4].toCompile > 0))  {
				if (gfSTEPCOUNT1 == 0) subMOT1();
			}
		}
		if  (gfStepSpeed2)  {
			if ((gfUserHeader[5].toCompile == 0xFFFFFFFE) || 
										(gfUserHeader[5].toCompile > 0))  {
				if (gfSTEPCOUNT2 == 0) subMOT2();
			}
		}
  #ifndef MINI
 		if  (gfStepSpeed3)  {
			if ((gfUserHeader[6].toCompile == 0xFFFFFFFE) || 
										(gfUserHeader[6].toCompile > 0))  {
				if (gfSTEPCOUNT3 == 0) subMOT3();
			}
		}
  #endif
#endif
		if (gfErrors) return;
		if (gfIP)  {				//mpdified 29Jan25
			gfIP ++;
			continue;
		}
		else  {						//execution finished by SEMIS or RETURN
			if (gfBackProcess == 3) sendString(finiStr);
			if (gfBackProcess)  gfBackProcess = 1;	//reset to user selection
			gfExecUser = 0;
			gfBackNumValid = 0;
			gfNestLevel = 0;
			return;
		}
	}
}


void intoUserCode(void)   // ***********************************************
// *************************************************************************
{
	gfNestLevel++;
	gfRP ++;
	*gfRP = (uint32_t)gfIP;						//gets restored by "op_SEMIS"
	gfIP = (uint32_t*)(*gfIP & 0xFFF7FFFF);		//delete RESTORE bit19 18Oct23
	
	if ((*gfIP & 0xFFF7FFFF) < KERNELSIZE)  {		//is Kernel Token
		if (gfBackProcess != 3)  gfKernelOps[*gfIP]();
		else  execNext();
	}
	else  {											//then it is a User Token
		if (checkVarcon() == 0) intoUserCode;		//new 19Jan25
	}
}


uint32_t checkVarcon(void)   // *********************************************
//new 19Jan25
// **************************************************************************
{
uint32_t result;
uint32_t *test;

	if (*gfIP >= (uint32_t)&gfUserHeader[0]) {	
									//is address of userHeader[x].toCompile
		test = (uint32_t*)(*gfIP);	//points to userHeader[x].attributes
		if (*(test+1) & A_VARCON)  {
			gfSP++;
			*gfSP = (uint32_t)(*(uint32_t*)*gfIP);
			result = 1;
		}
	}
	else  result = 0;

	if (gfBackProcess == 3)  {
		showPreStep(result);
		if (gfErrors) return result;
		showPostStep();
	}
	return result;
}


void execNext(void)   // ***************************************************
//abbreviation to save runtime if not in "step mode"
// *************************************************************************
{
    showPreStep(0);
    if (gfErrors) return;		//usually ESC = terminate User Thread
    gfKernelOps[*gfIP]();
    showPostStep();
}


void showPreStep(uint32_t varMode)  // **************************************
// First the name of the next code token is sent to terminal.
// Execution of this token is triggered by TAB key on terminal software.
// Finally (by "showPostStep") the result is sent to terminal (output + stack)
// can be used as a simple Forth-Decompiler, too
// to speed display up, keep key pressed
// **************************************************************************
{
uint32_t k, ipVal, indent;
uint32_t *localIP;
uint8_t *nameAddr;
uint32_t *value;

const uint8_t IPStr [] = "(IP->";

	if (gfErrors)  return;

	gfIPold = gfIP;					//keep in mind for showPostStep
	ipVal = *gfIP & 0xFFF7FFFF;		//delete RESTORE bit19 18Oct23
    indent = gfNestLevel * 3;

    sendTB(CRR);
    for (k=0; k < indent; k++) sendTB(SPACE);
		sendNX32((uint32_t)gfIPold & 0x0000FFFF);
		sendTB(SPACE);

	if (varMode)  {					//new 19Jan25
		valAdr2String(ipVal);
		sendTB(0x28);	//'('
		sendTB('=');
		value = (uint32_t*)ipVal;
		if (gfTempNumBase==16)  send0X32(*value);
		else  sendNS32(*value);
		sendTB(0x29);	//')'
	}
	else  {
		sendNX32 (ipVal & 0x0000FFFF);
		sendTB(SPACE);
		if (ipVal > KERNELSIZE)  {      //user token
			sendTB('-');
			sendTB('>');
			sendTB(SPACE);
		}
		token2String(ipVal);
	}
	
	if(((ipVal == IDX_DOIF) && (*gfSP == 0))  ||
			((ipVal == IDX_DOCASE) && (*gfSP != *(gfSP-1))) ||
			((ipVal == IDX_DOLTCAS) && (*(gfSP-1) >= *gfSP)) ||
			((ipVal == IDX_DOGTCAS) && (*(gfSP-1) <= *gfSP)) ||
			(ipVal == IDX_DOELSE) ||
			((ipVal == IDX_DOWHILE) && *gfSP) ||
			((ipVal == IDX_DOUNTIL) && (*gfSP == 0)) ||
			(ipVal == IDX_DOLOOP) || (ipVal == IDX_DOAGAIN)) {
		sendString(IPStr);
		sendNX32((*(gfIP+1) + 4) & 0x0000FFFF);
                        //gfIP is incremented before "executeTCode()" is left
        sendTB(0x29);	                    //')'
    }

    if (ipVal==IDX_DOLIT) {
        sendTB(0x28);	 //'('
        if (gfTempNumBase==16)   send0X32(*(gfIP+1));
        else  sendNS32(*(gfIP+1));
        sendTB(0x29);	 //')'
    }

    if (ipVal==IDX_DOWVA) {
        sendTB(0x28);	 //'('
		valAdr2String(*(gfIP+1));
        sendTB('@');
        send0X32(*(gfIP+1) & 0x0000FFFF);
        sendTB(0x29);	 //')'
    }

    if (ipVal==IDX_SEMIS) {
        sendTB(SPACE);
        sendTB('<');
        sendTB('-');
    }

	k = 0;
	while (k != 3)  {			//while terminal input != TAB
		if (gfBackProcess < 3) break;
		servSystem();
		if (checkTerminal())  {
			k = subQuery();
			if (k == 1)   {		//Terminal CRR input
				backProcEval();
				continue;	//background operation does not release step
			}
			if (k == 2)  {		//Terminal ESC input
				gfErrors |= ERR_CANCEL;
				return;
			}
//new 31Dec23:
			if (k == 4)  {		//Terminal CTRL_R input
				if (gfNestLevel != 0)  {
					gfBackNestLevel = gfNestLevel;
					gfBackProcess = 2;
				}
				else  k = 3; //continue;
			}
		}
	}

	sendTB(SPACE);
	if ((ipVal >= IDX_CONTI) && (ipVal <= IDX_ZBREAK))  {
		sendString(IPStr);
	}
	else  {
		sendTB('{');
		sendTB(SPACE);
	}
}


void showPostStep(void)  // ************************************************
// remark see "showPreStep()"
// *************************************************************************
{
uint32_t n, ipVal;

	if (gfErrors)  return;


    ipVal = *gfIPold & 0xFFF7FFFF;			//delete RESTORE bit19 18Oct23

	if ((ipVal >= IDX_CONTI) && (ipVal <= IDX_ZBREAK))  {
		sendNX32(((uint32_t)(gfIP+1)) & 0x00000FFF);
                        //gfIP is incremented before "executeTCode()" is left
        sendTB(0x29);	
		return;
	}
	sendTB('}');
    sendTB(SPACE);
    op_PRINTSTACK();

    if ((ipVal >= IDX_DODO) && (ipVal <= IDX_RP4DEC))  {
        sendTB(SPACE);
        sendTB('R');
        sendTB(':');
    }
    if ((ipVal == IDX_DOFOR) || (ipVal == IDX_DOLOOP) ){
        if (((uint32_t)(gfRP - gfReturnStack) > 2)) {
            sendNX32((*(gfRP-3) & 0x0000FFFF));
            sendTB(SPACE);
            sendTB(SPACE);
            sendNX32((*(gfRP-2) & 0x0000FFFF));
            sendTB(SPACE);
            sendTB('i');
            sendNX32((*(gfRP-1) & 0x0000FFFF));
            sendTB(SPACE);
            sendTB('x');
            sendNX32((*gfRP & 0x0000FFFF) + 4);
                //the compiled jump address is 1 less than effectively executed
        }
    }
    else if ((ipVal >= IDX_DODO) && (ipVal <= IDX_DOAGAIN))
                                            sendNX32(*gfRP & 0x0000FFFF);
}


void emptyForth(void)// ****************************************************
//load and install an empty project with default settings.
//Two VARs N2 and N4 (necessary for motor control) are installed as user token
//Two VARs V0, V1 and two  CONSTants P2, P3 are installed as user token
// *************************************************************************
{
uint32_t k;
uint32_t *pointer;

    pointer = (uint32_t*)&(gfUserHeader[0]);
    while (pointer < (uint32_t*)&gfUserHeader[MAXUSER])  {
        *pointer = 0;
        pointer++;
    }
    for (k = 0; k <= CODETOP; k++)  {
        gfTCode[k] = 0;
    }

	gProject = 0;		//set even if another project loaded with virgin flash
	gfNumBase = DEFAULT_NUMBASE;
	gfTempNumBase = DEFAULT_NUMBASE;
	gfBackProcess = 0;		//is default mode -- changed 26Jan25
	gfAuto = 0;
#ifndef MINI
	gMIDI_HEAD = 0;			//flush MIDI buffer
    gMIDI_TAIL = 0;
#endif
//default SPI setup:
	gfSpiMode = 0;
	gfSpiBits = 8;
	gfSpiClock = 50;		//about 50 kHz clock
//default PWM setup:
	gfFreq12 = 0;
    gfPWM1 = 0;
    gfPWM2 = 0;
	gfFreq34 = 0;
    gfPWM3 = 0;
    gfPWM4 = 0;
//optimize speed of background process
	gfStepSpeed1 = 0;
	gfStepSpeed2 = 0;
#ifndef MINI
	gfStepSpeed3 = 0;
#endif
#ifdef DMX
	gfDmxBusy  =1;			//DMX ON by default
#endif
//compile 4 VarCon for general use
	for (k = 0; k < 4;k++) {
		gfUserHeader[k].name[0] = 'V';
		gfUserHeader[k].name[1] = k + 0x31;  //V1...V4
		gfUserHeader[k].name[2] = 0;
		gfUserHeader[k].attributes = A_VARCON;
		gfUserHeader[k].toCompile = 0;
	}
//compile 3 VARCONs NSTP1,2,3(no.of steps) for op_MOT1,2,3 background control
	gfUserHeader[4].name[0] = 'N';
	gfUserHeader[4].name[1] = 'S';
	gfUserHeader[4].name[2] = 'T';
	gfUserHeader[4].name[3] = 'P';
	gfUserHeader[4].name[4] = '1';
	gfUserHeader[4].name[5] = 0;
	gfUserHeader[4].attributes = A_VARCON;
	gfUserHeader[4].toCompile = -1;		//start with DC motor
	gfUserHeader[5].name[0] = 'N';
	gfUserHeader[5].name[1] = 'S';
	gfUserHeader[5].name[2] = 'T';
	gfUserHeader[5].name[3] = 'P';
	gfUserHeader[5].name[4] = '2';
	gfUserHeader[5].name[5] = 0;
	gfUserHeader[5].attributes = A_VARCON;
	gfUserHeader[5].toCompile = -1;		//start with DC motor
	gfUserHeader[6].name[0] = 'N';
	gfUserHeader[6].name[1] = 'S';
	gfUserHeader[6].name[2] = 'T';
	gfUserHeader[6].name[3] = 'P';
	gfUserHeader[6].name[4] = '3';
	gfUserHeader[6].name[5] = 0;
	gfUserHeader[6].attributes = A_VARCON;
	gfUserHeader[6].toCompile = -1;		//start with DC motor
	gfNextNameCompi = 7;
//#endif
	gfCodeIndex = 0;
	op_ABORT();
}

// arithmetic ops  #########################################################

void op_ERR(void)  // *****************************************************
// ( -- )
// does nothing, except consume a little bit of runtime
// *************************************************************************
{
	gfErrors |= ERR_USERCODE;
}


void op_DROP(void)  // *****************************************************
// ( x1 x2 -- x1)
// forget the TOS and pop all lower stack entries one up
// *************************************************************************
{
	gfSP--;
	if (gfSP < gfDataStack + 3) gfErrors |= ERR_STACK;
}


void op_DUP(void)  // ******************************************************
// ( x -- x x)
// push all stack entries one down and copy the new NOS as TOS
// *************************************************************************
{
	gfSP++;
	if (gfSP >= gfDataStack+MAXSTACK - 4)  {
		gfErrors |= ERR_STACK;
		return;
	}
	*gfSP = *(gfSP-1);
}


void op_SWAP(void)  // *****************************************************
// (x1 x2 -- x2 x1)
// exchange TOS and NOS
// *************************************************************************
{
uint32_t TEMP;

	TEMP = *(gfSP-1);
	*(gfSP-1) = *gfSP;
	*gfSP = TEMP;
}


void op_OVER(void)  // *****************************************************
// ( x1 x2 -- x1 x2 x1)
// push all stack entries one down and copy the new 3rd of stack to TOS
// *************************************************************************
{
	gfSP++;
	*gfSP = *(gfSP-2);
}


void op_ROT(void)  // ******************************************************
// ( x1 x2 x3 -- x2 x3 x1 )
// rotate 3rd stack entry up
// *************************************************************************
{
uint32_t TEMP;

	TEMP = *(gfSP-2);
	*(gfSP-2) = *(gfSP-1);
	*(gfSP-1) = *gfSP;
	*gfSP = TEMP;
}


void op_DNROT(void)  // ****************************************************
// ( x1 x2 x3 -- x3 x1 x2 )
// rotate TOS down
// *************************************************************************
{
uint32_t TEMP;

	TEMP = *gfSP;
	*gfSP = *(gfSP-1);
	*(gfSP-1) = *(gfSP-2);
	*(gfSP-2) = TEMP;
}


void op_PICK(void)  // *****************************************************
// (  xn x(n-1) x2 x1 n  ---  xn x(n-1) ... x2 x1 x0 xn  )
// keep  n(=TOS) in mind and copy the (n+1)th element of stack on TOS
// 0 PICK performs DROP
// 1 PICK effectively performs DUP
// 2 PICK effectively performs OVER
// *************************************************************************
{
int32_t temp;

//first check available depth of stack
    temp = (int32_t)(gfSP - gfDataStack - 4) - (int32_t)*gfSP;
    if  (temp < 0) {
        gfErrors |= ERR_STACK;
        return;
    }
	temp = *gfSP;
	if (temp == 0) {
		gfSP--;
		return;
	}
	*gfSP = *(gfSP - temp);
}


void op_ROLL(void)  // *****************************************************
// (  xn x(n-1)...x2 x1 n --- x(n-1)  x2 x1 xn  )
// drops TOS, but keeps n in mind. Rotates the (new) n-th stack element 
// up to TOS. 
// 0 ROLL and 1 ROLL do nothing except dropping TOS
// 2 ROLL is equivalent to SWAP
// 3 ROLL is equivalent to ROT
// bigger values of "n" provide manipulation of deep stack items
// e.g. the sequence  "4 ROLL 4 ROLL" effectively performs 2SWAP
// *************************************************************************
{
uint32_t k, temp;

//first check available depth of stack
    k = (int32_t)(gfSP - gfDataStack - 4) - (int32_t)*gfSP;
    if  (k >= 0x80000000) {
        gfErrors |= ERR_STACK;
        return;
    }

	k = *gfSP;
	gfSP--;
	if (k < 2) return;
	temp = *(gfSP - k + 1); 
	while(k)  {
		*(gfSP - k + 1 ) = *(gfSP - k + 2);
		k -- ;
	}
	*gfSP = temp;
}


void op_INJECT(void)  // ****************************************************
// (  x(n+1) xn... x2 x1 n  ---  x1 x(n+1) xn... x2  )
// drops TOS, keeps n in mind. 
// then rotates (new) TOS(= x1) down below the (actual) n-th stack element.
// 0 INJECT and 1 INJECT simply drop TOS
// 2 INJECT effectively performs SWAP
// 3 INJECT effectively performs - ROT
// bigger values of "n" provide manipulation of deep stack items
// e.g.sequence "n PICK n INJECT" effectively doubles the deep n-th stack item
// **************************************************************************
{
uint32_t k, n, temp;

//first check available depth of stack
    k = (int32_t)(gfSP - gfDataStack - 4) - (int32_t)*gfSP;
    if  (k >= 0x80000000) {
        gfErrors |= ERR_STACK;
        return;
    }

	temp = *(gfSP - 1);
	n = *gfSP;
	gfSP --;
	if (n < 2) return;
	for (k=0;k < (n-1);k++)  {
		*(gfSP - k) = *(gfSP - k - 1);
	}
	*(gfSP - n + 1) = temp;
}


void op_PLUS(void)  // *****************************************************
// ( x1 x2  -- x1+x2)
// *************************************************************************
{
	*(gfSP-1) = (int32_t)*(gfSP-1) + (int32_t)*gfSP;
	gfSP--;
}


void op_INCT(void)  // *****************************************************
// ( x -- x+1)
// *************************************************************************
{
	*gfSP = *gfSP + 1;
}


void op_MINUS(void)  // ****************************************************
// ( x1 x2  -- x1-x2)
// *************************************************************************
{
	*(gfSP - 1) = (int32_t)*(gfSP - 1) - (int32_t)*gfSP;
	gfSP--;
}


void op_DECT(void)  // *****************************************************
// ( x -- x-1)
// *************************************************************************
{
	*gfSP = *gfSP - 1;
}


void op_ABS(void)  // ******************************************************
// ( x -- abs(x))
// *************************************************************************
{
int32_t abs;

	abs = (int32_t)*gfSP;
	if (abs < 0) *gfSP = -abs;
}


void op_SHIFT(void)  // ***************************************************
// if (X2>=0) ( x1 x2 -- x1<<x2 )
// if (X2<0) ( x1 x2 -- x1>>x2 )
// abs(x2) <= 31
// *************************************************************************
{
int32_t k,abs;

	k = (int32_t)*gfSP;
	abs = k;
	if (abs < 0) abs = -abs;
		if (abs > 31)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
	gfSP--;
	if (k >= 0)   {
		*(gfSP) = *(gfSP)<<k;
	}
	else  {
		*(gfSP) = *(gfSP)>>abs;
	}

}


void op_AND(void)  // ******************************************************
// ( x1 x2  -- x1&x2)
// *************************************************************************
{
	*(gfSP-1) &= *gfSP;
	gfSP--;
}


void op_OR(void)  // *******************************************************
// ( x1 x2  -- x1|x2)
// *************************************************************************
{
	*(gfSP-1) |= *gfSP;
	gfSP--;
}


void op_XOR(void)  // ******************************************************
// ( x1 x2  -- x1 XOR x2)
// *************************************************************************
{
	*(gfSP-1) = *gfSP ^ *(gfSP-1);
	gfSP--;
}


void op_NOT(void)  // ******************************************************
// ( x --- <one's complement of x>)
// logical inversion bit by bit (one's complement)
// in contrast to other Forth,this is an arithmetic operator differing from "0="
// *************************************************************************
{
	*gfSP = ~ (*gfSP);
}


void op_NEG(void)  // ******************************************************
// ( x --- <two's complement of x>)
// mathematical negation (two's complement)
// *************************************************************************
{
	*gfSP = -(int32_t)(*gfSP);
}


void op_MULT(void)  // *****************************************************
// ( factor factor  -- TOS=factor*factor)
// multiplication is made in 32bit signed-i.e. may overflow with factors>0xB000
// if overflow would happen, an error message is sent
// *************************************************************************
{
register float fact1, fact2;
    fact1 = (float)((int32_t)*gfSP);
    fact2 = (float)((int32_t)*(gfSP-1));
    fact1 = fact1 * fact2;
//float used for range check only:
    if ((fact1 > 2147483647.) || (fact1 < -2147483647.))  {
        gfErrors |= ERR_NUMVAL;
        return;
    }
   *(gfSP-1) = (int32_t)*gfSP * (int32_t)(*(gfSP-1));
	gfSP--;
}


void op_DIV(void)  // ******************************************************
// ( dividend divisor  -- x1/x2)
// 32 bit signed division
// *************************************************************************
{
	if (*gfSP == 0)  {
		gfErrors |= ERR_DIVZERO;
		return;
	}
	else  {
		*(gfSP-1) = (int32_t)*(gfSP-1) / (int32_t)*gfSP;
		gfSP--;
	}
}


void op_MULTDIV(void)  // **************************************************
// ( factor factor divisor-- TOS=(factor*factor)/divisor)
// multiplication is made in 32bit signed i.e. may overflow.
// if this case, the calculation is made in "float" and reconverted to int
// *************************************************************************
{
int32_t factor, dividend, divisor;
register float ffact, fdivid, fdivis;

    if (*gfSP == 0)  {
        gfErrors |= ERR_DIVZERO;
        return;
    }
	else  {
        ffact= (float)((int32_t)*(gfSP - 2));
        fdivid = (float)((int32_t)*(gfSP-1));
//        divisor =  (float)(int32_t)*gfSP;
        fdivid = ffact * fdivid;
        if ((fdivid > 2147483647.) || (fdivid < -2147483647.))  {
//use intermediate float calculations
            fdivis =  (float)(int32_t)*gfSP;
            fdivid = fdivid / fdivis;
            if ((fdivid > 2147483647.) || (fdivid < -2147483647.))  {
                gfErrors |= ERR_NUMVAL;
                return;
            }
            gfSP -= 2;
            * gfSP = (int32_t)fdivid;
        }
        else  {
//if 32bit range not exceeded, use integer calculation
            factor= (int32_t)*(gfSP - 2);
            dividend = (int32_t)*(gfSP-1);
            divisor =  (int32_t)*gfSP;
            dividend = factor * dividend;
            gfSP -= 2;
            * gfSP = dividend / divisor;
        }
	}
}


void op_MOD(void)  // ******************************************************
// return the remainder of a 32 bit signed integer division
// ( dividend divisor  -- remainder )
// the remainder has always the same sign as the dividend
// *************************************************************************
{
    if (*gfSP == 0)  {
        gfErrors |= ERR_DIVZERO;
        return;
    }
	else  {
		*(gfSP-1) = (int32_t)*(gfSP-1) % (int32_t)(*gfSP);
		gfSP--;
	}
}


void op_DIVMOD(void)  // ******************************************************
// return the remainder(NOS) and quotient(TOS) of a 32 bit signed int. division
// ( dividend divisor  -- remainder quotient )
// the remainder has always the same sign as the dividend
// *************************************************************************
{
int32_t dividend, divisor;

	if (*gfSP == 0)  {
		gfErrors |= ERR_DIVZERO;
		return;
	}
	else  {
		dividend = (int32_t)*(gfSP-1);
		divisor =  (int32_t)*gfSP;
		*(gfSP-1) = dividend % divisor;
		*gfSP = dividend / divisor;
	}
}


void op_BSET(void)  // ****************************************************
// (bit# -- ) read modify write
// Top of Stack (TOS) manipulation
// bit# = bit number to be set (0 to 31)
// *************************************************************************
{
	if (*gfSP > 31) {
		gfErrors |= ERR_NUMVAL;
	return;
	}
	*(gfSP-1) |= 1 << *gfSP;        //bitNo-->bitMask and set NOS
	gfSP--;                         //NOS-->TOS
}


void op_BCLR(void)  // *****************************************************
// (bit# -- ) read modify write
// Top of Stack (TOS) manipulation
// bit# = bit number to be set (0 to 31)
// *************************************************************************
{
	if (*gfSP > 31) {
		gfErrors |= ERR_NUMVAL;
		return;
	}
	*(gfSP-1) &= ~(1 << *gfSP);   //bitNo-->bitMask and invert NOS bitwise
	gfSP --;                        //NOS-->TOS
}


void op_BTST(void)  // *****************************************************
// (n bit# -- bitvalue)
// Top of Stack (TOS) manipulation
// bit# = bit number to be checked (0 to 31)
// new TOS = (bitvalue = boolean value of this bit (0 or 1)
// *************************************************************************
{
uint32_t temp;

	if (*gfSP > 31) {
		gfErrors |= ERR_NUMVAL;
		return;
	}
	temp = ~(1 << *gfSP);      //bitNo-->bitMask, invert
	temp = *(gfSP-1) & temp;
	gfSP--;
	if (temp) *gfSP = 1;
	else *gfSP = 0;
}


void op_RANDOM(void)  // ***************************************************
// (  -- number)
// number= uint 32bit random number
// renewed during each SysTick_Handler() interrupt (1kHz)
// *************************************************************************
{
	gfSP ++;
	*gfSP = gfRandom;
}

//comparison ops ###########################################################

void op_EQUAL(void)  // ****************************************************
//  ( x1 x2  -- TRUE|FALSE)
//	TRUE if x1==X2
// *************************************************************************
{
	if (*gfSP == *(gfSP-1)) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_UNEQU(void)  // ****************************************************
//  ( x1 x2  -- TRUE|FALSE)
//	TRUE if x1!=X2
// *************************************************************************
{
    if (*gfSP != *(gfSP-1)) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_ZEQ(void)  // ******************************************************
//  ( x   -- TRUE|FALSE)
//	TRUE if x==0, FALSE if x!=0
// *************************************************************************
{
	if (*gfSP == 0) *gfSP = 1;
	else *gfSP = 0;
}


void op_GRTR(void)  // *****************************************************
//  ( x1 x2  -- TRUE|FALSE)
// signed compare
//	TRUE if x1>X2
// *************************************************************************
{
	if ((int32_t)*(gfSP-1) > (int32_t)*gfSP) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_GRTREQ(void)  // ***************************************************
//  ( x1 x2  -- TRUE|FALSE)
// signed compare
//	TRUE if x1>=X2
// *************************************************************************
{
	if ((int32_t)*(gfSP-1) >= (int32_t)*gfSP) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_LESS(void)  // *****************************************************
//  ( x1 x2  -- TRUE|FALSE)
// signed compare
//	TRUE if x1<X2
// *************************************************************************
{
	if ((int32_t)*(gfSP-1) < (int32_t)*gfSP) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_ULESS(void)  // *****************************************************
//  ( x1 x2  -- TRUE|FALSE)
// unsigned compare
//	TRUE if x1<X2
// *************************************************************************
{
	if ((uint32_t)*(gfSP-1) < (uint32_t)*gfSP) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}


void op_LESSEQ(void)  // ***************************************************
//  ( x1 x2  -- TRUE|FALSE)
// signed compare
//	TRUE if x1<=X2
// *************************************************************************
{
	if ((int32_t)*(gfSP-1) <= (int32_t)*gfSP) *(gfSP-1) = 1;
	else *(gfSP-1) = 0;
	gfSP-- ;
}

// memory ops  #############################################################

void op_VARCON(void)  // ****************************************************
// ( value -- )
// syntax example:
// 23 VARCON CONNY
// defines the VariableConstant CONNY and initializes it with value 2?
// at runtime, a VARCON puts its value on TOS
// *************************************************************************
{
    if (compileUserSymbol(A_VARCON) == 0)  {
    	gfErrors |= ERR_STRINGTOKEN;
	return;
    }
    gfUserHeader[gfNextNameCompi].toCompile = *gfSP;
    gfSP--;
    gfNextNameCompi++;
    return;
}


void op_WVA(void)  // ******************************************************
// wrtite a new value to a VariableConstant
// special non-Forth typical syntax,
// example:
// 23 VARCON TEST
// 0xABCDE W TEST									----  changed 19Jan25
// (TEST must be scanned by W to get its .toCompile address)
// TEST .S   =returns 0xABCDE on stack
// the new value is inserted at any place where VARCON is threaded
// and will be used in future until a new value is inserted
// *************************************************************************
{
uint32_t n;

	gfTokenStart = gfTibPos;
	scanNext();
	if (findUserToken() == 0) {		//updates "gfParsed"=index of gfUserHeader
		gfErrors |= ERR_USERCONTEXT;
		return;
	}
	gfTokenStart = gfTibPos;

    if ((gfUserHeader[gfParsed].attributes & A_VARCON) == 0)  {
 		gfErrors |= ERR_CONTEXT;
		return;
    }
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
		}
	if (gfCompiState == 0)  {	    //interpreting
		gfUserHeader[gfParsed].toCompile = *gfSP; //re-init with stack value
		gfSP--;
	}
	else  {	    //compiles runtime "W" behaviour
	    compileTCode(IDX_DOWVA);
	    compileTCode((uint32_t)(&gfUserHeader[gfParsed].toCompile));
					//"gfUserHeader[gfParsed].toCompile" referenced by index !
	}
}


void op_ARRAY(void)	// *****************************************************
// 32bit signed integer array, 1000 entries. index = 0 cannot be used
// positive TOS=index: read value from array[-index] to TOS (index -- value)
// negative TOS=index: write NOS into array[-index] (value -index --)
// *************************************************************************
{
int32_t index;
static int32_t array[1000];

	index = (int32_t)*gfSP;				//copy TOS to index
	if ((index == 0) || (index > 1000) || (index < -1000))  {
		gfErrors |= ERR_NUMVAL;
		return;
	}
	if (index > 0)  {					//read array[-index] (-index --- value)
		*gfSP = array[index-1];			//adjust user index to array index
										//no stack pointer adjustment needed
	}
	else  {								//write NOS into array[-index]
										//TOS=index, NOS=value to be written
										// (value -index --- )
		index = - index;				//convert to effective index
		array[index-1] = *(gfSP-1);		//adjust user index to array index
										//(gfSP-1) is pointer to NOS
										//*(gfSP-1) is NOS value
		gfSP -= 2;						//adjust stack pointer:2 items removed
	}
}
 
// terminal ops  ###########################################################

void op_KEYQ(void)  // ****************************************************
//check if a character is received via serial terminal port (USB and/or USART)
//if NOT return 0 on TOS
//if YES return(number of bytes in RX buffer) on TOS
//exception: active background process grabs terminal input letters
//NO data byte is extracted from terminal input buffer
// *************************************************************************
{
    gfSP++;
    if (gfBackProcess && gfExecUser)  *gfSP = gfTermCheck;		
					//exclusively set '1' in subQuery(), cleared in op_KEY()
	else *gfSP = checkTerminal();
}


uint32_t checkTerminal(void)  // *******************************************
//check if a character is received via serial terminal port (USB and/or USART)
//if NOT return 0
//if YES return number of bytes in RX buffer
//SPLIT assignment:
//'0'=USB and USART both terminal,
//'1'=USB is terminal, USART is COM I/O port
//'2'=USART is terminal, USB is virtual COM I/O port
//'3'=USB is terminal, USART is MIDI interface
//'4'=USART is terminal, USB is MIDI interface
// *************************************************************************
{
#ifdef MINI
	if ((guConfigured == 0) || (gUSB_RXTAIL == gUSB_RXHEAD))  return 0;
	if (gUSB_RXTAIL < gUSB_RXHEAD)  {
		return ((uint32_t)(gUSB_RXHEAD - gUSB_RXTAIL));
	}
	else  return (RXBUFLEN + (uint32_t)(gUSB_RXTAIL - gUSB_RXHEAD));
#else
   if (gSplit == 0)  {
        if ((guConfigured) && (gUSB_RXTAIL != gUSB_RXHEAD))  {
            if (gUSB_RXTAIL < gUSB_RXHEAD)  {
                return ((uint32_t)(gUSB_RXHEAD - gUSB_RXTAIL));
            }
            else  return (RXBUFLEN + (uint32_t)(gUSB_RXTAIL - gUSB_RXHEAD));
        }
        if (gUSART_RXTAIL==gUSART_RXHEAD) return 0;
        if (gUSART_RXTAIL < gUSART_RXHEAD)  {
           return ((uint32_t)(gUSART_RXHEAD - gUSART_RXTAIL));
        }
        else return (RXBUFLEN+(uint32_t)(gUSART_RXTAIL-gUSART_RXHEAD));;
    }
    else if ((gSplit == 2) || (gSplit == 4))  {
        if (gUSART_RXTAIL==gUSART_RXHEAD) return 0;
        if (gUSART_RXTAIL < gUSART_RXHEAD)  {
            return ((uint32_t)(gUSART_RXHEAD - gUSART_RXTAIL));
        }
        else  return (RXBUFLEN+(uint32_t)(gUSART_RXTAIL-gUSART_RXHEAD));
    }
	else  {
		if( (guConfigured == 0) || (gUSB_RXTAIL == gUSB_RXHEAD))  return 0;;
        if (gUSB_RXTAIL < gUSB_RXHEAD)  {
            return ((uint32_t)(gUSB_RXHEAD - gUSB_RXTAIL));
        }
        else  return (RXBUFLEN + (uint32_t)(gUSB_RXTAIL - gUSB_RXHEAD));
    }
#endif
}


void op_KEY(void)  // ******************************************************
//wait until a character is received via serial terminal port (USB and/or USART)
//and return it on the stack
// *************************************************************************
{
uint32_t recvword;

    if (gfBackProcess && gfExecUser)  {
		if (gfTermCheck)  {
			gfSP++;
			*gfSP = gfTermChar;		//set in subQuery()
			gfTermCheck = 0;
		}
	}
	else  {
		while ((recvword = pullTermRxBuf())==0); 
									//watchdog is fed in the background
		gfSP++;
		*gfSP = recvword & 0x000000FF;
	}
}


void op_EMIT(void)  // *****************************************************
//send the lower 8 bits of TOS as BYTE!! via serial terminal port
// *************************************************************************
{
    sendTB((uint8_t)(*gfSP & 0x000000FF));
	gfSP--;
}


void op_PRINT(void)  // ****************************************************
//send the TOS number as ASCII text via serial terminal port (USB and/or USART)
//use system number base
//numbers in the range 0x20000000 to 0x50060C00 always are printed hex formatted
// *************************************************************************
{
    if ((gfTempNumBase == 16)||((*gfSP >= 0x20000000)&&(*gfSP < 0x50060C00))) {
        send0X32(*gfSP);
    }
    else sendNS32(*gfSP);
    sendTB(SPACE);
	gfSP--;
}


void op_PRINTHEX(void)  // *************************************************
//print the TOS number as hex style ASCII text via via serial terminal port
// *************************************************************************
{
	send0X32(*gfSP);
	sendTB(SPACE);
	gfSP--;
}


void op_PRINTDEC(void)  // *************************************************
//print the TOS number as hex style ASCII text via via serial terminal port
// *************************************************************************
{
	sendS32(*gfSP);;
	gfSP--;
}


void op_PRINTSTR(void)   // *********************************************
//compile the subsequent string to be sent at runtime via via serial terminal
//Note: between P" and the first character of the string MUST be a SPACE
//Non-printable characters (plus " and \) are entered as \+HexCode (w.o. 0x)
//examples: \5C sends Backslash, \20 sends SPACE, \22 sends '"', \00 sends 0
// *************************************************************************
{
	subPrintString(1);
}


void subPrintString(uint32_t out)   // *************************************
//out=1: send to Terminal, out=0: send to IO
//else remark see op_PRINTSTRING()
// *************************************************************************
{
uint32_t k, n, hibyte, lobyte, word;

	gfTokenStart = gfTibPos + 1;
	scanString();

	k= 0;
	n = 0;
	while (1)  {
		if (gfTokenstr[k] == 0x5C) n++; 	//Backslash '\'
						//each Backslash "transports" 3 bytes in Tokenstr[]
		if (gfTokenstr[k] == '"')  {
			gfTokenstr[k] = 0;
			break;
		}
		k++;
		if (k >= 80) {
			gfErrors |= ERR_SYNTAX;
			return;
		}
	}
	if (gfCompiState) {							//compiling
#ifdef MINI
		compileTCode(IDX_DOSTR);
#else
		if (out == 1)  compileTCode(IDX_DOSTR);
		else  compileTCode(IDX_DOIOSTR);
#endif		
		gfTCode[gfCodeIndex]= k - (n * 2);	//string length count
		gfCodeIndex ++;
		k = 0;
		n = 0;
		word = 0;
		do {
			if (gfTokenstr[k] != 0x5C)  {	//not Backslash '\'
				word |= gfTokenstr[k] << n;
				n += 8;
				if (n > 24)  {
					gfTCode[gfCodeIndex] = word;
					gfCodeIndex ++;	//emulates"compileTCode(tempstr[k])"
					n = 0;
					word = 0;
				}
				k++;
			}
			else  {
				lobyte = packBackslash(k);
				if (lobyte == 0xFFFFFFFF)  {
					gfErrors |= ERR_SYNTAX;
					return;
				}
				k += 3;
				word |= lobyte << n;
				n += 8;
				if (n > 24)  {
					gfTCode[gfCodeIndex] = word;
					gfCodeIndex ++;	//emulates"compileTCode(tempstr[k])"
					n = 0;
					word = 0;
				}
			}
		} while (gfTokenstr[k] != 0);
		if (n)  {
			gfTCode[gfCodeIndex] = word;	//string termination
			gfCodeIndex ++;
		}
	}
	else  {									//interpreting
		k = 0;
		do {
			if (gfTokenstr[k] != 0x5C)  {      //'\'
#ifdef MINI
				sendTB(gfTokenstr[k]);
#else
				if (out == 1) sendTB(gfTokenstr[k]);
				else sendIOB(gfTokenstr[k]);
#endif				
				k++;
			}
			else  {
				lobyte = packBackslash(k);
				if (lobyte == 0xFFFFFFFF)  {
					gfErrors |= ERR_SYNTAX;
					return;
				}
				k += 3;
#ifdef MINI
				sendTB(lobyte);
#else
				if (out == 1) sendTB(lobyte);	//bugfix 04Dec21
				else sendIOB(lobyte);
#endif
			}
		} while (gfTokenstr[k] != 0);
	}
}


uint32_t scanString(void)   // ********************************************
// ************************************************************************
{
uint32_t k;

    k=0;
	while (gfTib[gfTokenStart+k] != 0x22)  {  //SPACE){
                                    //search trailing " after scanned token
		gfTokenstr[k] = gfTib[gfTokenStart+k];
		k++;
		if (k >= 78) break;  //limit length of string
	}
	gfTokenstr[k] = 0x22;	//Apostroph
	k++;
	gfTokenstr[k] = 0;
	gfTibPos = gfTokenStart + k;
	return 1;
}


uint32_t packBackslash(uint32_t k)   // ***********************************
// ************************************************************************
{
uint32_t hibyte, lobyte;
	k++;
	hibyte = gfTokenstr[k];
	if ((hibyte >= 'a') && (hibyte <= 'z')) hibyte -= 0x20;
	if (hibyte < 0x30)  return 0xFFFFFFFF;
	hibyte -= 0x30;
	if (hibyte > 9) hibyte -= 7;
	if (hibyte > 0x0F)  return 0xFFFFFFFF;
	k++;
	lobyte = gfTokenstr[k];
//	k++;
	if ((lobyte >= 'a') && (lobyte <= 'z')) lobyte -= 0x20;
	if (lobyte < 0x30)  return 0xFFFFFFFF;
	lobyte -= 0x30;
	if (lobyte > 9) lobyte -= 7;
	if (lobyte > 0x0F)  return 0xFFFFFFFF;
	lobyte += hibyte * 16;
	return lobyte;
}


void op_DECIM(void)  // ****************************************************
//( -- )
//set the system number base to DECIMAL
//CForth internal operations use gTempNumBase, which is usually=gNumBase
//if op_Decim is called during compilation, only gTempNumBase is changed
//op_Semis() or the error handler restores gTempNumBase to gNumBase 
//during compilation number base may temporarily change without global change
// *************************************************************************
{
	gfTempNumBase = 10;
	if (gfCompiState == 0) gfNumBase = 10;
}


void op_HEX(void)  // *****************************************************
//( -- )
//set the system number base to HEX
//comment see op_DECIM
// *************************************************************************
{
	gfTempNumBase = 16;
	if (gfCompiState == 0) gfNumBase = 16;
}

#ifndef MINI
// serial I/O ops  #########################################################

void op_RXQ(void)  // ******************************************************
//check if a character is received via serial I/O (contained in the RX buffer)
//if NOT return 0 on TOS
//if YES return(number of bytes in RX buffer) on TOS
//for MIDI I/O: number of MIDI Messages (except SysEX) is returned
//SPLIT assignment:
//'0'=USB and USART both terminal,
//'1'=USB is terminal, USART is COM I/O port
//'2'=USART is terminal, USB is virtual COM I/O port
//'3'=USB is terminal, USART is MIDI interface
//'4'=USART is terminal, USB is MIDI interface
// *************************************************************************
{
   if (gSplit == 1)  {
        gfSP++;
        if (gUSART_RXTAIL==gUSART_RXHEAD) *gfSP=0;
        else  {
            if (gUSART_RXTAIL < gUSART_RXHEAD)
                *gfSP = (uint32_t)(gUSART_RXHEAD - gUSART_RXTAIL);
            else  *gfSP =RXBUFLEN+(uint32_t)(gUSART_RXTAIL-gUSART_RXHEAD);
        }
    }
    else if ((gSplit == 2) && guConfigured)  {
        gfSP++;
        if (gUSB_RXTAIL==gUSB_RXHEAD)  *gfSP=0;
        else  {
            if (gUSB_RXTAIL < gUSB_RXHEAD)
                *gfSP = (uint32_t)(gUSB_RXHEAD - gUSB_RXTAIL);
            else  *gfSP = RXBUFLEN + (uint32_t)(gUSB_RXTAIL - gUSB_RXHEAD);
        }
    }
    else if ((gSplit == 3) || ((gSplit == 4) && guConfigured)) {
		Rx2Midi(); //check if something in the low-level queue and pack->Midi
        Rx2Midi(); //try to pack a complete message
		Rx2Midi();
        gfSP++;
        if (gMIDI_TAIL==gMIDI_HEAD)  *gfSP=0;
        else  {
            if (gMIDI_HEAD > gMIDI_TAIL)  *gfSP = gMIDI_HEAD - gMIDI_TAIL;
            else  *gfSP = MIDI_BUFLEN - (gUSB_RXTAIL - gUSB_RXHEAD);
        }
    }
    else  gfErrors |= ERR_COMMACC;
}


void op_RX(void)  // ******************************************************
//wait until a character is received via serial I/O and return it on the stack
//for MIDI I/O: MIDI Messages (except SysEX) are returned as special words
//SPLIT assignment:
//'0'=USB and USART both terminal,
//'1'=USB is terminal, USART is COM I/O port
//'2'=USART is terminal, USB is virtual COM I/O port
//'3'=USB is terminal, USART is MIDI interface
//'4'=USART is terminal, USB is MIDI interface
// *************************************************************************
{
uint32_t recvword;

	if (gSplit == 1)  {
		while ((recvword = pullUsartRxBuf()) == 0)  {
									//watchdog is fed in the background
			if (checkBreak()) return;
		}
		gfSP++;
		*gfSP = recvword & 0x000000FF;
	}
	else if ((gSplit == 2) && guConfigured)  {
		while ((recvword = pullUsbRxBuf()) == 0)  {
									//watchdog is fed in the background
			if (checkBreak()) return;
		}
		gfSP++;
		*gfSP = recvword & 0x000000FF;
	}
	else if ((gSplit == 3) || ((gSplit == 4) && guConfigured)) {
		while ((recvword = pullMidiBuf()) == 0)  {
		//pullMidiBuf responds with != 0 when a complete MIDI msg is received
		//servSystem() is processed in the background
			Rx2Midi(); //check if something in the lowlevel queue and packMidi
			if (checkBreak()) return;
		}
		gfSP++;
		*gfSP = recvword & 0x00FFFFFF;
	}
	else  gfErrors |= ERR_COMMACC;
}


void Rx2Midi(void)  // *****************************************************
//transfer bytes from serial RX buffer to MidiScanner()
// *************************************************************************
{
uint32_t recvword;

	if (gSplit == 3)  {
		recvword = pullUsartRxBuf();  //USART is MIDI IN
		if (recvword)  MidiScanner(recvword & 0x000000FF);
	}
	if ((gSplit == 4) && (guConfigured)) {
		recvword = pullUsbRxBuf();  //USB is MIDI IN
		if (recvword)  MidiScanner(recvword & 0x000000FF);
	}
}


void op_TX(void)  // *******************************************************
//send the lower 8 bits of TOS as BYTE!! via serial I/O
// *************************************************************************
{
    sendIOB((uint8_t)(*gfSP & 0x000000FF));
	gfSP--;
}


void op_NTX(void)  // ************************************************
//send TOS times the lower 8 bits of next stack entries as BYTE!! via serial I/O
//example: 0x90 0x33 0x40 3 NTX sends a MIDI note
// *************************************************************************
{
uint32_t k,n;

    k = *gfSP;
    n = k;
    gfSP--;
    while (n)  {
        sendIOB((uint8_t)(*(gfSP - n + 1) & 0x000000FF));
        n--;
    }
    gfSP -= k;
}


void op_TXSTR(void)   // ***************************************************
//compile the subsequent string to be sent at runtime via serial I/O
//else see remark at op_PRINTSTRING()
// *************************************************************************
{
    subPrintString(0);
}
#endif //no MINI but RS-232

//time ops  ################################################################

void op_NOP(void)  // *****************************************************
// ( -- )
// does nothing, except consume a little bit of runtime
// *************************************************************************
{
	IWDG->KR = 0x0000AAAA;
//feed watchdog is necessary formally, else GNU tools ignore this emtpy code
}


void op_XNOP(void)  // *****************************************************
// ( delay -- )
// start a very short blocking delay
// duration is modified by stack parameter
// example for a minimal loop
// : TEST 0xA0 DO DUP OH 0 XNOP DUP OL 100 XNOP AGAIN;
// the HIGH period takes about 12us,the LOW period takes about 50us(48MHz clock)
// the HIGH period takes about 5us,the LOW period takes about 22us(144MHz clock)
// *************************************************************************
{
uint32_t loops;

    loops = *gfSP;
	gfSP--;
	while (loops) {
        IWDG->KR = 0x0000AAAA;  //feed watchdog, timeout ca. 0.5 seconds
        loops--;
	}
}


void op_MS(void)  // *******************************************************
// ( milliSeconds -- )
//blocks CForth execution for *gfSP milliSeconds
// example
// 1000 MS   blocks for 1 second
// *************************************************************************
{
    gfMSCOUNT = *gfSP;    		//in milliseconds
	gfSP--;
	while (gfMSCOUNT)  {
		servSystem();
		if (checkBreak()) return;	//feeds watchdog
	}
}


uint32_t checkBreak(void)  // **********************************************
// *************************************************************************
{
uint32_t tibstate;

	serveBack();

	if (gfTermCtrlBC == CTRL_C) {
		gfTermCtrlBC = 0;
		return 1;
	}
	if (gfBackProcess && gfExecUser)  {
//		if (checkTerminal())  {
			tibstate = subQuery();
			if (tibstate == 1)  backProcEval();	//terminal CRR input
			if (tibstate == 2)  {
				gfErrors |= ERR_CANCEL;
			}
			if (gfErrors) return 1;
		}
//	}
	return 0;
}


void serveBack(void)  // ****************************************************
//serve background processes
// **************************************************************************
{
	if (gfTermCtrlBC == CTRL_B)  op_BAKOP();	//set by USB lowlevel RX
#ifndef MINI
	if (gSplit >= 3) Rx2Midi();
#endif
	if  (gfStepSpeed1)  {
		if ((gfUserHeader[4].toCompile == 0xFFFFFFFE) || 
									(gfUserHeader[4].toCompile > 0))  {
			if (gfSTEPCOUNT1 == 0) subMOT1();
		}
	}
	if  (gfStepSpeed2)  {
		if ((gfUserHeader[5].toCompile == 0xFFFFFFFE) || 
									(gfUserHeader[5].toCompile > 0))  {
			if (gfSTEPCOUNT2 == 0) subMOT2();
		}
	}
#ifndef MINI
	if  (gfStepSpeed3)  {
		if ((gfUserHeader[6].toCompile == 0xFFFFFFFE) || 
									(gfUserHeader[6].toCompile > 0))  {
			if (gfSTEPCOUNT3 == 0) subMOT3();
		}
	}
#endif
}


void op_DECIS(void)  // ****************************************************
// ( deciSeconds -- )
//blocks CForth execution for *gfSP deciSeconds
// example
// 10 DCS   blocks for 1 second
// *************************************************************************
{
	*gfSP *= 100;		//deciSeconds = milliSeconds * 100;
	op_MS();
}


void op_TIX(void)  // ******************************************************
// ( time -- )
// start a countdown measured in tenths of seconds.
// countdown is controlled by WAIT (no-blocking) or BLOCK (blocking)
// *************************************************************************
{
    gfTIXCOUNT = *gfSP *100;       // internally scaled in milliseconds
	gfSP--;
}


void op_WAIT(void)  // *****************************************************
// ( -- time)
//returns remaing time in tenths of seconds 
//while countdown by "TIX" is not finished
//else returns 0 after countdown by "TIX" has finished
// *************************************************************************
{
    IWDG->KR = 0x0000AAAA;          //feed watchdog, timeout ca. 1-2 seconds
	gfSP++;
    *gfSP = gfTIXCOUNT/100;
}


void op_FREQ12(void)  // ****************************************************
// ( freq -- )
//start peripheral TIMER3 to work on PA6 and/or PA7 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfFreq12 = *gfSP;
    gfSP --;
	subFREQ12();
	subPWM1();
	subPWM2();
}


void subFREQ12(void)  // ****************************************************
// ( freq -- )
//start peripheral TIMER3 to work on PA6 and/or PA7 as output
// *************************************************************************
{
	if (gfFreq12 > 0x0000FFFF)  gfFreq12 = 0x0000FFFF;
	if (gfFreq12)  {
		if ((TIM3->CR1 & TIM_CR1_CEN) == 0)  {	//not yet initialized
			TIM3->SR = 0;
			TIM3->CR1 = 0;
			RCC->APB1ENR1 |= RCC_APB1ENR1_TIM3EN;
			TIM3->CCMR1 = (6<<12) | (6<<4);
								//Channel1 and 2,each PWM mode1,compare output
			TIM3->ARR = 2000;                   //about 11 bit resolution
												//but easier calculation
												//freq=24 gives 1kHz pulse
			TIM3->CR1 |= TIM_CR1_CEN;
		}
//initial or restart of timer:
#ifdef CPU_CLOCK144
/*
if the APB prescaler equals 1, the timer clock frequencies 
are set to the same frequency as that of the APB domain.
Otherwise the are set to twice(2x) the frequency of the APB domain
*/
	TIM3->PSC = (gfFreq12 * 3) - 1;
#else
	TIM3->PSC = gfFreq12 - 1;
#endif	
			//PSC is 16 bit register --> Timer3 freq about 24kHz down to 0.3Hz
			//1 is added in TIM3_PSC prescaler divisor
		if (gfPWM1 < 0) TIM3->CCER &= ~TIM_CCER_CC1E;//disable output handling
		else TIM3->CCER |= TIM_CCER_CC1E;			//enable output handling
		if (gfPWM2 < 0) TIM3->CCER &= ~TIM_CCER_CC2E;//disable output handling
		else TIM3->CCER |= TIM_CCER_CC2E;			//enable output	handling
	}
	else{	//FREQ = 0 -->OFF
		TIM3->CR1 &= ~TIM_CR1_CEN;				//disable timer
		gfPWM1 = -1;
		GPIOA->MODER &= ~(GPIO_MODER_MODE6);	//Set PA6 as input(default)
		GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD6;
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD6_1;		//pull-down
		TIM3->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		gfPWM2 = -1;
		GPIOA->MODER &= ~(GPIO_MODER_MODE7);	//Set PA7 as input(default)
		GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD7;
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD7_1;		//pull-down
		TIM3->CCER &= ~TIM_CCER_CC2E;			//disable output handling
	}
}


void op_PWM1(void)  // ******************************************************
// ( pwm -- )
//start peripheral TIMER3, channel 1 to work on PA6 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfPWM1 = (int32_t)*gfSP;
	gfSP --;

	if ((TIM3->CR1 & TIM_CR1_CEN) == 0)  {		//timer emergency activation
		gfFreq12 = 24;
        subFREQ12();
	}
	subPWM1();	
}


void subPWM1(void)  // *****************************************************
// ( pwm -- )
//start peripheral TIMER3, channel 1 to work on PA6 as output
// *************************************************************************
{
	GPIOA->MODER &= ~(GPIO_MODER_MODE6);	//temporarily as input
	GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD6;		//pull-up/down off,wasts current
//handle the different I/O cases:
	if (gfPWM1 < 0)  {
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD6_1;		//pull-down
		TIM3->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	if (gfPWM1 == 0)  {
		GPIOA->MODER |= GPIO_MODER_MODE6_0 ;	// Set PortA6 as output
		GPIOA->BSRR = GPIO_BSRR_BR_6;			// permanent LOW
		TIM3->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	if (gfPWM1 >= 2000)  {
		gfPWM1 = 2000;
		GPIOA->MODER |= GPIO_MODER_MODE6_0 ;	// Set PortA6 as output
		GPIOA->BSRR = GPIO_BSRR_BS_6;			// permanent HIGH
		TIM3->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	//set PA6 in alternate function mode
	GPIOA->MODER |= GPIO_MODER_MODE6_1	;		// Set PortA6 AF mode
	GPIOA->AFR[0] &= 0xF0FFFFFF;				//TIM3,CH1 is AF2
	GPIOA->AFR[0] |= 0x02000000;
	TIM3->CCER |= TIM_CCER_CC1E;				//enable output handling
	TIM3->CCR1 = (uint32_t)gfPWM1;				//H period ca.11bit resolution
}


void op_PWM2(void)  // ******************************************************
// ( pwm -- )
//start peripheral TIMER3, channel 2 to work on PA7 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfPWM2 = (int32_t)*gfSP;
	gfSP --;

	if ((TIM3->CR1 & TIM_CR1_CEN) == 0)  {	//timer	emergency activation
		gfFreq12 = 24;
        subFREQ12();
	}
	subPWM2();
}

void subPWM2(void)  // *****************************************************
// ( pwm -- )
//start peripheral TIMER3, channel 2 to work on PA7 as output
// *************************************************************************
{
	GPIOA->MODER &= ~(GPIO_MODER_MODE7);	//temporarily as input
	GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD7;		//pull-up/down off,wasts current
//handle the different I/O cases:
	if (gfPWM2 < 0)  {
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD7_1;		//pull-down
		TIM3->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	if (gfPWM2 == 0)  {
		GPIOA->MODER |= GPIO_MODER_MODE7_0 ;	// Set PortA7 as output
		GPIOA->BSRR = GPIO_BSRR_BR_7;			// permanent LOW
		TIM3->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	if (gfPWM2 >= 2000)  {
		gfPWM2 = 2000;
		GPIOA->MODER |= GPIO_MODER_MODE7_0 ;	// Set PortA7 as output
		GPIOA->BSRR = GPIO_BSRR_BS_7;			// permanent HIGH
		TIM3->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	//set PA7 in alternate function mode
	GPIOA->MODER |= GPIO_MODER_MODE7_1;			// Set PortA7 in AF mode
	GPIOA->AFR[0] &= 0x0FFFFFFF;				//TIM3,CH2 is AF2
	GPIOA->AFR[0] |= 0x20000000;
	TIM3->CCER |= TIM_CCER_CC2E;				//enable output	handling
	TIM3->CCR2 = (uint32_t)gfPWM2;				//H period ca.11bit resolution
}


void op_FREQ34(void)  // ****************************************************
// ( freq -- )
//start peripheral TIMER2 to work on PB3 and/or PA5 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfFreq34 = *gfSP;
    gfSP --;
	subFREQ34();
	subPWM3();
	subPWM4();
}


void subFREQ34(void)  // ****************************************************
// ( freq -- )
//start peripheral TIMER2 to work on PB3 and/or PA5 as output
// *************************************************************************
{
	if (gfFreq34 > 0x0000FFFF)  gfFreq34 = 0x0000FFFF;
	if (gfFreq34)  {
		if ((TIM2->CR1 & TIM_CR1_CEN) == 0)  {	//not yet initialized
			TIM2->SR = 0;
			TIM2->CR1 = 0;
			RCC->APB1ENR1 |= RCC_APB1ENR1_TIM2EN;
			TIM2->CCMR1 =  (6<<12) | (6<<4);
								//Channel1 and 2,each PWM mode1,compare output
			TIM2->ARR = 2000;			//about 11 bit resolution
										//but easier calculation
										//freq3=24 gives 1kHz pulse
			TIM2->CR1 |= TIM_CR1_CEN;
		}
//initial or restart of timer:
#ifdef CPU_CLOCK144
	TIM2->PSC = (gfFreq34 * 3) - 1;
#else
	TIM2->PSC = gfFreq34 - 1;
#endif	
			//PSC is 16 bit register --> Timer2 freq about 24kHz down to 0.3Hz
			//1 is added in TIM3_PSC prescaler divisor
		if (gfPWM3 < 0) TIM3->CCER &= ~TIM_CCER_CC2E;//disable output handling
		else TIM2->CCER |= TIM_CCER_CC2E;			//enable output	handling
		if (gfPWM4 < 0) TIM2->CCER &= ~TIM_CCER_CC1E;//disable output handling
		else TIM2->CCER |= TIM_CCER_CC1E;			//enable output handling
	}
	else{	//gfFreq34 = 0 -->OFF
		TIM2->CR1 &= ~TIM_CR1_CEN;				//disable timer
		gfPWM3 = -1;
		GPIOB->MODER &= ~(GPIO_MODER_MODE3);	//Set PB3 as input(default)
		GPIOB->PUPDR &= ~GPIO_PUPDR_PUPD3;
		GPIOB->PUPDR |= GPIO_PUPDR_PUPD3_1;		//pull-down
		TIM2->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		gfPWM4 = -1;
		GPIOA->MODER &= ~(GPIO_MODER_MODE5);	//Set PA5 as input(default)
		GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD5;
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD5_1;	//pull-down
		TIM2->CCER &= ~TIM_CCER_CC1E;			//disable output handling
	}
}

void op_PWM3(void)  // ******************************************************
// ( pwm -- )
//start peripheral TIMER2, channel 2 to work on PB3 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfPWM3 = (int32_t)*gfSP;
	gfSP --;

	if ((TIM2->CR1 & TIM_CR1_CEN) == 0)  {		//timer emergency activation
		gfFreq34 = 24;
        subFREQ34();
	}
	subPWM3();	
}


void subPWM3(void)  // *****************************************************
// ( pwm -- )
//start peripheral TIMER2, channel 2 to work on PB3 as output
// *************************************************************************
{
	GPIOB->MODER &= ~(GPIO_MODER_MODE3);	//temporarily as input
	GPIOB->PUPDR &= ~GPIO_PUPDR_PUPD3;		//pull-up/down off,wasts current
//handle the different I/O cases:
	if (gfPWM3 < 0)  {
		GPIOB->PUPDR |= GPIO_PUPDR_PUPD3_1;	//pull-down
	TIM2->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	if (gfPWM3 == 0)  {
		GPIOB->MODER |= GPIO_MODER_MODE3_0;		// Set PortB3 as output
		GPIOB->BSRR = GPIO_BSRR_BR_3;			// permanent LOW
		TIM2->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	if (gfPWM3 >= 2000)  {
		gfPWM3 = 2000;
		GPIOB->MODER |= GPIO_MODER_MODE3_0;		// Set PortB3 as output
		GPIOB->BSRR = GPIO_BSRR_BS_3;			// permanent HIGH
		TIM2->CCER &= ~TIM_CCER_CC2E;			//disable output handling
		return;
	}
	//set PB3 in alternate function mode
	GPIOB->MODER |= GPIO_MODER_MODE3_1			;// Set PortB3 in AF mode
	GPIOB->AFR[0] &= 0xFFFF0FFF;				//TIM2,CH2 is AF1
	GPIOB->AFR[0] |= 0x00001000;
	TIM2->CCER |= TIM_CCER_CC2E;				//enable output handling
	TIM2->CCR2 = (uint32_t)gfPWM3;				//H period ca.11bit resolution
}


void op_PWM4(void)  // ******************************************************
// ( pwm -- )
//start peripheral TIMER2, channel 1 to work on PA5 as output
// *************************************************************************
{
//	if (gfBackNumValid)  {
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfPWM4 = (int32_t)*gfSP;
	gfSP --;

	if ((TIM2->CR1 & TIM_CR1_CEN) == 0)  {		//timer emergency activation
		gfFreq34 = 24;
        subFREQ34();
	}
	subPWM4();	
}


void subPWM4(void)  // *****************************************************
// ( pwm -- )
//start peripheral TIMER2, channel 1 to work on PA5 as output
// *************************************************************************
{
	GPIOA->MODER &= ~(GPIO_MODER_MODE5);	//temporarily as input
	GPIOA->PUPDR &= ~GPIO_PUPDR_PUPD5;		//pull-up/down off,wasts current
//handle the different I/O cases:
	if (gfPWM4 < 0)  {
		GPIOA->PUPDR |= GPIO_PUPDR_PUPD5_1;	//pull-down
		TIM2->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	if (gfPWM4 == 0)  {
		GPIOA->MODER |= GPIO_MODER_MODE5_0 ;	// Set PortA5 as output
												//OTYPER as already config.
		GPIOA->BSRR = GPIO_BSRR_BR_5;			// permanent LOW
		TIM2->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	if (gfPWM4 >= 2000)  {
		gfPWM4 = 2000;
		GPIOA->MODER |= GPIO_MODER_MODE5_0 ;	// Set PortA5 as output
												//OTYPER as already config.
		GPIOA->BSRR = GPIO_BSRR_BS_5;			// permanent HIGH
		TIM2->CCER &= ~TIM_CCER_CC1E;			//disable output handling
		return;
	}
	//set PA5 in alternate function mode
	GPIOA->MODER |= GPIO_MODER_MODE5_1;			// Set PortA5 in AF mode
	GPIOA->AFR[0] &= 0xFF0FFFFF;				//TIM2,CH1 is AF1
	GPIOA->AFR[0] |= 0x00100000;
	TIM2->CCER |= TIM_CCER_CC1E;				//enable output handling
	TIM2->CCR1 = (uint32_t)gfPWM4;				//H period ca.11bit resolution
}


#ifdef MINI
  #ifdef MICRO
void op_MOT1(void)  // ***************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP1
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PA13 and PA4 (PA1 not wired out on proposed PCB)
//control of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SsyTick)
//gfUserHeader[4].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP1= 0:rotation stops, else number of steps to be performed
//NSTP1=-1:NORMAL (commutated) MOTOR drive  (steady H Bridge, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM1 is started with value of TOS
// else rotation counter clockwise,PWM1(PA6) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}

	if (gfUserHeader[4].toCompile == 0xFFFFFFFF)  { //signed int -1
		gfStepSpeed1 = 0;					//neutralize background process
		GPIOA->MODER &= ~(GPIO_MODER_MODE13 | GPIO_MODER_MODE14);
		GPIOA->MODER |= (GPIO_MODER_MODE13_0 | GPIO_MODER_MODE14_0);	//output
		if ((int32_t)*gfSP > 0)  {			//drive normal motor clockwise
			GPIOA->BSRR = GPIO_BSRR_BR_13;
			GPIOA->BSRR = GPIO_BSRR_BS_14;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOA->BSRR =  GPIO_BSRR_BS_13;
			GPIOA->BSRR =  GPIO_BSRR_BR_14;
		}
		else  {	//==0							//includes set PWM1->0
			GPIOA->BSRR =  GPIO_BSRR_BR_13;		//stop motor
			GPIOA->BSRR =  GPIO_BSRR_BR_14;
		}
		op_PWM1();	//start PWM @PA6 with value of TOS 
					//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed1 == 0)  {			//old value
		gfStepPhase1 = 0; 
		gfSTEPCOUNT1 = 0;				//restart new phase
//override user GPIO setup
		GPIOA->MODER &= ~(GPIO_MODER_MODE13 | GPIO_MODER_MODE14);
		GPIOA->MODER |= GPIO_MODER_MODE13_0 | GPIO_MODER_MODE14_0; //output
	}
	gfStepSpeed1 = *gfSP;
	gfSP--;
	if (gfStepSpeed1 == 0)  {			//new value
		GPIOA->BSRR = GPIO_BSRR_BR_13;	//switch motor power off externally!
		GPIOA->BSRR = GPIO_BSRR_BR_14;
		gfUserHeader[4].toCompile = 0;
	}
	gfSTEPCOUNT1 = 0;					//restart new phase, trigger subMOT1
	subMOT1();
}


void subMOT1(void)  // *****************************************************
//handles next stepper step in background process
//comment see above
// *************************************************************************
{
int32_t direction, nsteps;

//drive stepper motor
	nsteps = gfUserHeader[4].toCompile;			//user access:NSTP1
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed1;
	if (direction >= 0)  {	//run clockwise
		gfSTEPCOUNT1 = gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_13;
			GPIOA->BSRR = GPIO_BSRR_BR_14;		//repetition essential
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_14;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_13;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_14;
			break;
		}
	}
	else  {										//run counter clockwise
		gfSTEPCOUNT1 = -gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_13;		//repetition essential
			GPIOA->BSRR = GPIO_BSRR_BR_14;
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_13;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_14;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_13;
			break;
		}
	}
	gfStepPhase1 ++;
	if (gfStepPhase1 >= 4)  gfStepPhase1 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[4].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}

  #else  //Verobard version of MINI

void op_MOT1(void)  // ***************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP1
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PA0 and PA1
//control of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SsyTick)
//gfUserHeader[4].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP1= 0:rotation stops, else number of steps to be performed
//NSTP = -2-->permanent rotation in stepper mode
//NSTP1= -1:NORMAL (commutated) MOTOR drive  (steady H Bridge, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM1 is started with value of TOS
// else rotation counter clockwise,PWM1(PA6) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}

	if (gfUserHeader[4].toCompile == 0xFFFFFFFF)  { //signed int -1
		gfStepSpeed1 = 0;					//neutralize background process
		GPIOA->MODER &= ~(GPIO_MODER_MODE0 | GPIO_MODER_MODE1);
		GPIOA->MODER |= (GPIO_MODER_MODE0_0 | GPIO_MODER_MODE1_0);	//output
		if ((int32_t)*gfSP > 0)  {			//drive normal motor clockwise
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			GPIOA->BSRR = GPIO_BSRR_BS_1;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOA->BSRR =  GPIO_BSRR_BS_0;
			GPIOA->BSRR =  GPIO_BSRR_BR_1;
		}
		else  {	//==0							//includes set PWM1->0
			GPIOA->BSRR =  GPIO_BSRR_BR_0;		//stop motor
			GPIOA->BSRR =  GPIO_BSRR_BR_1;
		}
		op_PWM1();	//start PWM @PA6 with value of TOS 
					//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed1 == 0)  {			//old value->if not yet active:
		gfStepPhase1 = 0; 
		gfSTEPCOUNT1 = 0;				//restart new phase
//override user GPIO setup
		GPIOA->MODER &= ~(GPIO_MODER_MODE0 | GPIO_MODER_MODE1);
		GPIOA->MODER |= GPIO_MODER_MODE0_0 | GPIO_MODER_MODE1_0; //output
	}
//else simply change speed, don't disturb phases
	gfStepSpeed1 = *gfSP;
	gfSP--;
	if (gfStepSpeed1 == 0)  {			//new value
		GPIOA->BSRR = GPIO_BSRR_BR_0;	//switch motor power off externally!
		GPIOA->BSRR = GPIO_BSRR_BR_1;
		gfUserHeader[4].toCompile = 0;
	}
	gfSTEPCOUNT1 = 0;					//restart new phase, trigger subMOT1
	subMOT1();
}


void subMOT1(void)  // *****************************************************
//handles next stepper step in background process
//comment see above
// *************************************************************************
{
int32_t direction, nsteps;

//drive stepper motor
	nsteps = gfUserHeader[4].toCompile;			//user access:NSTP1
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed1;
	if (direction >= 0)  {	//run clockwise
		gfSTEPCOUNT1 = gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			GPIOA->BSRR = GPIO_BSRR_BR_1;		//repetition essential
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_0;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			break;
		}
	}
	else  {										//run counter clockwise
		gfSTEPCOUNT1 = -gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_0;		//repetition essential
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_0;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			break;
		}
	}
	gfStepPhase1 ++;
	if (gfStepPhase1 >= 4)  gfStepPhase1 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[4].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}


  #endif  //MICRO or else

#else  //not MINI

  #ifdef CASE48
void op_MOT1(void)  // ***************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP1
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PA0 and PA1
//control of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SsyTick)
//gfUserHeader[4].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP1= 0:rotation stops, else number of steps to be performed
//NSTP = -2-->permanent rotation in stepper mode
//NSTP1= -1:NORMAL (commutated) MOTOR drive  (steady H Bridge, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM1 is started with value of TOS
// else rotation counter clockwise,PWM1(PA6) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}

	if (gfUserHeader[4].toCompile == 0xFFFFFFFF)  { //signed int -1
		gfStepSpeed1 = 0;					//neutralize background process
		GPIOA->MODER &= ~(GPIO_MODER_MODE0 | GPIO_MODER_MODE1);
		GPIOA->MODER |= (GPIO_MODER_MODE0_0 | GPIO_MODER_MODE1_0);	//output
		if ((int32_t)*gfSP > 0)  {			//drive normal motor clockwise
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			GPIOA->BSRR = GPIO_BSRR_BS_1;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOA->BSRR =  GPIO_BSRR_BS_0;
			GPIOA->BSRR =  GPIO_BSRR_BR_1;
		}
		else  {	//==0							//includes set PWM1->0
			GPIOA->BSRR =  GPIO_BSRR_BR_0;		//stop motor
			GPIOA->BSRR =  GPIO_BSRR_BR_1;
		}
		op_PWM1();	//start PWM @PA6 with value of TOS 
					//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed1 == 0)  {			//old value->if not yet active:
		gfStepPhase1 = 0; 
		gfSTEPCOUNT1 = 0;				//restart new phase
//override user GPIO setup
		GPIOA->MODER &= ~(GPIO_MODER_MODE0 | GPIO_MODER_MODE1);
		GPIOA->MODER |= GPIO_MODER_MODE0_0 | GPIO_MODER_MODE1_0; //output
	}
//else simply change speed, don't disturb phases
	gfStepSpeed1 = *gfSP;
	gfSP--;
	if (gfStepSpeed1 == 0)  {			//new value
		GPIOA->BSRR = GPIO_BSRR_BR_0;	//switch motor power off externally!
		GPIOA->BSRR = GPIO_BSRR_BR_1;
		gfUserHeader[4].toCompile = 0;
	}
	gfSTEPCOUNT1 = 0;					//restart new phase, trigger subMOT1
	subMOT1();
}


void subMOT1(void)  // *****************************************************
//handles next stepper step in background process
//comment see above
// *************************************************************************
{
int32_t direction, nsteps;

//drive stepper motor
	nsteps = gfUserHeader[4].toCompile;			//user access:NSTP1
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed1;
	if (direction >= 0)  {	//run clockwise
		gfSTEPCOUNT1 = gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			GPIOA->BSRR = GPIO_BSRR_BR_1;		//repetition essential
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_0;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			break;
		}
	}
	else  {										//run counter clockwise
		gfSTEPCOUNT1 = -gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_0;		//repetition essential
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_0;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_0;
			break;
		}
	}
	gfStepPhase1 ++;
	if (gfStepPhase1 >= 4)  gfStepPhase1 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[4].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}

  #else  //STANDARD LQFP32

void op_MOT1(void)  // ***************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP1
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PA1 and PA3
//control of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SsyTick)
//gfUserHeader[4].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP1= 0:rotation stops, else number of steps to be performed
//NSTP = -2-->permanent rotation in stepper mode
//NSTP1= -1:NORMAL (commutated) MOTOR drive  (steady H Bridge, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM1 is started with value of TOS
// else rotation counter clockwise,PWM1(PA6) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}

	if (gfUserHeader[4].toCompile == 0xFFFFFFFF)  { //signed int -1
		gfStepSpeed1 = 0;					//neutralize background process
		GPIOA->MODER &= ~(GPIO_MODER_MODE1 | GPIO_MODER_MODE3);
		GPIOA->MODER |= (GPIO_MODER_MODE1_0 | GPIO_MODER_MODE3_0);	//output
		if ((int32_t)*gfSP > 0)  {			//drive normal motor clockwise
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			GPIOA->BSRR = GPIO_BSRR_BS_3;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOA->BSRR =  GPIO_BSRR_BS_1;
			GPIOA->BSRR =  GPIO_BSRR_BR_3;
		}
		else  {	//==0							//includes set PWM1->0
			GPIOA->BSRR =  GPIO_BSRR_BR_1;		//stop motor
			GPIOA->BSRR =  GPIO_BSRR_BR_3;
		}
		op_PWM1();	//start PWM @PA6 with value of TOS 
					//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed1 == 0)  {			//old value->if not yet active:
		gfStepPhase1 = 0; 
		gfSTEPCOUNT1 = 0;				//restart new phase
//override user GPIO setup
		GPIOA->MODER &= ~(GPIO_MODER_MODE1 | GPIO_MODER_MODE3);
		GPIOA->MODER |= GPIO_MODER_MODE1_0 | GPIO_MODER_MODE3_0; //output
	}
//else simply change speed, don't disturb phases
	gfStepSpeed1 = *gfSP;
	gfSP--;
	if (gfStepSpeed1 == 0)  {			//new value
		GPIOA->BSRR = GPIO_BSRR_BR_1;	//switch motor power off externally!
		GPIOA->BSRR = GPIO_BSRR_BR_3;
		gfUserHeader[4].toCompile = 0;
	}
	gfSTEPCOUNT1 = 0;					//restart new phase, trigger subMOT1
	subMOT1();
}


void subMOT1(void)  // *****************************************************
//handles next stepper step in background process
//comment see above
// *************************************************************************
{
int32_t direction, nsteps;

//drive stepper motor
	nsteps = gfUserHeader[4].toCompile;			//user access:NSTP1
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed1;
	if (direction >= 0)  {	//run clockwise
		gfSTEPCOUNT1 = gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			GPIOA->BSRR = GPIO_BSRR_BR_3;		//repetition essential
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_3;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_3;
			break;
		}
	}
	else  {										//run counter clockwise
		gfSTEPCOUNT1 = -gfStepSpeed1;
		switch (gfStepPhase1)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_1;		//repetition essential
			GPIOA->BSRR = GPIO_BSRR_BR_3;
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_1;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_3;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_1;
			break;
		}
	}
	gfStepPhase1 ++;
	if (gfStepPhase1 >= 4)  gfStepPhase1 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[4].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}

   #endif  //CASE48 or else

#endif  //MINI or else

void op_MOT2(void)  // ***************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP2
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PB7 and PB8
//deactivation of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SysTick)
//gfUserHeader[5].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP2= 0:rotation stops, else number of steps to be performed
//NSTP2= -2-->permanent rotation in stepper mode
//NSTP2= -1:NORMAL (commutated) MOTOR drive  (steady H Bridge, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM2 is started with value of TOS
// else rotation counter clockwise,PWM2(PA7) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	if (gfUserHeader[5].toCompile == 0xFFFFFFFF)  {//signed int -1 DC motor
		gfStepSpeed2 = 0;					//neutralize background process
		GPIOB->MODER &= ~(GPIO_MODER_MODE7 | GPIO_MODER_MODE8);
		GPIOB->MODER |= GPIO_MODER_MODE7_0 | GPIO_MODER_MODE8_0;  //output
		if ((int32_t)*gfSP > 0)  {				//drive normal motor clockwise
			GPIOB->BSRR =  GPIO_BSRR_BR_7;
			GPIOB->BSRR =  GPIO_BSRR_BS_8;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOB->BSRR = GPIO_BSRR_BS_7;
			GPIOB->BSRR = GPIO_BSRR_BR_8;
		}
		else  {	//==0							//includes PWM2->0
			GPIOB->BSRR = GPIO_BSRR_BR_7;		//stop motor
			GPIOB->BSRR = GPIO_BSRR_BR_8;
		}
		op_PWM2();						//start PWM2 @PA7 with value of TOS
										//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed2 == 0)  {			//old value->if not yet active:
		gfStepPhase2 = 0; 
		gfSTEPCOUNT2 = 0;				//restart new phase
//override user GPIO setup
		GPIOB->MODER &= ~(GPIO_MODER_MODE7 | GPIO_MODER_MODE8);
		GPIOB->MODER |= GPIO_MODER_MODE7_0 | GPIO_MODER_MODE8_0; //output
	}
//else simply change speed, don't disturb phases
	gfStepSpeed2 = *gfSP;
	gfSP--;
	if (gfStepSpeed2 == 0)  {			//new value
		GPIOB->BSRR = GPIO_BSRR_BR_7;	//switch motor power off externally!
		GPIOB->BSRR = GPIO_BSRR_BR_8;
		gfUserHeader[5].toCompile = 0;
	}
	gfSTEPCOUNT2 = 0;					//restart new phase, trigger subMOT2
	subMOT2();
}	


void subMOT2(void)  // ***************************************************
// background process of MOT2, comment see above
// *************************************************************************
{

int32_t direction, nsteps;

	nsteps = gfUserHeader[5].toCompile;		//user access: NSTP2
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed2;
	if (direction >= 0)  {					//run clockwise
		gfSTEPCOUNT2 = gfStepSpeed2;
		switch (gfStepPhase2)  {
		case 0:
			GPIOB->BSRR = GPIO_BSRR_BR_7;
			GPIOB->BSRR = GPIO_BSRR_BR_8;	//repetition essential
			break;
		case 1:
			GPIOB->BSRR = GPIO_BSRR_BS_8;
			break;
		case 2:
			GPIOB->BSRR = GPIO_BSRR_BS_7;
			break;
		case 3:
			GPIOB->BSRR = GPIO_BSRR_BR_8;
			break;
		}
	}
	else  {									//run counter clockwise
		gfSTEPCOUNT2 = -gfStepSpeed2;
		switch (gfStepPhase2)  {
		case 0:
			GPIOB->BSRR = GPIO_BSRR_BR_7;	//repetition essential
			GPIOB->BSRR = GPIO_BSRR_BR_8;
			break;
		case 1:
			GPIOB->BSRR = GPIO_BSRR_BS_7;
			break;
		case 2:
			GPIOB->BSRR = GPIO_BSRR_BS_8;
			break;
		case 3:
			GPIOB->BSRR = GPIO_BSRR_BR_7;
			break;
		}
	}
	gfStepPhase2 ++;
	if (gfStepPhase2 == 4)  gfStepPhase2 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[5].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}


#ifndef MINI
void op_MOT3(void)  // *****************************************************
// ( stepSpeed -- )
//starts background process, controlled by "system"user VARiable NSTP3
//2-Phase pulse scheme, must be expanded externally to 4 phases and amplified
//2 drive ports PA15 and PB4
//control of motor power must be handled externally
//stepSpeed=0:stepper off, negative value: rotation counter clockwise
//is counted down with 1ms clock (Arm SysTick)
//gfUserHeader[6].toCompile is the actual number of steps to be performed.
//							gets reset when number of steps were performed
//NSTP3= 0:rotation stops, else number of steps to be performed
//NSTP3= -2-->permanent rotation in stepper mode
//NSTP3= -1:NORMAL (commutated) MOTOR drive  (steady H Bridge PA15,PB4, no steps)
//then: 
// if(stepSpeed>0) rotation clockwise, PWM3 is started with value of TOS
// else rotation counter clockwise,PWM3(PB3) is started w.abs(value) of TOS
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	if (gfUserHeader[6].toCompile == 0xFFFFFFFF)  {	 //=signed int -1
		gfStepSpeed3 = 0;					//neutralize background process
		GPIOA->MODER &= ~(GPIO_MODER_MODE15);
		GPIOA->MODER |= GPIO_MODER_MODE15_0;	//output
		GPIOB->MODER &= ~(GPIO_MODER_MODE4);
		GPIOB->MODER |= GPIO_MODER_MODE4_0;	//output
		if ((int32_t)*gfSP > 0)  {			//drive normal motor clockwise
			GPIOA->BSRR = GPIO_BSRR_BR_15;
			GPIOB->BSRR = GPIO_BSRR_BS_4;
		}
		else if ((int32_t)*gfSP < 0)  {			//drive motor counterClockwise
			*gfSP = -(*gfSP);
			GPIOA->BSRR =  GPIO_BSRR_BS_15;
			GPIOB->BSRR =  GPIO_BSRR_BR_4;
		}
		else  {	//==0							//includes PWM2->0
			GPIOA->BSRR =  GPIO_BSRR_BR_15;		//stop motor
			GPIOB->BSRR =  GPIO_BSRR_BR_4;
		}
		op_PWM3();	//start PWM @PB3 with value of TOS 
					//PWM ratio limited to max 2000
		return;
	}

//else stepper motor
	if (gfStepSpeed3 == 0)  {				//old value->if not yet active:
		gfStepPhase3 = 0; 
		gfSTEPCOUNT3 = 0;					//restart new phase
//override user GPIO setup
		GPIOA->MODER &= ~(GPIO_MODER_MODE15);
		GPIOA->MODER |= GPIO_MODER_MODE15_0;
		GPIOB->MODER &= ~(GPIO_MODER_MODE4);
		GPIOB->MODER |= GPIO_MODER_MODE4_0;	//output
	}
//else simply change speed, don't disturb phases
	gfStepSpeed3 = *gfSP;
	gfSP--;
	if (gfStepSpeed3 == 0)  {			//new value
		GPIOA->BSRR = GPIO_BSRR_BR_15  ;//switch motor coils off externally!
		GPIOB->BSRR = GPIO_BSRR_BR_4;
		gfUserHeader[6].toCompile = 0;
	}
	gfSTEPCOUNT3 = 0;				//restart new phase, trigger subMOT2
	subMOT3();

}

void subMOT3(void)  // ***************************************************
// background process of MOT3, comment see above
// *************************************************************************
{
int32_t direction, nsteps;

	nsteps = gfUserHeader[6].toCompile;		//user access: MS2
	if ((nsteps == 0)||(nsteps == 0xFFFFFFFF)) return;//final or DC motor mode

	direction = gfStepSpeed3;
	if (direction >= 0)  {	//run clockwise
		gfSTEPCOUNT3 = gfStepSpeed3;
		switch (gfStepPhase3)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_15;
			GPIOB->BSRR = GPIO_BSRR_BR_4;	//repetition essential
			break;
		case 1:
			GPIOB->BSRR = GPIO_BSRR_BS_4;
			break;
		case 2:
			GPIOA->BSRR = GPIO_BSRR_BS_15;
			break;
		case 3:
			GPIOB->BSRR = GPIO_BSRR_BR_4;
			break;
		}
	}
	else  {									//run counter clockwise
		gfSTEPCOUNT3 = -gfStepSpeed3;
		switch (gfStepPhase3)  {
		case 0:
			GPIOA->BSRR = GPIO_BSRR_BR_15;	//repetition essential
			GPIOB->BSRR = GPIO_BSRR_BR_4;
			break;
		case 1:
			GPIOA->BSRR = GPIO_BSRR_BS_15;
			break;
		case 2:
			GPIOB->BSRR = GPIO_BSRR_BS_4;
			break;
		case 3:
			GPIOA->BSRR = GPIO_BSRR_BR_15;
			break;
		}
	}

	gfStepPhase3 ++;
	if (gfStepPhase3 >= 4)  gfStepPhase3 = 0;
	if (nsteps != 0xFFFFFFFE) gfUserHeader[6].toCompile = nsteps - 1;
				//update countdown , nsteps=0xFFFFFFFE=-2: permanent rotation
}
#endif  //not MINI

// peripheral ops ##########################################################

uint32_t checkValidIO(void)  // ********************************************
// *************************************************************************
{
uint32_t pinNo;

    pinNo = *gfSP;
    gfSP--;

#ifdef CASE48
	if (pinNo > 47)  {	//47=0x2F
		if ((pinNo < 0xA0)||((pinNo > 0xB8)&&(pinNo <0xCD)&&(pinNo >0xCF))) {
			gfErrors |= ERR_NUMVAL;				//only PC13,14,15 implemented
		}
		pinNo &= 0x0000005F;	//PortA<=0x0F,PortB<=0x1F,PortC<=0x4F
		if (pinNo > 0x1F) pinNo -= 0x20;	//range --> 0x2D..0x2F now
	}
#else
	if (pinNo > 24)  {
		if ((pinNo < 0xA0) || (pinNo > 0xB8)) gfErrors |= ERR_NUMVAL;
		pinNo &= 0x0000001F;	//PortB = bit4 set, result is 0...24
	}
#endif

	if (pinNo <= 15)  {				//PortA
#ifdef MINI
	//PA8 is LED, PA11,12 is USB
	//PB6 is DMX OUT, PA15 is DMX BREAK support
		if  ((pinNo == 8)||(pinNo == 11)||(pinNo == 12)||(pinNo == 15))  {
#else
  #ifdef STANDARD
	//PA8 is LED
	//PA9 is Vusb detect, PA10 is USART RX, PA11,12 is USB
			if ((pinNo >= 8) && (pinNo <= 12))  {
  #endif
  #ifdef STANDARD_DMX
	//PA8 is LED
	//PA9 is Vusb detect, PA10 is USART RX, PA11,12 is USB
	//PA0,2 is DMX OUT
		if ((pinNo == 0) || (pinNo == 2) ||
					((pinNo >= 8) && (pinNo <= 12))) {
  #endif
  #ifdef CASE48
	//PB15 is LED
	//PA9 is  USART TX, PA10 is USART RX, PA11,12 is USB
		if ((pinNo >= 9) && (pinNo <= 12)) {
  #endif
#endif	//MINI or not
			gfErrors |= ERR_NUMVAL;
		}
	}
	return pinNo;
}


void checkPWM(uint32_t pinNo)  // ************************************
// *************************************************************************
{
	if (pinNo == 4) DAC->CR &= ~DAC_CR_EN1;		//disable DAC
	if (pinNo == 5) DAC->CR &= ~DAC_CR_EN2;
	if (gfPWM1 != -1)  {
		if (pinNo == 6)  {						//PWM1 @ PA6
			TIM3->CCER &= ~TIM_CCER_CC1E;		//disable PWM1 output
			gfPWM1 = -1;						//PWM OFF
		}
	}	
	if (gfPWM2 != -1)  {
		if (pinNo == 7)  {						//PWM2 @ PA7
			TIM3->CCER &= ~TIM_CCER_CC2E;		//disable PWM2 output
			gfPWM2 = -1;						//PWM OFF
		}
	}
	if (gfPWM4 != -1)  {
		if (pinNo == 5)  {						//PWM4 @ PA5
			TIM2->CCER &= ~TIM_CCER_CC2E;		//disable PWM4 output
			gfPWM4 = -1;						//PWM OFF
		}
	}
}


void op_PINOH(void)  // ****************************************************
// (Port/Pin -- )
// sets GPIO to output HIGH in Push/Pull mode
// parameter: pin number: if Number Base= Hex:A0,A1,A4..A7,AF,B0,B3,B4,.,B8
// else pin numbers must be entered with leading 0x, e.g. 0xA0,0xAA,0xB7
// or as corresponding decimal number, e.g. 0xA0=160, 0xAA=170,0x B0=176
// 		more simple:subtract 160 from decimal value:e.g. PA15=15,PB0=16,PB8=24
// specific pin numbers are reserved for system or not implemented
// *************************************************************************
{
uint32_t pinNo, pos, mask;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	pos = (pinNo & 0x0000000F) << 1;
	mask = ~(3 << pos);

	if (pinNo <= 15)  {						//PortA
		checkPWM(pinNo);
		GPIOA->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOA->MODER &= mask;
		GPIOA->MODER |= 1 << pos;
		GPIOA->BSRR = 1 << pinNo;
		return;
	}
#ifdef CASE48
	else if (pinNo <= 31)  {				//PortB
		pinNo -= 0x10;
  #ifdef DMX
//PB0 is DMX Break switch, PB15 is LED
		if ((pinNo == 0) || (pinNo == 15))   {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #else
// PB15 is LED
		if (pinNo == 15)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOB->MODER &= mask;
		GPIOB->MODER |= 1 << pos;
		GPIOB->BSRR = 1 << pinNo;
		return;
	}
	else  {
		pinNo -= 0x20;
		if (pinNo < 13) {					//PortC only has pins 13,14,15
			gfErrors |= ERR_NUMVAL;
			return;
		}
		GPIOC->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOC->MODER &= mask;
		GPIOC->MODER |= 1 << pos;
		GPIOC->BSRR = 1 << pinNo;
	}
#else  //LQFP32	
	else  {  // Port B
		pinNo -= 0x10;
// PB1,2 not implemented (not checked for better runtime)
#ifdef MINI
		if (pinNo == 6)  {					//PB6 is DMX-TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#else
// PB6 is USART,
		if (pinNo == 6)						{//PB6 is USART TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOB->MODER &= mask;
		GPIOB->MODER |= 1 << pos;
		GPIOB->BSRR = 1 << pinNo;
	}
#endif
}


void op_PINOL(void)  // ****************************************************
// (Port/Pin -- )
// sets GPIO to output LOW in Push/Pull mode
// comment see op_PINOH
// *************************************************************************
{
uint32_t pinNo, pos, mask;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	pos = (pinNo & 0x0000000F) << 1;
	mask = ~(3 << pos);

	if (pinNo <= 15)  {						//PortA
		checkPWM(pinNo);
		GPIOA->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOA->MODER &= mask;
		GPIOA->MODER |= 1 << pos;
		GPIOA->BSRR =  1 << (pinNo + 16);
		return;
	}
#ifdef CASE48
	else if (pinNo <= 31)  {				//PortB
		pinNo -= 0x10;
  #ifdef DMX
//PB0 is DMX Break switch, PB15 is LED
		if ((pinNo == 0) || (pinNo == 15))   {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #else
// PB15 is LED
		if (pinNo == 15)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->PUPDR &= mask;				//pull-up/down off, wasts current		
		GPIOB->MODER &= mask;
		GPIOB->MODER |= 1 << pos;
		GPIOB->BSRR = 1 << (pinNo + 16);
		return;
	}
	else  {
		pinNo -= 0x20;
		if (pinNo < 13) {					//PortC only has pins 13,14,15
			gfErrors |= ERR_NUMVAL;
			return;
		}
		GPIOC->PUPDR &= mask;				//pull-up/down off, wasts current		
		GPIOC->MODER &= mask;
		GPIOC->MODER |= 1 << pos;
		GPIOC->BSRR = 1 << (pinNo + 16);
	}
#else  //LQFP32	
	else  {  // Port B
		pinNo -= 0x10;
// PB1,2 not implemented (not checked for better runtime)
#ifdef MINI
		if (pinNo == 6)  {					//PB6 is DMX-TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#else
// PB6 is USART,
		if (pinNo == 6) {					//PB6 is USART
			gfErrors |= ERR_NUMVAL;
			return;
		}
#endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->PUPDR &= mask;				//pull-up/down off, wasts current		
		GPIOB->MODER &= mask;
		GPIOB->MODER |= 1 << pos;
		GPIOB->BSRR = 1 << (pinNo + 16);
	}
#endif
}


void op_DAO4(void)  // ****************************************************
// (level -- )
// sets DAC1 (PA4) output to <level>
// parameter: 0 ... 0x1FF(4095)
// overrrides previous I/O configuration !
// *************************************************************************
{
    if ((uint32_t)*gfSP > 0xFFF) {  //(uint32_t) excludes negative numbers too
        gfErrors |= ERR_NUMVAL;
        return;
    }
    if ((DAC->CR & DAC_CR_EN1) == 0)  {
        RCC-> AHB2ENR |= RCC_AHB2ENR_DAC1EN;
		DAC->CR &= DAC_CR_EN1 | DAC_CR_EN2;	
										//clear all flags else
		DAC->CR |=  DAC_CR_EN1;
        wait5us();
    }
    DAC->DHR12R1 = *gfSP;
    gfSP--;
}


void op_DAO5(void)  // ****************************************************
// (level -- )
// sets DAC2 (PA5) output to <level>
// parameter: 0 ... 0xFFF(4095)
// overrrides previous I/O configuration !
// *************************************************************************
{
    if ((uint32_t)*gfSP > 0xFFF) {  //(uint32_t) excludes negative numbers too
        gfErrors |= ERR_NUMVAL;
        return;
    }
	if ((DAC->CR & DAC_CR_EN2) == 0)  {
		RCC-> AHB2ENR |= RCC_AHB2ENR_DAC1EN;
		DAC->CR &= DAC_CR_EN1 | DAC_CR_EN2;	
										//clear all flags else
		DAC->CR |=  DAC_CR_EN2;
		wait5us();
	}
	DAC->DHR12R2 = *gfSP;
	gfSP--;
}


void op_PINIPU(void)  // ***************************************************
// (Port/Pin -- )
// configures GPIO as input with internal pull-up resistor
// comment see op_PINOH
// *************************************************************************
{
uint32_t pinNo, pos, mask;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	pos = (pinNo & 0x0000000F) << 1;
	mask = ~(3 << pos);

	if (pinNo <= 15)  {						//PortA
		checkPWM(pinNo);
		GPIOA->MODER &= mask;				//configured as input
		GPIOA->PUPDR &= mask;				//first HiZ
		GPIOA->PUPDR |= 1 << pos;			//now pull-up
		return;
	}
#ifdef CASE48
	else if (pinNo <= 31)  {				//PortB
		pinNo -= 0x10;
  #ifdef DMX
//PB0 is DMX Break switch, PB15 is LED
		if ((pinNo == 0) || (pinNo == 15))   {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #else
// PB15 is LED
		if (pinNo == 15)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->MODER &= mask;
		GPIOB->PUPDR &= mask;				//first HiZ
		GPIOB->PUPDR |= 1 << pos;			//now pull-up
		return;
	}
	else  {
		pinNo -= 0x20;
		if (pinNo < 13) {					//PortC only has pins 13,14,15
			gfErrors |= ERR_NUMVAL;
			return;
		}
		GPIOC->MODER &= mask;
		GPIOC->PUPDR &= mask;				//first HiZ
		GPIOC->PUPDR |= 1 << pos;			//now pull-up
	}
#else  //LQFP32	
	else  {  // Port B
		pinNo -= 0x10;
// PB1,2 not implemented (not checked for better runtime), PB6 is USART,
#ifdef MINI
		if (pinNo == 6)  {					//PB6 is DMX-TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#else
// PB6 is USART,
		if (pinNo == 6) {					//PB6 is USART
			gfErrors |= ERR_NUMVAL;
			return;
		}
#endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->MODER &= mask;
		GPIOB->PUPDR &= mask;				//first HiZ
		GPIOB->PUPDR |= 1 << pos;			//now pull-up
	}
#endif
}


void op_PINIPD(void)  // ***************************************************
// (Port/Pin -- )
// configures GPIO as input with internal pull-down resistor
// comment see op_PINOH
// *************************************************************************
{
uint32_t pinNo, pos, mask;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	pos = (pinNo & 0x0000000F) << 1;
	mask = ~(3 << pos);

	if (pinNo <= 15)  {						//PortA
		checkPWM(pinNo);
		GPIOA->MODER &= mask;				//configured as input
		GPIOA->PUPDR &= mask;				//first HiZ
		GPIOA->PUPDR |= 2 << pos;			//now pull-down4 split
		return;
	}
#ifdef CASE48
	else if (pinNo <= 31)  {				//PortB
		pinNo -= 0x10;
  #ifdef DMX
//PB0 is DMX Break switch, PB15 is LED
		if ((pinNo == 0) || (pinNo == 15))   {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #else
// PB15 is LED
		if (pinNo == 15)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->MODER &= mask;
		GPIOB->PUPDR &= mask;				//first HiZ
		GPIOB->PUPDR |= 2 << pos;			//now pull-down
		return;
	}
	else  {
		pinNo -= 0x20;
		if (pinNo < 13) {					//PortC only has pins 13,14,15
			gfErrors |= ERR_NUMVAL;
			return;
		}
		GPIOC->MODER &= mask;
		GPIOC->PUPDR &= mask;				//first HiZ
		GPIOC->PUPDR |= 2 << pos;			//now pull-down
	}
#else  //LQFP32	
	else  {  // Port B
		pinNo -= 0x10;
// PB1,2 not implemented (not checked for better runtime)
#ifdef MINI
		if (pinNo == 6) {					//PB6 is DMX-TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#else
// PB6 is USART,
		if (pinNo == 6) {					//PB6is USART
			gfErrors |= ERR_NUMVAL;
			return;
		}
#endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->MODER &= mask;
		GPIOB->PUPDR &= mask;				//first HiZ
		GPIOB->PUPDR |= 2 << pos;			//now pull-down
	}
#endif
}


void op_PINIZ(void)  // ****************************************************
// (Port/Pin -- )
// configures GPIO as HI-Z input
// comment see op_PINOH
// *************************************************************************
{
uint32_t pinNo, pos, mask;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	pos = (pinNo & 0x0000000F) << 1;
	mask = ~(3 << pos);

	if (pinNo <= 15)  {						//PortA
		checkPWM(pinNo);
		GPIOA->MODER &= mask;				//configured as input
		GPIOA->PUPDR &= mask;				//first HiZ
		return;
	}
#ifdef CASE48
	else if (pinNo <= 31)  {				//PortB
		pinNo -= 0x10;
  #ifdef DMX
//PB0 is DMX Break switch, PB15 is LED
		if ((pinNo == 0) || (pinNo == 15))   {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #else
// PB15 is LED
		if (pinNo == 15)  {
			gfErrors |= ERR_NUMVAL;
			return;
		}
  #endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
		}
			GPIOB->PUPDR &= mask;				//pull-up/down off, wasts current
			GPIOB->MODER &= mask;
			return;
		}
	}
	else  {
		pinNo -= 0x20;
		if (pinNo < 13) {					//PortC only has pins 13,14,15
			gfErrors |= ERR_NUMVAL;
			return;
		}
		GPIOC->PUPDR &= mask;				//pull-up/down off, wasts current
		GPIOC->MODER &= mask;
	}
#else  //LQFP32	
	else  {  // Port B
		pinNo -= 0x10;
// PB1,2 not implemented (not checked for better runtime), PB6 is USART,
#ifdef MINI
		if (pinNo == 6)  {					//PB6 is DMX-TX
			gfErrors |= ERR_NUMVAL;
			return;
		}
#else
// PB6 is USART,
		if (pinNo == 6) {					//PB6 is USART,
			gfErrors |= ERR_NUMVAL;
			return;
		}
#endif
		if (gfPWM3 != -1)  {
			if (pinNo == 3)  {				//PWM3 @ PB3
				TIM2->CCER &= ~TIM_CCER_CC2E;//disable PWM3 output
				gfPWM3 = -1;				//PWM OFF
			}
		}
		GPIOB->MODER &= mask;				//configured as input
		GPIOB->PUPDR &= mask;				//HiZ
	}
#endif
}


void op_READDIG(void)  // *************************************************
// (Port/Pin -- digitalLevel)
// returns digital level of addresses GPIO
// pin is always read. No check if pin is available for common I/O
// else comment see op_PINOH
// *************************************************************************
{
uint32_t pinNo, level;

	pinNo = checkValidIO();
	if (gfErrors)  return;					//error flag is set

	if (pinNo <= 15) level = GPIOA->IDR & (1 << pinNo);
#ifdef CASE48
	else if (pinNo <= 31) {
		pinNo &= 0x0000000F;		//extract portPin number
		level = GPIOB->IDR & (1 << pinNo);
	}
	else  {
		pinNo &= 0x0000000F;		//extract portPin number
		level = GPIOC->IDR & (1 << pinNo);
		}
#else
	else  {
		pinNo &= 0x0000000F;		//extract portPin number
		level = GPIOB->IDR & (1 << pinNo);
	}
#endif
	gfSP++;
    if (level)  *gfSP = 1;
    else  *gfSP = 0;
}


void op_READADC(void)  // **************************************************
//(port -- analogLevel)
//only PA0..PA7,PB0 is supported(enter parameter 0..7 or 0xA0..0xA7,8 or 0xB0)
//returns analog level of addressed port. Initalizes pin for ADC automatically 
// *************************************************************************
{
uint32_t port, pos, mask, channel, k;
const uint8_t initStr[] = "  (ADC init) ";
const uint8_t offStr[] = "  (ADC off) ";

	port = *gfSP;
	if (port == 0xFFFFFFFF)  {		//i.e. signed -1
		gfSP--;
		disableADC();
		sendString(offStr);
		return;
	}
#ifndef LQFP48_DMX
	if (port == 0xB0)  {
		port = 8;
	}
	if ((port > 8) && ((port < 0xA0) || (port  > 0xA7)))  {
#else
	if ((port > 7) && ((port < 0xA0) || (port  > 0xA7)))  {
#endif
		gfErrors |= ERR_NUMVAL;
		return;
	}

	port &= 0x0000000F;
	pos =  port << 1;
	mask = 3 << (port << 1);

    if ((RCC->AHB2ENR & RCC_AHB2ENR_ADC12EN) == 0)  {
       initADC();
       sendString(initStr);
    }
    ADC1->ISR = 0x000007FF;         //clear relevant all bits by writing 1

    switch (port)  {
#ifndef STANDARD_DMX
    case 0:
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
		channel = 1;
        break;
#endif
    case 1:
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 2;
        break;
#ifndef DMX
    case 2:
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 3;
        break;
#endif
#ifdef MINI		//here DMX is PB6/PA15
    case 2:
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 3;
        break;
#endif
    case 3:
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 4;
        break;
    case 4:
        DAC->CR &= ~DAC_CR_EN1;
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
		channel = 17;
        break;
    case 5:
		if (gfPWM4 != -1)  {
			TIM2->CCER &= ~TIM_CCER_CC1E;		//disable PWM4 output
			gfPWM4 = -1;						//PWM OFF
		}
        DAC->CR &= ~DAC_CR_EN2;
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
		channel = 13;
        break;
    case 6:
		if (gfPWM1 != -1)  {
			TIM3->CCER &= ~TIM_CCER_CC1E;	//disable output handling
			gfPWM1 = -1;
		}
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 3;
        break;
    case 7:
		if (gfPWM2 != -1)  {
			TIM3->CCER &= ~TIM_CCER_CC2E;	//disable output handling
			gfPWM2 = -1;
		}
        GPIOA->MODER |= mask;				//analog input
        GPIOA->PUPDR &= ~mask;				//HiZ
        channel = 4;
        break;
#ifndef LQFP48_DMX
    case 8:		//PB0
        GPIOB->MODER |= 3;					//analog input
        GPIOB->PUPDR &=0xFFFFFFFC;			//HiZ
        channel = 15;
        break;
#endif
	default:
		gfErrors |= ERR_NUMVAL;
		return;
    }

	if ((port < 4) || (port == 8)) {
		ADC1->SQR1 = channel << 6;      //number of scanned channels always=1
		ADC1->CR |= ADC_CR_ADSTART;
		while ((ADC1->ISR & ADC_ISR_EOS) == 0) ;
		*gfSP = (uint32_t)ADC1->DR;
	}
	else  {
		ADC2->SQR1 = channel << 6;      //number of scanned channels always=1
		ADC2->CR |= ADC_CR_ADSTART;;
		while ((ADC2->ISR & ADC_ISR_EOS) == 0) ;
		*gfSP = (uint32_t)ADC2->DR;
	}	
}


void initADC(void)  // ***************************************************
// ( -- )
// initalizes ADC in general
// once initialized, it will be active during the session.
// takes some extra current
// *************************************************************************
{
    if (RCC->AHB2ENR & RCC_AHB2ENR_ADC12EN)  return;  //dont init twice

	RCC->AHB2RSTR &= ~(RCC_AHB2RSTR_ADC12RST);
	RCC->AHB2ENR |= RCC_AHB2ENR_ADC12EN;	//enables ADC1 and ADC2 clock
//keep runtime overhead low, initialize both ADC1, ADC2
//ADC12 is connected with PA0(IN1),PA1(IN2) ->use with ADC1
//ADC1 is connected with PA2(IN3),PA3(IN4),PB0(IN15)
//ADC2 is connected with PA4(IN17),PA5(IN13),PA6(IN3),PA7(IN4)
	ADC1->CR &= ~ADC_CR_DEEPPWD;			//release from Deep Power Down
	ADC1->CR |= ADC_CR_ADVREGEN;			//enable ADC power regulator
	ADC2->CR &= ~ADC_CR_DEEPPWD;			//release from Deep Power Down
	ADC2->CR |= ADC_CR_ADVREGEN;			//enable ADC power regulator
//AVR voltage regulator starting time is regular 20us
    msDelay(2);     //as a precaution
//ADC1 channels   are used
	ADC1->DIFSEL &= 0xFFFF0001;				//force all channels to "single ended"
//ADC2 channels   are used
	ADC2->DIFSEL &= 0xFFFF0001;				//force all channels to "single ended"
    ADC12_COMMON->CCR = 3<<16;				//no special ADC channels active
									//and CKMODE=11:ADC clock source = HLK/4= 12MHz
    wait5us();      //as a precaution

    ADC1->ISR |= ADC_ISR_ADRDY;				//clear by writing 1 !
    ADC1->CR |= ADC_CR_ADEN;
    while ((ADC1->ISR & ADC_ISR_ADRDY) == 0) ;
    ADC1->ISR |= ADC_ISR_ADRDY;
    ADC1->CFGR = 0x80000000;//single conversion,right aligned,12bit resolution
    ADC1->SMPR1 = 0x24924924;	
					//sampling time for each channel:47.5+2.5 ADC clock cycles
													//=about 5us plus overhead
    ADC1->SMPR2 = 0x04924924;	
					//sampling time for each channel:47.5+2.5 ADC clock cycles
													//=about 5us plus overhead

    ADC2->ISR |= ADC_ISR_ADRDY;				//clear by writing 1 !
    ADC2->CR |= ADC_CR_ADEN;
    while ((ADC2->ISR & ADC_ISR_ADRDY) == 0) ;
    ADC2->ISR |= ADC_ISR_ADRDY;
    ADC2->CFGR = 0x80000000;	//single conversion,right aligned,12bit resolution
    ADC2->SMPR1 = 0x24924924;	
					//sampling time for each channel:47.5+2.5 ADC clock cycles
													//=about 5us plus overhead
    ADC2->SMPR2 = 0x04924924;	
					//sampling time for each channel:47.5+2.5 ADC clock cycles
													//=about 5us plus overhead
    ADC1->SQR1 = 0;				//always sample only 1 channel, insert Ch# later
    ADC1->SQR2 = 0;
    ADC1->SQR3 = 0;
    ADC1->SQR4 = 0;
    ADC2->SQR1 = 0;				//always sample only 1 channel, insert Ch# later
    ADC2->SQR2 = 0;
    ADC2->SQR3 = 0;
    ADC2->SQR4 = 0;
    wait5us();
    wait5us();
}

void disableADC(void)  // ***************************************************
// ( -- )
// initalizes ADC in general
// once initialized, it will be active during the session.
// takes some extra current
// *************************************************************************
{
	ADC1->ISR |= ADC_ISR_ADRDY;			//clear by writing 1 
	ADC1->CR |= ADC_CR_ADDIS;			//set ADDIS bit
	ADC2->ISR |= ADC_ISR_ADRDY;			//clear by writing 1 
	ADC2->CR |= ADC_CR_ADDIS;			//set ADDIS bit
	while ((ADC1->CR & ADC_CR_ADEN) != 0);
    RCC->AHB2ENR &= ~(RCC_AHB2ENR_ADC12EN);
	RCC->AHB2RSTR |= RCC_AHB2RSTR_ADC12RST;
}


void op_INITSPI(void)  // ***************************************************
// (Mode(0,1,2,3)  Bits(4..31) Clock -- )  //3 stack parameter !!
// SPI is always Master
// *************************************************************************
{
    gfSpiClock = *gfSP;
    gfSP--;
    gfSpiBits = *gfSP;
    gfSP--;
    gfSpiMode = *gfSP;
    gfSP--;
    subInitSpi();
}


void subInitSpi(void)  // **************************************************
//mode0:data sampled on rising clk edge and shifted out at the falling edge
//		clock idle low
//mode1::data sampled on falling clk edge and shifted out at the rising edge
//		clock idle low
//mode2:data sampled on falling clk edge and shifted out at the rising edge
//		clock idle high
//mode3:data sampled on rising clk edge and shifted out at the falling edge
//		clock idle high
// *************************************************************************
{
	if ((gfSpiMode>3)||(gfSpiBits<4)||(gfSpiBits>32)||(gfSpiClock>0xFFFF)) {
		gfErrors |= ERR_NUMVAL;
		return;
	}

//PB3 is SPI1_SCK   configure as output push/pull
//PB4 is SPI1_MISO  configure as input pull down
//PB5 is SPI1-MOSI, configure as output push/pull
	GPIOB->MODER &= ~(GPIO_MODER_MODE3 | GPIO_MODER_MODE4 | GPIO_MODER_MODE5);
											//all, especially PB4, are input
	GPIOB->MODER|=(GPIO_MODER_MODE3_0|GPIO_MODER_MODE5_0);//PB3/5 push/pull out
	GPIOB->PUPDR &= ~GPIO_PUPDR_PUPD4;                
	GPIOB->PUPDR |= GPIO_PUPDR_PUPD4_1;				//PB4=MISO now pull-down

	if (gfSpiMode >= 2) GPIOB->BSRR = GPIO_BSRR_BS3;//preset idle SCLK High
	else  GPIOB->BSRR = GPIO_BSRR_BR3;				//idle SCLK Low
}


void op_SPI(void)  // ******************************************************
//(sendWord(or Byte) --	recvWord(or	Byte))
//uses default parameters or user parameters if project	is loaded from Flash,
//else these must be initialised before with INITSPI()
// /CS must	be handled completely by the user, how ever
// *************************************************************************
{
uint32_t spiTX,	spiRX;
int32_t	k, loops;

	if ((GPIOB->MODER & 0x00000FC0) != 0x00000440) subInitSpi();
												//i.e.PB3,5=output,PB4=input
				//if this does not fit->emergency init with actual parameters

//as precaution, usually not necessary:
	if (gfSpiMode >= 2) GPIOB->BSRR = GPIO_BSRR_BS_3;	//idle SCLK High
	else  GPIOB->BSRR = GPIO_BSRR_BR_3;					//idle SCLK Low

	spiTX = *gfSP;
	spiRX = 0;
	k = gfSpiBits;
	IWDG->KR = 0x0000AAAA;								//precaution

	do  {
		k--;
		if (spiTX & (1<< k)) GPIOB->BSRR = GPIO_BSRR_BS_5;//MOSI high
		else GPIOB->BSRR = GPIO_BSRR_BR_5;				//MOSI Low

		if ((gfSpiMode == 0) || (gfSpiMode == 3))  {
			GPIOB->BSRR = GPIO_BSRR_BR_3;					//SCLK	Low
		}
		else GPIOB->BSRR = GPIO_BSRR_BS_3;					//SCLK	High

		loops = gfSpiClock;
		while (loops) {
			loops--;
		}

		spiRX = spiRX << 1;
		spiRX |= (GPIOB->IDR >> 4) & 1;						//PB4-->bit0

		if ((gfSpiMode == 0) || (gfSpiMode == 3))  {
			GPIOB->BSRR = GPIO_BSRR_BS_3;					//SCLK	High
		}
		else GPIOB->BSRR = GPIO_BSRR_BR_3;					//SCLK	Low

		loops = gfSpiClock;
		while (loops) {
			loops--;
		}
	}  while (k);

	if (gfSpiMode >= 2) GPIOB->BSRR = GPIO_BSRR_BS_3;	//idle SCLK High
	else  GPIOB->BSRR = GPIO_BSRR_BR_3;					//idle SCLK Low
	* gfSP = spiRX;
}


void op_INITI2C(void)  // **************************************************
// ( SlaveAddress --- )
//If additionally TOS bit8 is set, 400kHz clock is used (add256 to SADD)
//else standard is 100kHz clock.
//enter the slave device address here 
//	(7 bit address from manual, will be shifted <<1 and R/W bit inserted later)
//Is primarily stored internally and later inserted in any I2C RX/TX command.
//Can be changed command by command for multi-slave communication
//only I2C master transmitter/receiver, 7bit address, 100kHz and 400kHz clock
//SMBus is not supported
//if these parameters shall be modified, recall INITI2C when needed 
//else one call before first Tx/Rx operation is sufficient
//PB7 is I2C1/SDA
//PB8 is I2C1/SCL
//these pins are specified by STM as FT,Fm+, i.e. 5Volt tolerant, high drive
//max.safe output:SDA and SCL pull up to +5V:4.7kOhm +10kOhm divider to Ground
//results in ca. 3.4V idle voltage at PB7,8
// *************************************************************************
{
uint32_t param;

	param = *gfSP;
	gfSP--;

	I2C1->CR1 &= ~(I2C_CR1_PE);					//PE bit OFF
	RCC->APB1ENR1 |= RCC_APB1ENR1_I2C1EN;		//activate I2C clock

//without this instruction line PB7,8 work as push-pull out:
	GPIOB->OTYPER |= (GPIO_OTYPER_OT7 | GPIO_OTYPER_OT8);
//set PB7,PB8 into alternate function mode
	GPIOB->MODER &= ~(GPIO_MODER_MODE7 | GPIO_MODER_MODE8);
	GPIOB->MODER |= (GPIO_MODER_MODE7_1 | GPIO_MODER_MODE8_1);
	GPIOB->AFR[0] &= 0x0FFFFFFF;    //I2C1 is AF4
	GPIOB->AFR[0] |= 0x40000000;
	GPIOB->AFR[1] &= 0xFFFFFFF0;
	GPIOB->AFR[1] |= 0x00000004;
//as a precaution for 5V operation, set inputs in HiZ mode
	GPIOB->PUPDR &= ~(GPIO_PUPDR_PUPDR7 | GPIO_PUPDR_PUPDR8);

	RCC->CCIPR &= ~RCC_CCIPR_I2C1SEL;			//connect I2C1 with APB1 bus

	I2C1->CR2 = (param & 0x0000007F)<<1; 		//slave address:SADD[7:1]

//init Timing Register
	if (param & 0x00000100)   {	//400kHz I2C clock, prescaled APB1 freq = 8 MHz
#ifdef CPU_CLOCK144
	I2C1->TIMINGR = 8<<28;		//tPRESC = 72MHz/9
#else
	I2C1->TIMINGR = 5<<28;		//tPRESC = 48MHz/6
#endif
	I2C1->TIMINGR |= ((2<<20) | (4<<16) | (5<<8) | 10);
								//tSCLDEL=(3+1)*tPRESC, tSDADEL=3*tPRESC
								//tSCLH=(5+1)*tPRESC, tSCLL=(9+1)*tPRESC
								//internal delay 500ns assumed(see RefMan p1845)
	}
	else  {
#ifdef CPU_CLOCK144
//100kHz I2C clock, prescaled APB1 freq = 4,5 MHz
	I2C1->TIMINGR = 15<<28;		//tPRESC = 72MHz/16
	I2C1->TIMINGR |= ((1<<20) | (7<<16) | (18<<8) | 23);
								//tSCLDEL=(4+1)*tPRESC, tSDADEL=2*tPRESC
								//tSCLH=(16+1)*tPRESC, tSCLL=(20+1)*tPRESC
								//internal delay 500ns assumed(see RefMan p1845)
#else
//100kHz I2C clock, prescaled APB1 freq = 4 MHz
	I2C1->TIMINGR = 11<<28;		//tPRESC = 48MHz/12
	I2C1->TIMINGR |= ((0<<20) | (6<<16) | (16<<8) | 20);
								//tSCLDEL=(4+1)*tPRESC, tSDADEL=2*tPRESC
								//tSCLH=(16+1)*tPRESC, tSCLL=(20+1)*tPRESC
								//internal delay 500ns assumed(see RefMan p1845)
#endif
	}
	I2C1->ISR = 0;				//interrupt and status register
	I2C1->ICR = 0x738;			//clear all interrupts
	I2C1->CR1 |= I2C_CR1_PE;	//enable I2C1

//	sendX32(GPIOB->IDR);

	if (((GPIOB->IDR & GPIO_IDR_IDR_7) == 0) ||
				((GPIOB->IDR & GPIO_IDR_IDR_8) == 0))  {	//no I2C hardware
		gfErrors |= ERR_COMMACC;
		return;	
	}
}
// #endif

void op_I2CTX1(void)  // ****************************************************
// ( byte -- )  n > 0
//1 byte is sent to the addressed I2C slave
// *************************************************************************
{
uint32_t sendByte;

	sendByte = *gfSP & 0x000000FF;
	gfSP --;

	gTIMEOUT = 200;							//0.2 seconds emergency brake

	while ((I2C1->ISR & (I2C_ISR_TXE | I2C_ISR_RXNE )) == 0);  //as precaution
	I2C1->CR2 &= 0xFC00FBFF;				//clear NBYTES area and RD_WRN bit
	I2C1->CR2 |= (I2C_CR2_AUTOEND | (1<<16));					//enter 1 byte

	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0) 	{
		if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
		//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;		//reset all relevant flags

	I2C1->TXDR = sendByte;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0) 	{
		if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
		//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;		//reset all relevant flags
	I2C1->CR2 |= I2C_CR2_STOP;
}


void op_I2CTXN(void)  // ****************************************************
// ( b1 b2 ... bn n -- )  n > 0
//n bytes are sent to the addressed I2C slave
//the bytes to be sent are handed over via Stack.
//TOS is the number of bytes to be sent
//the bytes are on stack in order to be sent (first byte = deepest in stack)
//
//the meaning of each byte,especially b1,depends on the slave device protocol
//EEPROM for example needs b1= start address followed by (n-1) data bytes
// *************************************************************************
{
uint32_t k, nBytes;

	nBytes = *gfSP;
	gfSP -= (nBytes + 1);
	if (gfSP < (gfDataStack + 3))  {
		gfErrors |= ERR_STACK;
		return;
	}
	if (nBytes == 0) {
		gfErrors |= ERR_NUMVAL;
		return;
	}

	gTIMEOUT = 200;							//0.2 seconds emergency brake

	while ((I2C1->ISR & (I2C_ISR_TXE | I2C_ISR_RXNE )) == 0);  //as precaution
	I2C1->CR2 &= 0xFC00FBFF;				//clear NBYTES area and RD_WRN bit
	I2C1->CR2 |= (I2C_CR2_AUTOEND | (nBytes<<16)); 				//enter NBYTES

	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0) 	{
		if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
		//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;		//reset all relevant flags

	for (k=0; k < nBytes; k++) {
		I2C1->TXDR = *(gfSP + k+1) & 0x000000FF;
		while ((I2C1->ISR & I2C_ISR_TXE) == 0) 	{
			if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
			//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
				gfErrors |= ERR_COMMACC;
				return;
			}
		}
		I2C1->ICR &= 0x000087F8;		//reset all relevant flags
	}
	I2C1->CR2 |= I2C_CR2_STOP;
}


void op_I2CRX1(void)  // ****************************************************
// ( -- byte)
//one bytes is received from the addressed I2C slave
//when used with common EEPROM, this is a "Current Address Read"
// *************************************************************************
{
	gTIMEOUT = 200;							//0.2 seconds emergency brake

	while ((I2C1->ISR & (I2C_ISR_TXE | I2C_ISR_RXNE )) == 0); //as precaution
	I2C1->CR2 &= 0xFC00FFFF;							//clear NBYTES area
	I2C1->CR2 |= (I2C_CR2_AUTOEND | I2C_CR2_RD_WRN);
	I2C1->CR2 |= 1 << 16;								//enter 1 BYTE

	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0)  {
		if (I2C1->ISR & I2C_ISR_NACKF)  {
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;				//reset all relevant flags
	while ((I2C1->ISR & I2C_ISR_RXNE) == 0)  {
		if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
			//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	gfSP++;
	*gfSP =	I2C1->RXDR & 0x000000FF;
	I2C1->ICR &= 0x000087F8;				//reset all relevant flags
}


void op_I2CRXN(void)  // ****************************************************
// (n -- b1 b2 ... bn)
//n bytes are received from the addressed I2C slave
//when used with common EEPROM, this is a "Current Address Read"
// *************************************************************************
{
uint32_t k, nBytes;

	nBytes = *gfSP;
	gfSP --;
	if (nBytes == 0)   {
		gfErrors |= ERR_NUMVAL;
		return;
	}

	gTIMEOUT = 200;							//0.2 seconds emergency brake

	while ((I2C1->ISR & (I2C_ISR_TXE | I2C_ISR_RXNE )) == 0); //as precaution
	I2C1->CR2 &= 0xFC00FFFF;							//clear NBYTES area
	I2C1->CR2 |= (I2C_CR2_AUTOEND | I2C_CR2_RD_WRN);
	I2C1->CR2 |= nBytes << 16;					//enter n bytes to be received

	//start immediately in receive mode
	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0)  {
		if (I2C1->ISR & I2C_ISR_NACKF)  {
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;					//reset all relevant flags
//read data bytes
	for (k=0; k < nBytes; k++) {
		while ((I2C1->ISR & I2C_ISR_RXNE) == 0)  {
			if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
				//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
				gfErrors |= ERR_COMMACC;
				return;
			}
		}
		I2C1->ICR &= 0x000087F8;				//reset all relevant flags
		gfSP++;
		*gfSP =	I2C1->RXDR & 0x000000FF;
	}
}


void op_I2CADRXN(void)  // **************************************************
// ( RegAddr n -- b1 b2 ...bn)
// first one byte with slave register address is sent to slave, then
// n bytes are received from the addressed I2C slave
// *************************************************************************
{
uint32_t k, nBytes;

	if (gfSP < (gfDataStack + 3))  {
		gfErrors |= ERR_STACK;
		return;
	}
	nBytes = *gfSP;
	gfSP --;
	if (nBytes == 0)   {
		gfErrors |= ERR_NUMVAL;
		return;
	}

	gTIMEOUT = 200;							//0.2 seconds emergency brake

	while ((I2C1->ISR & (I2C_ISR_TXE | I2C_ISR_RXNE )) == 0); //as precaution
	I2C1->CR2 &= 0xFC00FBFF;			//clear NBYTES area,AUTOEND,RD_WRN
	I2C1->CR2 |= 0x00010000;			//enter 1 BYTE, no AUTOEND
// transmit slave address
	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0)  {
		if (I2C1->ISR & I2C_ISR_NACKF)  {
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;			//reset all relevant flags

//transmit register control byte
	I2C1->TXDR = *gfSP & 0x000000FF;
	gfSP--;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0)  {
		if (I2C1->ISR & I2C_ISR_NACKF)  {
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	I2C1->ICR &= 0x000087F8;			//reset all relevant flags
//switch to receive mode
	I2C1->CR2 &= 0xFC00FFFF;				//clear NBYTES area
	I2C1->CR2 |= (I2C_CR2_AUTOEND | I2C_CR2_RD_WRN);
	I2C1->CR2 |= nBytes << 16;				//enter n bytes to be received
// transmit slave address
	I2C1->CR2 |= I2C_CR2_START;
	while ((I2C1->ISR & I2C_ISR_TXE) == 0)  {
		if (I2C1->ISR & I2C_ISR_NACKF)  {
			gfErrors |= ERR_COMMACC;
			return;
		}
	}
	for (k=0; k < nBytes; k++) {
//		dummy = I2C1->RXDR;						//formally to reset RXNE flag
		while ((I2C1->ISR & I2C_ISR_RXNE) == 0)  {
			if ((I2C1->ISR & 0x00000F50) || (gTIMEOUT==0))  {
			//any of PECERR,OVR,ARLO,BERR,TC,NACKF flag set
				gfErrors |= ERR_COMMACC;
				return;
			}
		}
		gfSP++;
		*gfSP =	I2C1->RXDR & 0x000000FF;
		I2C1->ICR &= 0x000087F8;				//reset all relevant flags
	}
}


void op_REGWRITE(void)  // *************************************************
// (value  peripheral_register_address -- )
// writes "value" the addressed peripheral register
// register address: see Reference Manual and/or STM32G431xx.h
// *************************************************************************
{
uint32_t* RegAddr;


    if ((*gfSP < 0x40000000)||(*gfSP > 0x50060C00)) {
										//BOT&TOP see RefMan
        gfErrors |= ERR_NUMVAL;
        return;
    }
    RegAddr = (uint32_t*)*gfSP;
    *RegAddr = *(gfSP-1);
    gfSP -= 2;
}


void op_REGREAD(void)  // **************************************************
// (peripheral_register_address -- value)
// returns the content of the addressed peripheral register
// register address: see Reference Manual and/or STM32G431xx.h
// *************************************************************************
{
uint32_t* RegAddr;

    if ((*gfSP < 0x40000000) || (*gfSP > 0x50060C00)) { 
										//BOTTOM&TOP see RefMan

        gfErrors |= ERR_NUMVAL;
        return;
    }
    RegAddr = (uint32_t*)*gfSP;
    *gfSP = *RegAddr;
}


void op_REGBSET(void)  // ***************************************************
// (bitNo  peripheral_register_address -- )
// SETS bitNo of the the addressed peripheral register
// register address: see Reference Manual and/or STM32G431xx.h
// *************************************************************************
{
uint32_t* RegAddr;
uint32_t temp;

    if ((*gfSP < 0x40000000)||(*gfSP > 0x50060C00)) {
										//BOT&TOP see RefMan
        gfErrors |= ERR_NUMVAL;
        return;
    }
	if (*(gfSP-1) > 31) {
		gfErrors |= ERR_NUMVAL;
        return;
	}
    RegAddr = (uint32_t*)*gfSP;
	temp = 1 << *(gfSP-1);				//bitNo-->bitMask
    *RegAddr |= temp;
	gfSP -= 2;
}

void op_REGBCLR(void)  // ***************************************************
// (bitNo  peripheral_register_address -- )
// CLEARS/RESETS bitNo of the the addressed peripheral register
// register address: see Reference Manual and/or STM32G431xx.h
// *************************************************************************
{
uint32_t* RegAddr;
uint32_t temp;

    if ((*gfSP < 0x40000000)||(*gfSP > 0x50060C00)) {
										//BOT&TOP see RefMan
        gfErrors |= ERR_NUMVAL;
        return;
    }
    if (*(gfSP-1) > 31) {
		gfErrors |= ERR_NUMVAL;
		return;
	}
    RegAddr = (uint32_t*)*gfSP;
	temp = ~(1 << *(gfSP-1));			//bitNo-->bitMask and invert bitwise
    *RegAddr &= temp;
	gfSP -= 2;
}


void op_REGBTST(void)  // ***************************************************
// (bitNo  peripheral_register_address -- bitLevel)
// CLEARS/RESETS bitNo of the the addressed peripheral register
// register address: see Reference Manual and/or STM32G431xx.h
// *************************************************************************
{
uint32_t* RegAddr;
uint32_t temp;

    if ((*gfSP <0x40000000)||(*gfSP >0x50060C00))  {
										//BOT&TOP see RefMan
        gfErrors |= ERR_NUMVAL;
        return;
    }
	if (*(gfSP-1) > 31) {
		gfErrors |= ERR_NUMVAL;
		return;
	}
    RegAddr = (uint32_t*)*gfSP;
	temp = 1 << *(gfSP-1);				//bitNo-->bitMask
    temp = *RegAddr & temp;
    gfSP--;
	if (temp) *gfSP = 1;
	else *gfSP = 0;
}


// structuring ops  ########################################################

void op_IF(void)  // *******************************************************
//( -- addr )
//compiles doIF
//starts compilation of a IF ... ELSE ... THEN conditional
//At compile time,it compiles doIF, provides a placeholder for jump address
//and sends its code address via stack to ELSE or THEN
// *************************************************************************
{
	compileTCode(IDX_DOIF);
	gfSP++;
	*gfSP = gfCodeIndex;        //keep in mind for ELSE or THEN
	compileTCode(0xFFFFFFFF);   //placeholder, will be filled by ELSE or THEN
}


void op_CASE(void)  // *******************************************************
//( -- addr )
//compiles doIF
//starts compilation of a CASE ... ELSE ... THEN conditional
//At compile time,it compiles doCASE, provides a placeholder for jump address
//and sends its code address via stack to ELSE or THEN
// *************************************************************************
{
	compileTCode(IDX_DOCASE);
	gfSP++;
	*gfSP = gfCodeIndex;        //keep in mind for ELSE or THEN
	compileTCode(0xFFFFFFFF);   //placeholder, will be filled by ELSE or THEN
}


void op_LTCASE(void)  // *************************************************
//( -- addr )
//compiles doLTCAS
//starts compilation of a <CASE ... ELSE ... THEN conditional
//At compile time,it compiles do<CASE,provides a placeholder for jump address
//and sends its code address via stack to ELSE or THEN
// *************************************************************************
{
	compileTCode(IDX_DOLTCAS);
	gfSP++;
	*gfSP = gfCodeIndex;        //keep in mind for ELSE or THEN
	compileTCode(0xFFFFFFFF);   //placeholder, will be filled by ELSE or THEN
}


void op_GTCASE(void)  // *************************************************
//( -- addr )
//compiles doGTCAS
//starts compilation of a >CASE ... ELSE ... THEN conditional
//At compile time,it compiles do<CASE,provides a placeholder for jump address
//and sends its code address via stack to ELSE or THEN
// *************************************************************************
{
	compileTCode(IDX_DOGTCAS);
	gfSP++;
	*gfSP = gfCodeIndex;        //keep in mind for ELSE or THEN
	compileTCode(0xFFFFFFFF);   //placeholder, will be filled by ELSE or THEN
}


void op_ELSE(void)  // *****************************************************
//( addr -- addr)
//compiles doELSE
//compiles middle part of a IF ... ELSE ... THEN conditional
//At compile time, it compiles doELSE ,its code address into the placeholder
//behind doIF and provides a placeholder for jump address.
//The stack entry with the code address provided by IF is
//exchanged with an entry that contains the code address of doELSE.
// *************************************************************************
{
uint32_t  temp;
	compileTCode(IDX_DOELSE);
	temp = gfCodeIndex;		            //keep in mind for THEN
	compileTCode(0xFFFFFFFF);		    //placeholder, will be filled by THEN
	gfTCode[*gfSP] = (uint32_t)(gfTCode + gfCodeIndex -1);
							//write jump target into compilation of IF or ELSE
							//word pointer index offset is bytes*4
	//calculated 1 word less because pointer is incremented in executeTCode()
	*gfSP = temp;
}


void op_THEN(void)  // ****************************************************
//( addr -- )
//compiles termination of a IF ... ELSE ... THEN conditional
//Essentially it compiles the terminating code address
//into the placeholder behind doIF (or doELSE if present).
// *************************************************************************
{
	gfTCode[*gfSP] = (uint32_t)(gfTCode + gfCodeIndex -1);
							//write jump target into compilation of IF or ELSE
							//word pointer index offset is bytes*4
	//calculated 1 word less because pointer is incremented in executeTCode()
	gfSP--;
}


void op_DO(void)  // *******************************************************
//( -- addr )
//compiles doDO
//starts compilation of a DO ... WHILE or DO ... AGAIN loop.
//Essentially it provides a placeholder for the BREAK code address
//and sends its code address via stack to WHILE or AGAIN
// *************************************************************************
{
    compileTCode(IDX_DODO);//at runtime DODO puts BREAK adddress on Return Stack
	gfSP++;
	*gfSP = gfCodeIndex;        //sends loopback address to "WHILE" or "AGAIN"
	compileTCode(0xFFFFFFFF);   //placeholder for BREAK adddress
                                //will be filled by WHILE or AGAIN
}


void op_UNTIL(void)  // ****************************************************
//( addr -- )
//compiles doUNTIL
//finishes compilation of a DO ... UNTIL loop
//Essentially it compiles its code address for BREAK behind doDO
//and compiles the loopback code adress provided by DO behind doUNTIL.
// *************************************************************************
{
	compileTCode(IDX_DOUNTIL);
	prepareBack();
}


void op_WHILE(void)  // ****************************************************
//( addr -- )
//compiles doWHILE
//finishes compilation of a DO ... WHILE loop
//Essentially it compiles its code address for BREAK behind doDO
//and compiles the loopback code adress provided by DO behind doWHILE.
// *************************************************************************
{
	compileTCode(IDX_DOWHILE);
	prepareBack();
}


void op_AGAIN(void)  // ***************************************************
//( addr -- )
//compiles doAGAIN
//finishes compilation of a DO ... AGAIN loop
//Essentially it inserts its code address for BREAK compilation behind doDO
//and compiles the loopback code adress provided by DO behind doAGAIN.
// *************************************************************************
{
	compileTCode(IDX_DOAGAIN);
	prepareBack();
}


void prepareBack(void)  // ********************************************************
// *************************************************************************
{
	compileTCode((uint32_t)(gfTCode + *gfSP));	//loopback address sent by DO
	gfTCode[*gfSP] = (uint32_t)(gfTCode + gfCodeIndex -1);
//word pointer index offset is bytes*4
//gfCodeIndex was incremented by "compileTCode()" ie. points to RPDEC then.
//must be decremented here because gfIP is increased when doBREAK etc.is left
	compileTCode(IDX_RPDEC);	//compile ReturnStack reconstruction  handler
	gfSP--;
}


void op_FOR(void)  // ******************************************************
//( -- addr )
//compiles doFOR
//starts compilation of a FOR ... LOOP loop
//The runtime behaviour is very similar to "C" programming:
//doFOR expects 3 DataStack entries: startIndex, endIndex, Increment
//Increment can be positive or negative and can be greater than 1.
//For runtime details of the FOR...LOOP loop see doFOR and doLOOP below.
//Essentially FOR provides a placeholder for the BREAK code address
//and sends its code address via stack to LOOP.
// *************************************************************************
{
    compileTCode(IDX_DOFOR);//at runtime DOFOR puts BREAK adddress on Ret.Stack
	gfSP++;                     //sends the return address to "LOOP"
	*gfSP = gfCodeIndex;
	compileTCode(0xFFFFFFFF);   //placeholder, will be filled by LOOP
}


void op_LOOP(void)  // *****************************************************
//( addr -- )
//compiles doLOOP
//finishes compilation of a FOR ... LOOP loop
//Essentially it inserts its code address for BREAK compilation behind doFOR
//and compiles the loopback code adress provided by FOR behind doLOOP.
// *************************************************************************
{
	compileTCode(IDX_DOLOOP);
	compileTCode((uint32_t)(gfTCode + *gfSP));//compile jump back to behind FOR
    //word pointer index offset is bytes*4
	gfTCode[*gfSP] = (uint32_t)(gfTCode + gfCodeIndex -1);
									//sends the CONTI & BREAK address to "FOR"
//gfCodeIndex was incremented by "compileTCode()" ie. points to RP4DEC then.
//must be decremented here because gfIP is incremented when doBREAK etc. is left
	compileTCode(IDX_RP4DEC);
	gfSP--;
}


void op_IDX(void)  // ******************************************************
//( -- index)
//at runtime= copies 2nd Return Stack entry = actual loop parameter on TOS
// *************************************************************************
{
    gfSP++;
    *gfSP = *(gfRP - 1);
}


void op_CONTI(void)  // ****************************************************
//( --- )
//initiates unconditional jump to behind doDO or doFOR (starts new loop)
//performs same action as BREAK, but 2 words shorter to the loopback
//gfIP is increased before execution, so stored always one word less before
//More than 1 CONTI may be within a loop, each with individual decsion
//gfIP will be incremented when op_DOCONTI is left
// *************************************************************************
{
uint32_t test;
uint32_t * addr;

	IWDG->KR = 0x0000AAAA;	//feed watchdog to keep permanent CONTI alive

	addr = (uint32_t*)(*gfRP-4);	//word pointer increment = 4 bytes!
	test = *addr;
//first check which action is necessary for this situation:
	if (test == IDX_DOLOOP) gfIP = (uint32_t*)(addr-1);
//in this case, CONTI will perform doLOOP with loop index update
	else  {
//in these cases, CONTI will get the loopback address and go there
		addr++ ;					//word pointer increment = 4 bytes!
		addr = (uint32_t*)(*addr);
		gfIP = (uint32_t*)(addr);
								
	}
}


void op_BREAK(void)  // ****************************************************
//( --- )
//initiates unconditional jump to (value from Return Stack)
// *************************************************************************
{
	gfIP = (uint32_t*)*gfRP;
        //the BREAK IP is stored in *gfRP by doDO or doFOR
        //gfIP will be incremented in "executeTcode()" when op_DOBREAK is left
        //gfRP is reset by RP-1 (DO loop) or RP-4 (FOR loop)
}


void op_ZBREAK(void)  // ****************************************************
//( flag --)
//if (flag==0) linear loop progression
//else it initiates a jump to (value from Return Stack)
// *************************************************************************
{
    if (*gfSP == 0) gfIP = (uint32_t*)*gfRP;     //perform BREAK
    //else do nothing
    gfSP--;
}


void op_RETURN(void)  // ***************************************************
//( --- )
//leaves a user thread before end (equivalent to "C" return;)
//see op_SEMIS , too
// *************************************************************************
{
	gfIP = (uint32_t *)*gfRP;
	if (gfRP > gfReturnStack) {
//new 30Dec23:
		if (gfBackProcess == 2)  {
			if (gfNestLevel == gfBackNestLevel)  {
				gfBackNestLevel = 0;
				gfBackProcess = 3;
			}
		}
		gfRP--;				//else gfRP==gfReturnStack-->finished
		gfNestLevel --;
	}
	else  {
		gfIP =  0;		// signal termination to executeTcode()
		if (gfBackProcess == 2)  gfBackProcess = 3;
	}
}

// compiler ops  ##############################################################

void op_PRINTSTACK(void)  // ***********************************************
// simple debug function: displays all stack entries. TOS is rightmost
//for debug purposes exceptionally it can be compiled into User Threads
// *************************************************************************
{
uint32_t k;
const uint8_t UnderflowStr [] = "underflow! ";
const uint8_t emptyStr [] = " --- ";

    sendTB('[');
	if (gfSP == (gfDataStack + 3))  sendString(emptyStr);
    else if (gfSP < gfDataStack + 3)  {
        sendString(UnderflowStr);
        gfSP = gfDataStack; //re-init stack
        *gfSP=0;
        gfSP++;
        *gfSP = 0;          //four stack items as underflow reserve
        gfSP++;
        *gfSP = 0;
        gfSP++;
        *gfSP = 0;
    }
    else  {
        k = 4;
        while ((gfSP > gfDataStack + 3)&&(k <= (uint32_t)(gfSP-gfDataStack))) {
                    //4 stack items as underflow reserve not displayed
            sendTB(SPACE);
            if (gfTempNumBase == 16)  send0X32(gfDataStack[k]);
            else  sendNS32(gfDataStack[k]);
			sendTB(SPACE);
            k++ ;
        }
    }
    sendTB(']');
}


void op_FLUSH(void)  // ****************************************************
// *************************************************************************
{
    gfSP = gfDataStack;
	*gfSP=0;
	gfSP++;
	*gfSP = 0;        	//four stack items as underflow reserve
    gfSP++;
	*gfSP = 0;
	gfSP++;
	*gfSP = 0;
}


void op_OPS(void)  // ******************************************************
//display all kernel operators: their names and their token values
// *************************************************************************
{
uint32_t k, n, p;
const uint8_t KernelStr [] = "Kernel:";
#ifdef DMX
const uint8_t DmxStr [] = "DMX:";
#endif

    sendString(KernelStr);
    sendTB(CRR);
    k = 1;				//don't show {ERR} dummy name
    n = 0;
	while (k < KERNELSIZE) {
        sendTB(SPACE);
        for (p=0;p<9;p++)  {
            if (gfKernelNames[k][p]) sendTB(gfKernelNames[k][p]);
            else break;
        }
        while(p < 7) {
            sendTB(SPACE);
            p++;
        }

        sendTB('#');
        sendX8((uint8_t)k);
        sendTB(SPACE);
        sendTB(SPACE);
        sendTB(SPACE);
#ifdef DMX
		if (k == DMXFIRST)  {		//defined in globaldefines.c
			n = 4;
			sendTB(CRR);
			sendString(DmxStr);
		}
#endif
		k++;
		n++;
		if (n==5) {
            sendTB(CRR);
			n = 0;
		}
	}
}


void op_USER(void)  // *****************************************************
//display all user token: their names and their token values
// *************************************************************************
{
uint32_t k, p, wrap;
const uint8_t VarconStr [] = "(VarCon=";

    k = 0;
    wrap = 0;
	while (k < gfNextNameCompi)  {
        if (k<10) sendTB(SPACE);
        sendNS32(k);
        sendTB(':');
        for (p=0;p<16;p++)  {
            if (gfUserHeader[k].name[p]) sendTB(gfUserHeader[k].name[p]);
            else break;
        }

		if (gfUserHeader[k].attributes & A_VARCON)  {
			sendString(VarconStr);
            if (gfTempNumBase == 16)  send0X32(gfUserHeader[k].toCompile);
            else sendNS32(gfUserHeader[k].toCompile);
            sendTB(SPACE);
            sendTB('@');
            send0X32(((uint32_t)&gfUserHeader[k].toCompile) & 0x0000FFFF);
			sendTB(0x29);     //')'
            wrap += 2;
        }
        sendTB(SPACE);
        sendTB(SPACE);
		k++;
		wrap++;
		if (wrap >= 6) {
			sendTB(CRR);
			sendTB(SPACE);
			wrap = 0;
		}
	}
	if (wrap) sendTB(CRR);
}


void op_SEE(void)  // ******************************************************
//show the compiled body of a user thread
// *************************************************************************
{
uint32_t k, n, wrap, w, count, word, byte, tCode, address;
uint32_t *vtest;
uint8_t *letter;
const uint8_t KernelStr [] = "can't SEE Kernel Ops";
const uint8_t UserStr [] = "UserHeader_Struct=";
const uint8_t VarconStr [] = "VarCon:Value=";


	gfTokenStart = gfTibPos;

	scanNext();
	
    if (findUserToken() == 0) {     //updates "gfParsed"
        if (findKernelToken())  {
            sendString(KernelStr);
            return;
        }
		gfErrors |= ERR_STRINGTOKEN;
		return;
	}
	gfTokenStart = gfTibPos;
	sendString(UserStr);
	send0X32(((uint32_t)gfUserHeader + gfParsed*24));	//24 bytes/struct
	sendTB(CRR);

	tCode = gfUserHeader[gfParsed].toCompile;
	if (tCode & 0x00080000)  {
		tCode &= 0xFFF7FFFF;					//delete RESTORE bit19 18Oct23
		k = (tCode - (uint32_t)gfTCode) / 4;
	}
	else k = gfUserHeader[gfParsed].attributes & 0x00000FFF;

	wrap = 0;
	if ((gfUserHeader[gfParsed].attributes & A_VARCON) == 0) {
														//"normal" Function
	    for (n = k;gfTCode[n] != IDX_SEMIS;n++) {
			wrap++;
			if (wrap >= 8) {
				sendTB(CRR);
				wrap = 0;
			}
			address = (uint32_t)(&(gfTCode[n])) & 0x0000FFFF;
			sendNX32(address);	//leading 0x2000.. and 0x suppressed
			sendTB(':');

			if (gfTCode[n] >= (uint32_t)&gfUserHeader[0]) { //new 19Jan25
				vtest = (uint32_t*)gfTCode[n] +1;
				if (*vtest & A_VARCON)  {
					valAdr2String(gfTCode[n]);	//VARCON never REPACEd
					sendTB(0x28);	 //'('
					sendTB('=');
					vtest = (uint32_t*)gfTCode[n];
					if (gfTempNumBase==16)  send0X32(*vtest);
					else  sendNS32(*vtest);
					sendTB(0x29);	 //')'
					sendTB(SPACE);
					sendTB(SPACE);
					continue;
				}
			}

			token2String(gfTCode[n]);
			tCode = gfTCode[n] & 0xFFF7FFFF;	//delete RESTORE bit19 18Oct23

			if ((tCode >= IDX_DOIF) && (tCode <= IDX_DOLOOP)) {
				n++;
				sendTB(0x28);	 //'('
				send0X32((gfTCode[n] + 4) & 0x0000FFFF);
									//word pointer increment = 4 bytes!
				sendTB(0x29);	 //')'
				wrap++;
			}
//bugfix 18Jan24: below here "else" added, 
//else wrong interpolation of literal values by next "if" possible
			else if (tCode==IDX_DOLIT) {
				n++;
				sendTB(0x28);	 //'('
				if (gfTempNumBase==16)  send0X32(gfTCode[n]);
				else  sendNS32(gfTCode[n]);
				sendTB(0x29);	 //')'
				wrap++;
			}
			else if (tCode==IDX_DOWVA) {
				n++;
				sendTB(0x28);	 //'('
				valAdr2String(gfTCode[n]);	//VARCON never REPACEd
				sendTB('@');
				send0X32(gfTCode[n] & 0x0000FFFF);
				sendTB(0x29);	 //')'
				wrap++;
			}
			else if (tCode==IDX_DOSTR) {
				sendTB('"');
				n++;
				count = gfTCode[n];		//byte count
				n++;
				w= 0;
				word= 0;
				while (count)  {
					if (w == 0) {
						word = gfTCode[n];
						n++;
				}
					byte = (word >> w) & 0xFF;
					if ((byte >= 0x20) && (byte < 0x7F))  sendTB(byte);
					else  {
						sendTB(0x5C);        //Backslash
						sendNibble((byte >> 4) & 0x0000000F);
						sendNibble(byte & 0x0000000F);
					}
					w += 8;
					if (w > 24) w = 0;
					count --;
				}
				n-- ;	//during while{} already increment by for{ preempted
				sendTB('"');
				wrap += 3;
			}
			sendTB(SPACE);
			sendTB(SPACE);
	    }
		address = (uint32_t)(&(gfTCode[n])) & 0x0000FFFF;
		sendNX32(address);	//leading 0x2000.. and 0x suppressed
		sendTB(':');
	    sendTB(';');
	}
	else {
		sendString(VarconStr);
		if (gfTempNumBase == 16)  send0X32(gfUserHeader[gfParsed].toCompile);
		else  sendS32(gfUserHeader[gfParsed].toCompile);
		sendTB(SPACE);
		sendTB('@');
		send0X32((uint32_t)(&gfUserHeader[gfParsed].toCompile) & 0x0000FFFF);
						//address always in hex, leading 0x2000.. suppressed
	}
    sendTB(CRR);
}


void token2String(uint32_t tokenval)   // **********************************
// *************************************************************************
{
uint32_t k, p;

	if (tokenval < KERNELSIZE) {        //Kernel token
        for (p=0;p<16;p++)  {
            if (gfKernelNames[tokenval][p]) sendTB(gfKernelNames[tokenval][p]);
            else break;
        }
	}
	else {                              //User token
		for (k=gfNextNameCompi; k >= 0; k--)  {
    		if (gfUserHeader[k].toCompile == tokenval) break;
		}
        for (p=0;p<16;p++)  {
            if (gfUserHeader[k].name[p]) sendTB(gfUserHeader[k].name[p]);
            else break;
        }
	}
}


void valAdr2String(uint32_t addr)   // ********************************
//reconstruct VAR or CONST name from address of .toCompile  struct entry
// *************************************************************************
{
uint32_t p;
uint8_t* byteAddress;
uint8_t letter;

    byteAddress = (uint8_t*)addr - 16;	//pointer to .name[0

	sendTB(0x27);	//= '
    for (p=0;p<16;p++)  {
        letter = *(byteAddress + p);
        if (letter) sendTB(letter);
        else break;
    }
	sendTB(0x27);
}


void op_MEM(void)  // ******************************************************
//(addr -- )
//send 32 Token Code entries starting from "addr" (16bit, w.o. leading 0x2000..)
//for debug purposes exceptionally it can be compiled into User Threads
// *************************************************************************
{
uint32_t *address;
uint32_t k, value, wrap;
const uint8_t StartStr [] = "128 SRAM bytes (32 words) starting from 0x";

	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	value =  *gfSP;
	gfSP--;
    value |= 0x20000000;
    address = (uint32_t *)value;
    if ((uint32_t)address > 0x20007F80) {
        gfErrors |= ERR_NUMVAL;
        return;
    }
    sendTB(CRR);
    sendString(StartStr);
    sendX32((uint32_t)address);

    wrap = 3;
    for (k=0; k<32; k++)  {
        value = (uint32_t)(address + k);
        wrap++;
        if (wrap == 4) {
            sendTB(CRR);
            sendX32(value);
            sendTB(':');
            wrap = 0;
        }
        value = *(address + k);
        sendX8((uint8_t)(value >> 24));
        sendX8((uint8_t)((value >> 16) & 0x000000FF));
        sendX8((uint8_t)((value >> 8) & 0x000000FF));
        sendX8((uint8_t)((value) & 0x000000FF));
        sendTB('|');
    }
    sendTB(CRR);
}


void op_STEP(void)  // ****************************************************
//( --- )
// activates single step execution, trigger next step with TAB key
// special cases: key "ESC" terminates execution at once (leave endless loops)
//                key CTRL_B terminates long BLOCK waiting immediately
//for debug purposes exceptionally it can be compiled into User Threads
// *************************************************************************
{
     gfBackProcess = 3;

}


void op_RUN(void)  // ****************************************************
//( --- )
//default state of background process, always active after system start 
//de-activates single step execution, but keeps the background process active
// special cases: key "ESC" terminates execution at once (leave endless loops)
//                key CTRL_B terminates long BLOCK waiting immediately
//for debug purposes exceptionally it can be compiled into User Threads
// *************************************************************************
{
    gfBackProcess = 1;
}


void op_NOBAK(void)  // ****************************************************
//( --- )
// de-activates background process. Token Code speed is approx 40% faster
//for speed control exceptionally it can be compiled into User Threads
// *************************************************************************
{
    gfBackProcess = 0;
}


void op_BAKOP(void)  // ****************************************************
//( --- )
// restores RUN mode when background process is OFF
// *************************************************************************
{
const uint8_t backStr [] = " ->BackOP ";

	if (gfBackProcess == 0)  {
		sendString(backStr);
		gfBackProcess = 1;
		gfTermCtrlBC = 0;
	}
}


void op_FORGET(void)  // ***************************************************
//delete the argument user function and all later defined ones
//example:
// : XYZ .......
// FORGET XYZ
// *************************************************************************
{
uint32_t k;
uint32_t *pointer;

	gfTokenStart = gfTibPos;
	scanNext();
	if (findUserToken() == 0) {
		gfErrors |= ERR_USERSTRING;
		return;
	}
	gfTokenStart = gfTibPos;
//update compiler pointers
	if (gfParsed < 7)  {		//NSTP1..3, V1..4 cannot be deleted
		gfErrors |= ERR_CONTEXT;
		return;
	}
	gfNextNameCompi = gfParsed;
	gfUserHeader[gfParsed].name[0] = 0;
	gfCodeIndex = gfUserHeader[gfParsed].attributes & 0x00000FFF;
//works with VARCON too, VARCON compilation doesn't modify gfCodeIndex
	gfCodeStartIndex = gfCodeIndex;
//as precaution, set the rest of gfUserHeader[] and gfTCode completely to zero
    pointer = (uint32_t*)&(gfUserHeader[gfNextNameCompi]);
    while (pointer < (uint32_t*)&gfUserHeader[MAXUSER])  {
        *pointer = 0;
        pointer++;
    }
    for (k = gfCodeIndex; k < CODETOP; k++)  gfTCode[k] = 0;
}


void op_ABORT()  // ********************************************************
//warm start. Re-initalizes runtime parameters and stacks,
//			  but keeps variable values and compiled functions
// *************************************************************************
{
uint32_t k;
uint32_t *pointer;

	gfComment = 0;
	gfCompiState = 0;
	gfExecUser = 0;
	gfBackNumValid = 0;
	gfNestLevel = 0;
	if (gfBackProcess)  gfBackProcess = 1;		//reset to user selection
	gfTermCheck = 0;
	gfTempNumBase = gfNumBase;

    for (k = 0; k < MAXUSER; k++)  {		//clear thread of token code
        if (gfUserHeader[k].name[0] == 0) break;
    }
    gfNextNameCompi = k;

    k = CODETOP;
    while (1) {
       if (gfTCode[k] == IDX_SEMIS) break;
       if (k == 0) break;
       k--;
    }

    if (k > 0) gfCodeIndex = k + 1;
    else gfCodeIndex = 0;
    gfCodeStartIndex = gfCodeIndex;

//as precaution, set the rest of gfUserHeader[] and gfTCode completely to zero
    pointer = (uint32_t*)&(gfUserHeader[gfNextNameCompi]);
    while (pointer < (uint32_t*)&gfUserHeader[MAXUSER])  {
        *pointer = 0;
        pointer++;
    }
    for (k = gfCodeIndex; k < CODETOP; k++)  {
        gfTCode[k] = 0;
    }

	gfRP = gfReturnStack;
	*gfRP = 0;	            //repeat as precaution, is essential when compiling
	gfIP = gfTCode;
	gfErrors = 0;
	op_FLUSH();
}

void op_REPLACE(void)  // ***************************************************
// replace the token code of <target name> with that of <supplier name>
// Syntax:  REPLACE  <supplier name>  <target name>
// ( -- )  no stack parameters
// <supplier name> is always the first hit searched from TOP of userToken[]
// <target name> is always the first hit searched from TOP of userToken[]
// only token code between supplier and target is replaced
// original state is restored with "RESTORE".
// This PROCEDURE WORKS ONLY 
// if supplier is the first hit
// AND target is the first hit of search from TOP of userToken 

// REPLACE inserts the supplier code address in gfUserHeader[target].toCompile
// with bit19 set.
// but keeps the original gfTCode_index in gfUserHeader[target].attributes. 
// In 32bit token code format, the UserToken.attributes value 
// is it's startAddress offset in gfTCode (32bit words, i.e. *4=offset in bytes)
// *************************************************************************
{
uint32_t   supIndex, targIndex;
uint32_t   supToken, targToken;
uint32_t k;

//parse <supplier name> from TIB
    supIndex = getToken();                 //index in gfUserHeader[]
    if (supIndex == 0xFFFFFFFF)  {
        gfErrors |= ERR_USERSTRING;
        return;
    }
    if (gfUserHeader[supIndex].attributes & A_VARCON)  {
        gfErrors |= ERR_CONTEXT;    //replacement of VARCON makes no sense
        return;
    }

//parse <target name> from TIB
    targIndex =  getToken();                //index in gfUserHeader[]
    if (targIndex == 0xFFFFFFFF)  {
        gfErrors |= ERR_USERSTRING;
        return;
    }
    if (gfUserHeader[targIndex].attributes & A_VARCON)  {
        gfErrors |= ERR_CONTEXT;    //replacement of VARCON makes no sense
        return;
    }
    if (supIndex < targIndex)  {
		gfErrors |= ERR_CONTEXT;
		return;
	}
    supToken = gfUserHeader[supIndex].toCompile  | 0x00080000;//mark REPLACEd;
    targToken = gfUserHeader[targIndex].toCompile;
	gfUserHeader[targIndex].toCompile = supToken;
	supIndex = gfUserHeader[supIndex].attributes & 0x00000FFF;
					//i.e.all code >= target and < supplier will be replaced

	for (k = 0;k <= gfCodeStartIndex; k++) {
		if (gfTCode[k] == targToken) {
			gfTCode[k] = supToken;
		}
	}
}


void op_RESTORE(void)  // **************************************************
// Syntax:  RESTORE <original name>
// ( -- )  no stack parameters
// <original name> is the first hit searched down from TOP of userToken[].
// If this is "REPLACE"d (wears "suppliers" name plus bit19 set, 
// the threaded code is reset to the original.
// If nothing has been "REPLACE"d, nothing gets changed
// The procedure is made possible as follows:
// REPLACE inserts the supplier code address in gfUserHeader[target].toCompile
// with bit19 set.
// but keeps the original gfTCode index in gfUserHeader[target].attributes. In
// In 32bit token code format, the UserToken.attributes value 
// is it's startAddress offset in gfTCode (32bit words, i.e. *4=offset in bytes)
// *************************************************************************
{
 uint32_t  index, maxIndex, origIndex, actualCode, origCode, k;

//parse <name of originalized token> from TIB
    index = getToken();                    //index in gfUserHeader[]
    if (index == 0xFFFFFFFF)  {
        gfErrors |= ERR_USERSTRING;
        return;
    }
    actualCode = gfUserHeader[index].toCompile;
//gfUserHeader[index].attributes contains another formulation of orig "toCompile"
	origIndex = gfUserHeader[index].attributes & 0xFFF;
	origCode = (uint32_t)(&gfTCode[origIndex]);
//protect "native actualCode"
	if ((actualCode & 0xFFF7FFFF) <= origCode)  {
		gfErrors |= ERR_CONTEXT;
		return;
	}

	gfUserHeader[index].toCompile = origCode;	//restore in gfUserHeader[]

	for (k = 0;k <= gfCodeStartIndex; k++) {
		if (gfTCode[k] == actualCode)  {
			gfTCode[k] = origCode;		//restore in gfTCode[]
		}
    }
}


uint32_t getToken(void)   // ***********************************************
// *************************************************************************
{
int32_t k;
uint32_t n;
uint8_t refChar, userChar;

	gfTokenStart = gfTibPos;
	scanNext();
	gfTokenStart = gfTibPos;
	k = gfNextNameCompi - 1;		//k is index of last compiled USER Token
									//search DOWN from the LAST COMPILED
    while (k != -1)  {
        for (n=0; n < 16;n++) {
            userChar = gfUserHeader[k].name[n];
            refChar = gfTokenstr[n];
//transform to capital letters:
            if ((refChar >= 'a')&&(refChar <= 'z')) refChar -= 0x20;
            if (userChar != refChar)  break;
            if (userChar == 0) return k;//="gfUserHeader[]"index of found token
        }
        k--;
    }
	return 0xFFFFFFFF;                  //if named token not found
}


void op_COLON(void)  // ****************************************************
//( --- )
// start the compilation of a user function.
// *************************************************************************
{
uint32_t userSymbol;

    userSymbol = compileUserSymbol(0);
	if (userSymbol == 0)  return;
	if (userSymbol == 2)  gfUserHeader[gfNextNameCompi].name[0] = 0;	
							//compilation cancelled, mark end of Symbol Table
	else  {
//will be a "normal" token, no VARCON
		gfCodeStartIndex = gfCodeIndex;
		gfCompiState = 1;
	}
}


void op_SEMIS(void)  // *****************************************************
//( --- )
//terminates every compilation, i.e. must be the last thread token
// **************************************************************************
{
	if (gfCompiState != 0)  {            //during compilation
		compileTCode(IDX_SEMIS);
		gfCompiState = 0;
		gfNextNameCompi ++;
		gfUserHeader[gfNextNameCompi].name[0] = 0;	//mark end of Symbol Table
//update compiler pointers
		gfCodeStartIndex = gfCodeIndex;
	}
	else  {								//interpreting (runtime) state
		gfIP = (uint32_t*)*gfRP;
		if (gfRP > gfReturnStack) {
			if (gfBackProcess == 2)  {
				if (gfNestLevel == gfBackNestLevel)  {
					gfBackNestLevel = 0;
					gfBackProcess = 3;
				}
			}
			gfRP--;				//else gfRP==gfReturnStack-->finished
			gfNestLevel --;
		}
		else  {
			gfIP =  0;		// signal termination to executeTcode()
			if (gfBackProcess == 2)  gfBackProcess = 3;
		}
	}
	gfTempNumBase = gfNumBase;
}


void op_DOSTRING(void)  // *************************************************
//( --- )
//runtime operator, compiled by P"
//In ThreadCode memory it is followed by the bytes to be transmitted.
//differing from initial CForth versions, the string is "byte counted".
// *************************************************************************
{
	subDoString(1);
}

#ifndef MINI
void op_DOIOSTR(void)  // **************************************************
//( --- )
//runtime operator, compiled by TX"
//remark see op_DOSTRING
// *************************************************************************
{
	subDoString(0);
}
#endif

void subDoString(uint32_t out)  // *****************************************
// *************************************************************************
{
uint32_t n, count, word;

	gfIP++;
	count = *gfIP;
	gfIP++;
	n = 0;
	while (count)  {
		if (n == 0) {
			word = *gfIP;
			gfIP++;
		}
		if (out == 1) sendTB((word >> n) & 0xFF);
#ifndef MINI
		else sendIOB((word >> n) & 0xFF);
#endif
		n += 8;
		if (n > 24) n = 0;
		count --;
	}
	gfIP--;		//during while{} already increment by executeTCode() preempted
}


void op_DOLIT(void)  // ****************************************************
//( -- number)
//In the compiled thread, doLIT puts the content of the next ThreadCode entry
//(=number value) on TOS at runtime.
// *************************************************************************
{
	gfSP++;
	gfIP++;
	*gfSP = *gfIP;
}


void op_DOWVA(void)  // *****************************************************
//(value -- )
//In the compiled thread,the "value" address of VARCON is comiled behind doWVA
//At runtime, "value" is written into this ".toCompile" field.
// *************************************************************************
{
uint32_t* address;

	gfIP++;
	address = (uint32_t*)*gfIP;
	*address = *gfSP;
	gfSP--;
}


void op_DOIF(void)  // *****************************************************
//( flag --)
//compiled by IF
//takes a flag from TOS, if == 0, a jump is initiated:
//behind doELSE if present,
//or to code address compiled by THEN
//if != 0, execution is performed until doELSE.
//or continues linear (if no doELSE)
// *************************************************************************
{
	if (*gfSP != 0)  gfIP++;
	else  gfIP = (uint32_t*)(*(gfIP+1));	//word pointer increment = 4 bytes!
    gfSP--;
}


void op_DOCASE(void)  // *****************************************************
//( reference --)
//compiled by CASE
//compare reference with NOS
//if equal execution is performed until doELSE.
//or continues linear (if no doELSE)
//if != 0, a jump is initiated behind doELSE if present,
//or to code address compiled by THEN
//!! after this operation former NOS is TOS now for furter handling. 
//!! DROP if not needed
// *************************************************************************
{
uint32_t temp;
	
    temp = *(gfSP-1);
	if (temp == *gfSP)  gfIP++;
	else  gfIP = (uint32_t*)(*(gfIP+1));	//word pointer increment = 4 bytes!
    gfSP--;									//drop reference value
}


void op_DOLTCAS(void)  // ***************************************************
//( reference --)
//compiled by <CASE
//compare reference with NOS
//if NOS is less than reference execution is performed until doELSE.
//			or continues linear (if no doELSE)
//if greater or equal, a jump is initiated behind doELSE if present,
//			or to code address compiled by THEN
//!! after this operation former NOS is TOS now for furter handling. 
//!! DROP if not needed
// *************************************************************************
{
int32_t temp;
	
    temp = (int32_t)*(gfSP-1);
	if (temp < (int32_t)*gfSP)  gfIP++;
	else  gfIP = (uint32_t*)(*(gfIP+1)); //word pointer increment = 4 bytes!
    gfSP--;								//drop reference value
}


void op_DOGTCAS(void)  // ***************************************************
//( reference --)
//compiled by >CASE
//compare reference with NOS
//if NOS is less than reference execution is performed until doELSE.
//			or continues linear (if no doELSE)
//if greater or equal, a jump is initiated behind doELSE if present,
//			or to code address compiled by THEN
//!! after this operation former NOS is TOS now for furter handling. 
//!! DROP if not needed
// *************************************************************************
{
int32_t temp;
	
    temp = (int32_t)*(gfSP-1);
	if (temp > (int32_t)*gfSP)  gfIP++;
	else  gfIP = (uint32_t*)(*(gfIP+1)); //word pointer increment = 4 bytes!
    gfSP--;								//drop reference value
}


void op_DOELSE(void)  // ****************************************************
//( --- )
//compiled by ELSE
//if the flag of doIF != 0, doELSE manages a jump
//behind the code of the ELSE part (compiled by THEN).
//if the flag of doIF == 0, execution is continued behind doELSE.
//At the end of the ELSE part, execution is continued linar.
// *************************************************************************
{
	gfIP = (uint32_t*)*(gfIP+1);	//word pointer increment = 4 bytes!
}


void op_DODO(void)  // *****************************************************
//( --- )
//compiled by DO
//doDO puts the BREAK code address on the ReturnStack
// *************************************************************************
{
    gfRP++;                 //prepare Return Stack for doWHILE or doAGAIN
    gfIP++;                 //address next index of gfTCode
	*gfRP = *gfIP;
// *gfIP contains the IP jump offset relative to gfTCode base address
//this is adjusted for jump to op_RPDEC.
//CONTI needs 2 less go to doWHILE,doAGAIN
//the jump address is compiled 1 less,
//because it is incremented by executeTCode() when BREAK etc. is leaving there
}


void op_DOUNTIL(void)  // **************************************************
//( flag ---)
//compiled by UNTIL
//doWHILE takes the flag from TOS.
//If == 0, execution goes forward via RP-1 for linear progression
//If != 0, execution loops back to behind (doDO + 1word)
// *************************************************************************
{
    if (*gfSP != 0)  {
        gfRP--;						//remove address for CONTI or BREAK
        gfIP += 2;					//jump behind RP1DEC
	}
	else  {
		IWDG->KR = 0x0000AAAA;	//feed watchdog to keep loop longtime alive
		gfIP = (uint32_t*)*(gfIP + 1);	//jump behind (doDO + 1word)
								//word pointer increment = 4 bytes!
	}
    gfSP--;
}


void op_DOWHILE(void)  // **************************************************
//( flag ---)
//compiled by WHILE
//doWHILE takes the flag from TOS.
//If == 0, execution goes forward via RP-1 for linear progression
//If != 0, execution loops back to behind (doDO + 1word)
// *************************************************************************
{
    if (*gfSP == 0)  {
        gfRP--;						//remove address for CONTI or BREAK
        gfIP += 2;					//jump behind RP1DEC
	}
	else  {
		IWDG->KR = 0x0000AAAA;	//feed watchdog to keep loop longtime alive
		gfIP = (uint32_t*)*(gfIP + 1);	//jump behind (doDO + 1word)
								//word pointer increment = 4 bytes!
	}
    gfSP--;
}


void op_DOAGAIN(void)  // ***************************************************
//( --- )
//compiled by AGAIN
//execution loops always back to behind (doDO + 1word)
// *************************************************************************
{
	IWDG->KR = 0x0000AAAA;	//feed watchdog to keep loop longtime alive
	gfIP = (uint32_t*)*(gfIP + 1);
}


void op_DOFOR(void)  // ****************************************************
//( startIndex, endIndex increment -- )
//is RUNTIME, compiled by FOR
//3 items are copied from the DataStack to the ReturnStack:
//TOS is the increment/decrement of loop index per loop
//NOS is the loop end index for loop termination
//3rd DataStack entry -> 2nd ReturnStack entry:start index=loop index
//Furthermore the BREAK code address is put on top of the ReturnStack.
// *************************************************************************
{
//put items to Return Stack:
    gfRP += 4;
	*(gfRP - 1) = (int32_t)*(gfSP - 2);	//actual loop counter (==start index)
	*(gfRP - 2) = (int32_t)*(gfSP -1);	//loop end index (for op_DOLOOP)
	*(gfRP - 3) = (int32_t)*gfSP;	//loop index inc/decrement(for op_DOLOOP)
	gfSP -= 3;
	gfIP++;					//address next index of gfTCode
	*gfRP = *gfIP;			//put address for CONTI or BREAK on Return Stack
// *gfIP contains the IP jump offset relative to gfTCode base address
//this is adjusted for jump to op_RP4DEC. CONTI needs 2 less go to doLOOP
//the jump address is compiled 1 less,
//because it is incremented by executeTCode() when BREAK etc. is leaving there
}


void op_DOLOOP(void)  // ***************************************************
//( --- )
//compiled by LOOP
//counts the 2nd ReturnStack entry up or down (depends on 4th entry),
//compares the 2nd entry with the 3rd one,
//and intiates loopback to behind doFOR (if compare is in range)
//or execution goes forward via RP-4 for linear progression.
// *************************************************************************
{
int32_t index, endIndex, deltaIndex;

	index = *(gfRP - 1);
	endIndex = *(gfRP - 2);
    deltaIndex = *(gfRP - 3);
//    index += deltaIndex;					//changed 20April24

	if (deltaIndex >= 0)  {
    	if (index < endIndex)  {
			index += deltaIndex;			//changed 20April24
			IWDG->KR = 0x0000AAAA;	//feed watchdog, timeout ca. 1-2 seconds
            *(gfRP - 1) = index; 
			gfIP = (uint32_t*)*(gfIP+1);	//jump behind doFOR
    	}
    	else  {
            gfRP -= 4;						//tidy up Return Stack
            gfIP += 2;						//proceed behind RP4DEC
    	}
	}
	else  {
		if (index > endIndex)  {
			index += deltaIndex;			//changed 20April24
			IWDG->KR = 0x0000AAAA;  //feed watchdog, timeout ca. 1-2 seconds
            *(gfRP - 1) = index;
			gfIP = (uint32_t*)*(gfIP+1);	//jump behind doFOR
     	}
        else  {
            gfRP -= 4;						//tidy up Return Stack
            gfIP += 2;						//proceed behind RP4DEC
        }
    }
}


void op_RPDEC(void)  // ****************************************************
//( --- )
//compiled by WHILE and AGAIN
//At runtime it removes the top entry from the ReturnStack
// *************************************************************************
{
	gfRP--;
}


void op_RP4DEC(void)  // ***************************************************
//( --- )
//compiled by LOOP
//At runtime it removes tfour top entries from the ReturnStack
// *************************************************************************
{
	gfRP -= 4;
}

// sytem ops  ##############################################################

void op_AUTO(void)  // *****************************************************
//( --- )
//Syntax: AUTOEXE <User Thread name>
//The specified user thread is automatically executed at system start.
//To get it available at system start, the system must start from Flash.
//For details see LOAD and SAVE
// *************************************************************************
{
	gfTokenStart = gfTibPos;
	scanNext();					//copies next Tib word to gfTokenstr[]
	if (findUserToken() == 0) {
		if (gfTokenstr[0] == 0x30)  gfAuto = 0;
		else  gfErrors |= ERR_USERSTRING;
		return;
	}
    gfAuto = gfParsed + 1;          //+1 to distinguish from gfAuto = 0
}


#ifndef MINI
void op_SPLIT(void)  // ****************************************************
//(type -- )
//type:'0'=USB and USART both terminal,
//'1'=USB is terminal, USART is COM I/O port
//'2'=USART is terminal, USB is virtual COM I/O port
//'3'=USB is terminal, USART is MIDI interface
//'4'=USART is terminal, USB is MIDI interface
//the non-terminal part is used for external I/O serial communication (RX, TX)
// *************************************************************************
{
uint32_t split;
const uint8_t tempStr [] = " preset Baud Rate restored! ";
const uint8_t usbStr [] = "USB Terminal_Com will get lost!";
const uint8_t usartStr [] = "USART Terminal_Com will get lost!";
const uint8_t midiStr [] = "USART will change to MIDI I/O!";

	if (*gfSP > 4) {
		gfErrors |= ERR_NUMVAL;
		return;
	}

    split = *gfSP;
    gfSP --;
	if (split == gSplit)  return;
	if ((split == 2) || (split == 4))  {
		sendString(usbStr);
		sendTB(CRR);
		if (conFirm() == 0) return;
	}
	if (split == 1)   {
		sendString(usartStr);
		sendTB(CRR);
		if (conFirm() == 0) return;
	}
	if (split == 3)  {
		sendString(midiStr);
		sendTB(CRR);
		if (conFirm() == 0) return;
        gTempBaud = 31;				//MIDI baudRate
		initUsart1();
	}
	gSplit = split;
	if (((gSplit != 1) && (gSplit != 3)) && (gTempBaud != gBaud)) {
		gTempBaud = gBaud;
		initUsart1();
		sendString(tempStr);
		sendTB(CRR);
	}
}

void op_BAUD(void)  // ******************************************************
// (Baud -- )
// parameters: ONLY FIRST 2 DIGITS of decimal baudrate are accepted:
// 11=115200,57=57600,38=38400,31=31250(MIDI),19=19200,96=9600,
// BaudRate input is always interpreted DECIMAL, independent of number base
// **************************************************************************
{
uint32_t temp;
const uint8_t tempStr [] = "BaudRate changes temporarily!";
const uint8_t nextBaudStr [] =
"new Baud Rate same for all Projects. Not active before next system start.";

	temp = *gfSP;
	gfSP--;

    if (temp == 0x11) temp = 11;
    if (temp == 0x19) temp = 19;
    if (temp == 0x31) temp = 31;
    if (temp == 0x38) temp = 38;
    if (temp == 0x57) temp = 57;
	if (temp == 0x96) temp = 96;

	if ((temp == 11) || (temp == 19) || (temp == 31) || (temp == 38) ||
											(temp == 57)|| (temp == 96)) {
        if ((gSplit == 1) || (gSplit == 3))  {	//USART is serial/MIDI I/O
            sendString(tempStr);
            sendTB(CRR);
            gTempBaud = temp;
			initUsart1();
        }
        else  {
			sendString(nextBaudStr);		
			gBaud = temp;
			gTempBaud = temp;
			pushFlashPage23();
			gBaudSet = 0;
        }
		sendTB(CRR);
    }
    else  gfErrors |= ERR_NUMVAL;
}

void op_CHANNEL(void)  // **************************************************
// (Channel -- )
// is diminuished by 1 (0xFFFFFFFF, 0...15) to be used for parsing MIDI messages
// *************************************************************************
{
const uint8_t channelStr [] = " 0 CHANNEL accepts ANY MidiCh in received msgs ";

    if (*gfSP <= 16)  {
		gChannel = *gfSP;
        if (gChannel == 0) sendString(channelStr);
		gfSP--;
		pushFlashPage23();
	}
	else  gfErrors |= ERR_NUMVAL;
}
#endif  //not MINI

void op_LED(void)  // ******************************************************
// (LedBusy -- )
// if (gLedBusy == 0) LED is always OFF
// if (gLedBusy == 1) LED is always ON
// if (gLEDBusy == 2) LED blinks when RS-232 data are received 
// if (gLEDBusy == 3) LED blinks when RS-232 data are received or transmitted
//					  (default)
// *************************************************************************
{
	gLedBusy = *gfSP;
	gfSP--;
    if (gLedBusy > 3) {
		gfErrors |= ERR_NUMVAL;
		return;
	}		
#ifdef CASE48  //GREYPILL
	if (gLedBusy == 0)  GPIOB->BSRR = GPIO_BSRR_BR_15;	//PB15 front LED
	if (gLedBusy == 1)  GPIOB->BSRR = GPIO_BSRR_BS_15;
#else
	if (gLedBusy == 0)  GPIOA->BSRR = GPIO_BSRR_BR_8;	//PA8 fron LED
	if (gLedBusy == 1)  GPIOA->BSRR = GPIO_BSRR_BS_8;
#endif
}

void op_VIDPID(void)  // ***************************************************
// (VID PID -- )
// to avoid errors, VID should be entered in HEX (4 digits)
// be careful not to shoot your USB access (appropriate Windows .inf file !)
// effective VID is only changed after system restart
// *************************************************************************
{
uint32_t vid, pid;
const uint8_t VidStr [] = "VID = ";
const uint8_t PidStr [] = "PID = ";
	
	op_SWAP();
	vid = *gfSP;
    sendString(VidStr);
    op_PRINTHEX();			//drops TOS
    sendTB(SPACE);
	pid = *gfSP;
    sendString(PidStr);
    op_PRINTHEX();			//drops TOS
    sendTB(SPACE);

	if (conFirm() == 0) return;
	gVid = vid;
	gPid = pid;
	pushFlashPage23();
}


void op_QUEST(void)  // *****************************************************
// ( -- )
//  displays a list of active system coonfiguration
// **************************************************************************
{
uint32_t k, q, r;
const uint8_t stackStr [] = "Stack=";
const uint8_t threadStr [] = "CodeThread=";
//const uint8_t emptyStr [] = "Empty_";
const uint8_t flashStr [] = "Flash";
//const uint8_t projectStr [] = "Project";
const uint8_t autoStr [] = " loaded  AUTOEXE=";
const uint8_t clockStr [] = "SysClock=";
const uint8_t ledStr [] = "  LEDmode=";
#ifndef MINI
const uint8_t splitStr [] = "  Terminal_Split=";
const uint8_t baudStr [] = "MHz  BaudRate=";
const uint8_t channelStr [] = "  MidiCh=";
const uint8_t pendStr [] = "(pending)";
#endif
const uint8_t debugStr [] = "  BackOP_";

const uint8_t spiStr [] = "   SPI=";
const uint8_t freq12Str [] = "Freq12=";
const uint8_t pwm1Str [] = "Hz PWM1=";
const uint8_t pwm2Str [] = ", PWM2=";
const uint8_t freq3Str [] = "   Freq34=";
const uint8_t pwm3Str [] = "Hz PWM3=";
const uint8_t pwm4Str [] = ", PWM4=";
const uint8_t onStr [] = "ON";
const uint8_t offStr [] = "OFF";

const uint8_t mot1Str [] = "MOT1:";
const uint8_t mot2Str [] = "   MOT2:";
const uint8_t mot3Str [] = "   MOT3:";
const uint8_t tstepStr [] = "Step=";
const uint8_t nstepsStr [] = "ms N=";
const uint8_t dcmotStr [] = "DC/PWM";
#ifdef DMX
const uint8_t dmxStr [] = "DMX:";
const uint8_t dashStr [] = " ------";
const uint8_t addressStr [] = "Work on DMXchannel:";
const uint8_t fadetimeStr [] = "(FadeTime=";
const uint8_t cycleStr [] = "s Cycle=";
const uint8_t bootsceneStr [] = " BootScene=";
#endif

#ifdef MINI
	sendModel();
#endif
	sendTB('D');
	sendTB('a');
	sendTB('t');
	sendTB('a');
	sendString(stackStr);
	sendX32((uint32_t)(gfDataStack + 3));
	sendTB('R');
	sendTB('e');
	sendTB('t');
	sendTB('u');
	sendTB('r');
	sendTB('n');
	sendString(stackStr);
	sendX32((uint32_t)gfReturnStack);
	sendString(threadStr);
	sendX32((uint32_t)gfTCode);
	sendTB('U');
	sendTB('s');
	sendTB('e');
	sendTB('r');
	sendTB('=');
	sendX32((uint32_t)gfUserHeader);
	sendTB(CRR);
	if (gProject == 0)  sendString(gfEmptyStr);
	sendString(gfProjectStr);
	if (gProject != 0) {
		sendTB('#');
		sendTB(gProject + 0x30);
	}
	sendString(autoStr);
	if (gfAuto == 0 )  {
		sendTB('n');
		sendTB('o');
		sendTB('n');
		sendTB('e');
	}
	else  {	
		for (k = 0;k < 16;k++)  {
			if (gfUserHeader[gfAuto-1].name[k] == 0)  break;
			sendTB(gfUserHeader[gfAuto-1].name[k]);
		}
	}
	sendString(ledStr);
	sendTB(gLedBusy | 0x30);
#ifdef MINI
	sendString(debugStr);
	if (gfBackProcess == 0)  sendString(offStr);
	else  sendString(onStr);
	sendTB(CRR);

	sendString(clockStr);
	sendNS32(SystemCoreClock / 1000000);
	sendTB('M');
	sendTB('H');
	sendTB('z');
#else
	sendString(splitStr);
	sendTB(gSplit + 0x30);
	sendString(debugStr);
	if (gfBackProcess == 0)  sendString(offStr);
	else  sendString(onStr);
	sendTB(CRR);

	sendString(clockStr);
	sendNS32(SystemCoreClock / 1000000);
	sendString(baudStr);
	sendNS32(gTempBaud);
	if (gTempBaud == 31)  {
		sendTB('2');
		sendTB('5');
		sendTB('0');
	}
	else  {
		if (gTempBaud == 11)  {
			sendTB('5');
			sendTB('2');
		}
		if (gTempBaud == 19)  sendTB('2');
		if (gTempBaud == 38)  sendTB('4');
		if (gTempBaud == 57)  sendTB('6');
		sendTB('0');
		sendTB('0');
	}
	if ((gBaudSet == 0) && (gSplit != 1) && (gSplit != 3)) sendString(pendStr);
//gSplit=0,2,4: USART is Terminal, else USART is I/O, then immediate Baud change
	sendString(channelStr);
	if (gChannel == 0) {
	sendTB('a');
	sendTB('l');
	sendTB('l');
	}
	else sendNS32(gChannel);
#endif  //MINI or not
	sendString(spiStr);
	sendNS32(gfSpiMode);
	sendTB(',');
	if (gfTempNumBase==16)  {
		send0X32(gfSpiBits);
		sendTB(',');
		send0X32(gfSpiClock);
	}
	else  {
		sendNS32(gfSpiBits);
		sendTB(',');
		sendNS32(gfSpiClock);
	}
	sendTB(CRR);
	sendString(freq12Str);
    k = gfFreq12 * 2000;
	if (gfFreq12 <= 0) sendNS32(0);
	else  {
		q = 48000000 / k;
		sendNS32(q);
		r = (48000000 - (q * k)) * 10;
		sendTB('.');
        q = r / k;
		sendNS32(q);
		r = (r - (q * k)) * 10;
        q = r / k;
		r = (r - (q * k)) * 10;
        k = r / k;
        if (k >= 5) q++;
		sendNS32(q);
	}
	sendString(pwm1Str);
	if (gfPWM1 < 0) sendString(offStr);
	else {
		if (gfTempNumBase==16)  send0X32(gfPWM1);
		else  sendNS32(gfPWM1);
	}
	sendString(pwm2Str);
	if (gfPWM2 < 0) sendString(offStr);
	else {
		if (gfTempNumBase==16)  send0X32(gfPWM2);
		else  sendNS32(gfPWM2);
	}
	sendString(freq3Str);
    k = gfFreq34 * 2000;
	if (gfFreq34 <= 0) sendNS32(0);
	else  {
		q = 48000000 / k;
		sendNS32(q);
		r = (48000000 - (q * k)) * 10;
		sendTB('.');
        q = r / k;
		sendNS32(q);
		r = (r - (q * k)) * 10;
        q = r / k;
		r = (r - (q * k)) * 10;
        k = r / k;
        if (k >= 5) q++;
		sendNS32(q);
	}
	sendString(pwm3Str);
	if (gfPWM3 < 0) sendString(offStr);
	else {
		if (gfTempNumBase==16)  send0X32(gfPWM3);
		else  sendNS32(gfPWM3);
	}
	sendString(pwm4Str);
	if (gfPWM4 < 0) sendString(offStr);
	else {
		if (gfTempNumBase==16)  send0X32(gfPWM4);
		else  sendNS32(gfPWM4);
	}
	sendTB(CRR);
	
	sendString(mot1Str);
	if (gfUserHeader[4].toCompile == -1)  {
		if (GPIOA->ODR & 2) sendTB('-');			//PA1 High = CCW
		sendString(dcmotStr);
		sendTB('1');
	}
	else  {
		sendString(tstepStr);
		sendNS32(gfStepSpeed1);
		sendString(nstepsStr);
		sendNS32(gfUserHeader[4].toCompile);
	}
	sendString(mot2Str);
	if (gfUserHeader[5].toCompile == -1)  {
		if (GPIOA->ODR & 0x10) sendTB('-');			//PA4 High = CCW
		sendString(dcmotStr);
		sendTB('2');
	}
	else  {
		sendString(tstepStr);
		sendNS32(gfStepSpeed2);
		sendString(nstepsStr);
		sendNS32(gfUserHeader[5].toCompile);
	}
#ifndef MINI
	sendString(mot3Str);
	if (gfUserHeader[6].toCompile == -1)  {
		if (GPIOB->ODR & 0x80) sendTB('-');			//PB7 High = CCW
		sendString(dcmotStr);
		sendTB('3');
	}
	else  {
		sendString(tstepStr);
		sendNS32(gfStepSpeed3);
		sendString(nstepsStr);
		sendNS32(gfUserHeader[6].toCompile);
	}
#endif
	sendTB(CRR);

#ifdef DMX
	sendString(dmxStr);
	if (gfDmxBusy == 0)  sendString(offStr);
	else  sendString(onStr);
	sendString(dashStr);
	sendTB(CRR);
	sendString(addressStr);
	sendS32(gdDmxAddr + 1);				//added 13July20
	sendTB('T');
	sendTB('x');
	sendTB('=');
	sendS32(gdDMXBUF[gdDmxAddr]); 
	sendTB('F');
	sendTB('i');
	sendTB('n');
	sendTB('a');
	sendTB('l');
	sendTB('=');
	sendS32(gdFINALMEM[gdDmxAddr]);
	sendString(fadetimeStr);
	k = gdFadeTime/100;		//fadetimeH
	sendNS32(k);
	sendTB('.');
	sendNS32((gdFadeTime/10) - (k*10));
	sendString(cycleStr);
	sendNS32(gdDmxMax);
	sendString(bootsceneStr);
	sendNS32(gdBootScene);	
	sendTB(0x29);						//')'
	sendTB(CRR);	
#endif	
}

//const uint8_t gfProjectStr [] = " Project no.";
//const uint8_t gfEmptyStr [] = "(=Empty Project)";

void op_SAVEFLASH(void)  // ************************************************
// save this project state in microcontroller internal flash
// *************************************************************************
{
uint32_t k, n;
const uint8_t saveStr [] = "Save ";

    n = *gfSP;
	gfSP--;
#ifdef DMX	
	if (n > 2)  {
#else
	if (n > 5)  {
#endif
		gfErrors |= ERR_NUMVAL;
		return;
	}

	sendString(saveStr);
	if (n == 0) sendString(gfEmptyStr);
	sendString(gfProjectStr);
    if (n)  {
		sendTB('#');
		sendTB(n + 0x30);
	}
//	if (n == 0) sendString(gfEmptyStr);
	sendTB(' ');
	sendTB('?');

    if (conFirm() == 0) return;
    gProject = n;
    pushFlashPage23();	//essenially save "gProject", BaudRate, DMX fadeTime

    if (gProject) {
		if (pushFlashProject() == 0)  gfErrors |= ERR_USERCODE;
	}
}


void op_LOADFLASH(void)  // ************************************************
// reload previously saved project from microcontroller internal flash
// *************************************************************************
{
uint32_t k, n;
const uint8_t loadStr [] = "Load ";

    n = *gfSP;
	gfSP--;
#ifdef DMX	
	if (n > 2)  {
#else
	if (n > 5)  {
#endif
		gfErrors |= ERR_NUMVAL;
		return;
	}
	sendString(loadStr);
//	sendString(gfProjectStr);
//    sendTB(n + 0x30);
//	if (n == 0) sendString(gfEmptyStr);
	if (n == 0) sendString(gfEmptyStr);
	sendString(gfProjectStr);
    if (n)  {
		sendTB('#');
		sendTB(n + 0x30);
	}
	sendTB(' ');
	sendTB('?');

    if (conFirm() == 0) return;
	gProject = n;		//valid temporarily until next CForth start	
    if (gProject == 0)  {
		emptyForth();
		popFlashPage23();		//reload global setup
		gProject = 0;			//override existing global entry
	}
    else  popFlashProject();
#ifdef DMX
	if (gdBootScene)  {
		gfSP++;
		*gfSP = gdBootScene;
		op_DMXLOAD();
	}
#endif
}


uint32_t conFirm(void)  // *************************************************
//save some code length
// *************************************************************************
{
uint32_t k;
const uint8_t confirmStr [] = " confirm w.upper case 'Y' ";
const uint8_t cancelStr [] = "cancelled";

    sendString(confirmStr);
 	while ((k= pullTermRxBuf()) == 0);
 	if ((k & 0x000000FF) != 'Y')  {
        sendString(cancelStr);
        return 0;
 	}
	sendTB(CRR);
    return 1;
}

#ifdef DMX
//DMX related ops  #########################################################

void supplyBack(void)
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
}

void op_DMXS(void)	// *****************************************************
// (slot --   )
// set the DMX slot (channel) which is addressed for subsequent DMX actions
// *************************************************************************
{
uint32_t slot;

	supplyBack();

	slot = *gfSP;
	gfSP--;
	if ((slot > DMX_BUFLEN) || (slot == 0)) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	gdDmxAddr = (uint32_t)slot - 1;	
}


void op_DMXV(void)	// *****************************************************
// *************************************************************************
{
uint32_t level;

	supplyBack();

	level = *gfSP;
	gfSP--;
	if (level > 255) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	setDmxLevel(level);
}


void op_DMXRGB(void)	// *************************************************
// (RGBWord ---  )
// set the actually addressed DMX channel to bits 16...23 of RGBWord
// set (the actually addressed DMX channel + 1) to bits 8...15 of RGBWord
// set (the actually addressed DMX channel + 2) to bits 0..7 of RGBWord
// ignore bits 24...31 of RGBWord
// e.g:TOS = 0x007F00FF will set the RGB lamp to a shade of bluish magenta
// so a group of subsequently addressed RGB lamps can be set w.o. re-address
// *************************************************************************
{
uint32_t level, oldAddr;

	supplyBack();

	oldAddr = gdDmxAddr;
	level = *gfSP;
	gfSP--;
	setDmxLevel((level >> 16) & 0x000000FF);
	gdDmxAddr++;
	if (gdDmxAddr > 512) {
		gfErrors |= ERR_NUMVAL;
		gdDmxAddr = oldAddr;
		return;
	}
	setDmxLevel((level >> 8) & 0x000000FF);
	gdDmxAddr++;
	if (gdDmxAddr > 512) {
		gfErrors |= ERR_NUMVAL;
		gdDmxAddr = oldAddr;
		return;
	}
	setDmxLevel(level & 0x000000FF);
	gdDmxAddr = oldAddr;
}


void op_DMXNXTRGB(void)	// **************************************************
// (RGBWord ---  )
// PRE-INCREASE the selected DMX CHANNEL(slot) BY 3
// set the actually addressed DMX channel to bits 16...23 of RGBWord
// set (the actually addressed DMX channel + 1) to bits 8...15 of RGBWord
// set (the actually addressed DMX channel + 2) to bits 0..7 of RGBWord
// ignore bits 24...31 of RGBWord
// e.g:TOS = 0x007F00FF will set the RGB lamp to a shade of bluish magenta
// so a group of subsequently addressed RGB lamps can be set w.o. re-address
// *************************************************************************
{
uint32_t level, oldAddr;

	supplyBack();

	level = *gfSP;
	gfSP--;
	oldAddr = gdDmxAddr;
	gdDmxAddr += 3;
	if (gdDmxAddr > 512) {
		gfErrors |= ERR_NUMVAL;
		gdDmxAddr = oldAddr;
		return;
	}
	oldAddr = gdDmxAddr;
	setDmxLevel((level >> 16) & 0x000000FF);
	gdDmxAddr++;
	if (gdDmxAddr > 512) {
		gfErrors |= ERR_NUMVAL;
		gdDmxAddr = oldAddr;
		return;
	}
	setDmxLevel((level >> 8) & 0x000000FF);
	gdDmxAddr++;
	if (gdDmxAddr > 512) {
		gfErrors |= ERR_NUMVAL;
		gdDmxAddr = oldAddr;
		return;
	}
	setDmxLevel(level & 0x000000FF);
	gdDmxAddr = oldAddr;
}


void op_DMXSET(void)	// **************************************************
// (slot level --   )
// first address DMX slot(channel) and set it to 'level'
// **************************************************************************
{
uint32_t level;

	level = *gfSP;
	gfSP--;
	if (level > 255) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	op_DMXS();
	setDmxLevel(level);
}


void op_DMXNEXT(void)	// *************************************************
// PRE-INCREASE the selected DMX CHANNEL(slot)
// *************************************************************************
{
uint32_t level;

	level = *gfSP;
	gfSP--;
	if ((level > 255) || (gdDmxAddr >= (DMX_BUFLEN - 1)))  { 
		gfErrors |= ERR_NUMVAL;
		return;
	}		
	gdDmxAddr++ ;
	setDmxLevel(level);
}


void op_DMXEQ(void)	// *****************************************************
//(number  --   )
// *************************************************************************
{
uint32_t k, number, address;
uint8_t level;

	number = *gfSP;
	gfSP--;
	if ((number == 0) || (number > (DMX_BUFLEN - gdDmxAddr))) {
		gfErrors |= ERR_NUMVAL;
		return;
	}		
	level = gdFINALMEM[gdDmxAddr];
	address = gdDmxAddr;
	for (k=0;k<number;k++)  {
		gdDmxAddr++;
		setDmxLevel(level);
		if (gdDmxAddr == (DMX_BUFLEN - 1))  break;
	}
	gdDmxAddr = address;				//store initial value back
}


void op_DMXFT(void)	// *****************************************************
//enter fade time in 1/10s units. Internally fade process updated with 100Hz
// *************************************************************************
{
uint32_t time;

	supplyBack();

	time = *gfSP;
	gfSP--;
	if (time > 127) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	gdFadeTime = (uint32_t)time * 10;
}


void op_DMXPLUS(void)	// *************************************************
// *************************************************************************
{
	if (gdDMXBUF[gdDmxAddr] != 0xFF) {
		gdDMXBUF[gdDmxAddr]++ ;
		gdFINALMEM[gdDmxAddr]++ ;
	}
}


void op_DMXMINUS(void)	// *************************************************
// *************************************************************************
{
	if (gdDMXBUF[gdDmxAddr] != 0x00) {
			gdDMXBUF[gdDmxAddr]-- ;
			gdFINALMEM[gdDmxAddr]-- ;
	}
}


void op_DMXDARK(void)	// *************************************************
// (  --  )
// set all DMX channels to level 0
// *************************************************************************
{
uint32_t k;

	for (k=0; k < DMX_BUFLEN; k++)  {	
		gdDMXBUF[k] = 0 ;
		gdFINALMEM[k] = 0 ;
	}
}


void op_DMXFLASH(void)	// *************************************************
// (scene time   --  )
// during "time" (deci-seconds !!) set all DMX channels to flash preset "scene"
//	then restore previous DMX lighting situation
// *************************************************************************
{
uint32_t k, fadetime;
	

	k = *(gfSP-1);						//perform "SWAP"
	*(gfSP-1) = *gfSP;
	*gfSP = k;							//TOS holds "scene", NOS holds "time"

	for (k=0; k < DMX_BUFLEN; k++)  {	
		gdDMXTEMP[k] = gdFINALMEM[k] ;
	}

	fadetime = gdFadeTime;
	gdFadeTime = 0;
	op_DMXLOAD();						//load immediately
	gdFadeTime = fadetime;
	if (gfErrors) return;

	*gfSP *= 100;						//deciSeconds = milliSeconds * 100;
	op_MS();
	 
	for (k=0; k < DMX_BUFLEN; k++)  {	
		gdFINALMEM[k] = gdDMXTEMP[k];
		gdDMXBUF[k] = gdDMXTEMP[k];		//previously active fade cancelled
	}
}


void op_DMXSAVE(void)	// *************************************************
//save lighting scene into Flash
//	(scene -- ), scene = 1..96, internally scaled 0..95
// *************************************************************************
{
uint32_t k, scene, page, offset;
uint8_t *ptr;

	scene = *gfSP - 1;
	gfSP--;
	if (scene > 95) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	page = scene >> 2;						//0,1,2,...15
	offset = (scene - (page<<2)) * 512;		//0,512,1024,1536	
	ptr = (uint8_t *)(PAGE40 + (page*2048));	
	for (k=0; k<2048;k++)  {
		gfTempBuf[k] = ~(*ptr);
		ptr++;
	}	
	for (k=0;k < 512;k++)  {	
		gfTempBuf[offset+k] = gdDMXBUF[k];
	}
	pushDmxFlash(page);
}


void op_DMXCYCL(void)	// *************************************************
// set length of DMX transmit cycle
//	(cycle -- )
//	(24 <=cycle <= 512)
// *************************************************************************
{
uint32_t cycle;

	cycle = *gfSP;
	gfSP--;
	if ((cycle < 24)||(cycle > 512)) {//first check then change//BUGFIX21Nov21
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	gdDmxMax = cycle;
}


void op_DMXBOOT(void)	// *************************************************
// enter which scene shall be loaded automatically at system start
//	(scene -- )
//	scene==0: dark scene, all channels black. (1 <=scene <=96) load from Flash
// *************************************************************************
{
uint32_t scene;

	scene = *gfSP;
	gfSP--;
	if (scene >96) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	gdBootScene = scene;
	pushFlashPage23();			//stored as global parameter
}


void op_DMXLOAD(void)	// *************************************************
//	(scene -- )
//	scene==0: dark scene, all channels black. (1 <=scene <=96) load from Flash
// *************************************************************************
{
uint32_t k, scene, page, offset;
uint32_t * flashPtr;
uint32_t * dmxPtr;

	supplyBack();

	scene = *gfSP;
	gfSP--;
	if (scene > 96) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
	if (scene == 0)  {
		for (k=0; k < DMX_BUFLEN;k++)  gdFINALMEM[k] = 0;//"dark" scene
	}
	else  {
		scene--;
		page = scene >> 2;								//0,1,2,...15
		offset = (scene - (page<<2)) * 512;				//0,512,1024,1536	
		flashPtr = (uint32_t *)(PAGE40 + (2048 * page) + offset);
		dmxPtr = (uint32_t *) gdFINALMEM;
		for (k=0; k < 128;k++)  {
			*dmxPtr = ~(*flashPtr);
			flashPtr++;
			dmxPtr++;
		}
	}
	for (k=0;k < DMX_BUFLEN;k++)  {
		if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
		if (gdFadeTime == 0)  gdDMXBUF[k] = gdFINALMEM[k];
		else startFader(k);
	}

}


void op_DMXLDPART(void)	// *************************************************
//(scene startAddr number -- )
//load number bytes from Flash offeset = (scene-1)*2048 + startAddr 
//to active DMX output starting from selected DMX Slot XS++ 
// *************************************************************************
{
uint32_t k, scene, number,startAddr;
uint32_t page, offset;
uint8_t *ptr, *ptrMax;
	
	number = *gfSP;
	startAddr = *(gfSP - 1) -1;		//DMX channel n is startAddr 0
	scene = *(gfSP - 2);			//user 1..96, internal 0..95
	gfSP -= 3;

	if ((number==0)||(number>(DMX_BUFLEN-gdDmxAddr)) ||
								(startAddr>=DMX_BUFLEN) || (scene > 96)) {
		gfErrors |= ERR_NUMVAL;
		return;
	}		

	if (scene == 0)  {
		for (k=gdDmxAddr;k < (gdDmxAddr + number);k++)  {
			gdFINALMEM[k] = 0;
			if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
			if (gdFadeTime == 0) gdDMXBUF[k] = gdFINALMEM[k];
			else startFader(k);
			if (k == DMX_BUFLEN)  break;
		}
	}
	else  {
		scene--;		
		page = scene >> 2;								//0,1,2,...15
		offset = (scene - (page<<2)) * 512;				//0,512,1024,1536	
		ptr = (uint8_t *)(PAGE40 + (2048 * page) + offset);
		ptrMax = (uint8_t *)(ptr + DMX_BUFLEN); 
		ptr += startAddr;
		
		for (k=gdDmxAddr;k < (gdDmxAddr + number);k++)  {
			gdFINALMEM[k] = ~(*ptr);
			ptr++;
			if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
			if (gdFadeTime == 0) gdDMXBUF[k] = gdFINALMEM[k];
			else startFader(k);
			if (k == DMX_BUFLEN)  break;
			if (ptr == ptrMax)  break;
		}
	}
}


void op_DMXR(void)	// *****************************************************
//(number --   )
//send the levels of number DMX channeles starting from selected DMX channel
// *************************************************************************
{
//uint32_t address;
uint32_t number, address;
const uint8_t readStr [] = "Read from S=";

	number = *gfSP;
	gfSP--;
	if (number > DMX_BUFLEN) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
//	sendTB(CRR);
	sendString(readStr);
	sendNS32(gdDmxAddr + 1);
	sendTB(':');
	sendTB(SPACE);
	sendTB('V');
	sendTB('=');
	address = gdDmxAddr;
	sendNS32(gdDMXBUF[address]);
	address++ ;
	number -- ;
	while (1)  {
		if ((address == DMX_BUFLEN) || (number== 0)) break;
		sendTB(',');
		sendNS32((int32_t)gdDMXBUF[address]);
		address++;
		number--;
	}
	sendTB(CRR);
}


void op_DMXBUS(void)  // ******************************************************
// (DmxBusy -- )
// *************************************************************************
{
	if (gfExecUser && gfBackNumValid)  {
		gfBackNumValid = 0;
		gfSP++;
		*gfSP = gfBackProcData;
	}
	gfDmxBusy = *gfSP;
	gfSP--;
    if (gfDmxBusy > 1) {
		gfErrors |= ERR_NUMVAL;
		return;
	}	
}

#endif //DMX

// ### END #################################################################

